/*	NAME:
		jGNEFilter.c

	WRITTEN BY:
		Dair Grant
				
	DESCRIPTION:
		This file contains a code resource to be installed as a jGNEFilter.

	NOTES:
		We intercept each keyDown event, and draw the relevent character to
		the right of the Apple menu.

	___________________________________________________________________________
*/
//=============================================================================
//		Include files																	 
//-----------------------------------------------------------------------------
#include <Gestalt.h>
#include "A4Stuff.h"
#include "SetupA4.h"
#include "ES Address Table.h"
#include "WK Constants.h"















//=============================================================================
//		Global Variables																 
//-----------------------------------------------------------------------------
ProcPtr				gNextFilter;					// The address of the next filter
Boolean				gAlreadyRan = false;			// Have we already been called?
GrafPort			gMyPortRec;						// A port to draw in
GrafPtr				gMyPort;						// Pointer to gMyPortRect
ESAddressTable		*gTheAddressTable;				// Default Extension Shell Address Table















//=============================================================================
//		Private function prototypes							 
//-----------------------------------------------------------------------------
void		main(void);
ProcPtr		DojGNEFilter(EventRecord *theEvent : __A1) : __A0;















//=============================================================================
//		main : Entry point to our code resource.
//-----------------------------------------------------------------------------
//		Note :	We save some scratch registers, call DojGNEFilter, restore the
//				registers, then jump to the next filter.
//
//				DojGNEFilter takes the EventRecord in A1, and returns the
//				address of the next filter in the chain in A0. Since A1 is the
//				Event Record on entry, we don't need to do anything special to
//				set it up.
//-----------------------------------------------------------------------------
asm void main(void)
{

	// Save some registers, and call DojGNEFilter to do the actual filter.
	// The EventRecord we're looking at is in A1 when we're called,
	// which is where CWrapper expects it.
	movem.l		d0-d2/a1-a2, -(a7)
	jsr			DojGNEFilter
	
	
	// Restore the registers, and jump to A0, which DojGNEFilter has left
	// as the address of the next filter in the chain.
	movem.l	(a7)+, d0-d2/a1-a2
	jmp		(a0)
}















//=============================================================================
//		DojGNEFilter : The C jGNE Filter.
//-----------------------------------------------------------------------------
//		Note :	Written in C for convenience.
//
//				We first check to see if we've ever been called before - if
//				we haven't, we initialise our global variables accordingly.
//
//				For filtering events, we watch for keyDown events. We then
//				draw the character for the keypress in a GrafPort to the right
//				of the Apple menu.
//-----------------------------------------------------------------------------
ProcPtr DojGNEFilter(EventRecord *theEvent : __A1) : __A0
{	long			oldA4;
	GrafPtr			savePort;
	ProcPtr			nextFilter;
	Rect			theRect;

	

	// Set up A4 so that we can access our globals
#ifndef powerc
	oldA4 = SetCurrentA4();
#endif



	// Do our one time initialisation stuff
  	if (!gAlreadyRan)
		{
		// Get the address of the next thing on the chain
		Gestalt(kWhatKeyAddressTable, (long *) &gTheAddressTable);		
		gNextFilter = (ProcPtr) ((long) gTheAddressTable->theTable[kWhatKey]);
		
		
		// Save the current port, then open ours to intialise it
		GetPort(&savePort);
		gMyPort = &gMyPortRec;
		OpenPort(gMyPort);
		SetPort(savePort);

		gAlreadyRan = true;
		}




	// Examine the event - we only keyDown and autoKey events. If we
	// wanted to catch other events, this is where we would do it.
	// If we wanted to swallow events, we would set theEvent->what to
	// a nullEvent.
	switch (theEvent->what) {
		case keyDown:
		case autoKey:
			// Set things up for drawing
			GetPort(&savePort);
			SetPort(gMyPort);
			TextFont(geneva);
			TextSize(9);
			
			
			// Wipe out anything that's there already
			SetRect(&theRect, 5, 5, 15, 15);
			EraseRect(&theRect);
			
			
			// Draw the character
			MoveTo(5, 12);
			DrawChar(theEvent->message & charCodeMask);
			
			
			// Restore things
			SetPort(savePort);
			break;
			
		default:
			break;
	}

	


	
	// Put the address of the next filter into a local variable,
	// restore A4, and return the address of the next filter.
	nextFilter = gNextFilter;
#ifndef powerc
	SetA4(oldA4);
#endif
	return(nextFilter);
}
