#pragma once
/*****
 *
 *	MyConfiguration.h
 *
 *	You will want to modify the constants in this header file to
 *	set the code to compile with the particular options you want.
 *	Use 1 (one) to turn an option on, 0 (zero) to turn it off.
 *
 *	This is a support file for "Grant's CGI Framework".
 *	Please see the license agreement that accompanies the distribution package
 *	for licensing details.
 *
 *	Copyright 1995 by Grant Neufeld
 *	grant@acm.com
 *	http://arpp1.carleton.ca/grant/
 *
 *****/

/***  CONSTANT DECLARATIONS  ***/

/**  Threads  **/

/* if off, the thread manager must be present for the app to run */
#define kCompileWithThreadsOptional		1
/* if on, all Apple Events (including the CGI sdoc event) will be
	spawned off as separate threads. If off, you are on your own for
	threading the CGI events. */
#define kCompileWithThreadedAppleEvents	0


/**  Quitting  **/

/* if on, app will try to quit after a set time of idling */
#define kCompileWithQuitOnLongIdle		1
/* if on, app will not quit on idle if it was opened with no special
	event - starts off by receiving an OpenApplication AppleEvent */
#define kCompileWithIdleQuitOnOpenApp	1
/* if memory gets 'precariously' low, the application will be set to quit.
	This _may_ (note: untested) speed up the application. Needs profiling */
#define kCompileWithQuitOnLowMemory		1


/**  Modules  (CGI/ListSTAR)  **/

/* the following determine which modules are compiled */
/* CGI */
#define kCompileWithCGICode				1
/* ListSTAR */
#define kCompileWithListSTARCode		0


/**  CGI Support  **/

/* Send-partial event support (CGIAESendPartial).
	Also includes the connection field of the CGIRecord. */
#define kCompileWithCGISendPartial		1
/* Form handling code (all the CGIFormFields... functions).
	If you turn this off, the formFields and totalFields will not
	be available in the CGIHandles. */
#define kCompileWithCGIFormHandling		1
/* automatically format form data before calling MyCGIProcess. */
#define kCompileWithCGIFormAutoProcess	1
/* automatically decode parameters (convert %xx values, and special chars)
	Parameters affected: path_args, http_search_args */
#define kCompileWithCGIAutoDecode		1

/* support use as a user defined action */
#define kCompileWithCGIActionSupport	1

/**  CGI Parameters  **/

/* '----' path_args			*/
#define kCompileWithCGIpath_args		1
/* 'kfor' http_search_args	*/
#define kCompileWithCGIhttp_search_args	1
/* 'user' username			*/
#define kCompileWithCGIusername			1
/* 'pass' password			*/
#define kCompileWithCGIpassword			1
/* 'frmu' from_user			*/
#define kCompileWithCGIfrom_user		1
/* 'addr' client_address	*/
#define kCompileWithCGIclient_address	1
/* 'post' post_args			*/
#define kCompileWithCGIpost_args		1
/* 'meth' method			*/
#define kCompileWithCGImethod			1
/* 'svnm' server_name		*/
#define kCompileWithCGIserver_name		1
/* 'svpt' server_port		*/
#define kCompileWithCGIserver_port		1
/* 'scnm' script_name		*/
#define kCompileWithCGIscript_name		1
/* 'ctyp' content_type		*/
#define kCompileWithCGIcontent_type		1
/* 'refr' referer			*/
#define kCompileWithCGIreferer			1
/* 'Agnt' user_agent		*/
#define kCompileWithCGIuser_agent		1
/* 'Kcip' client_ip			*/
#define kCompileWithCGIclient_ip		1
/* 'Kfrq' full_request		*/
#define kCompileWithCGIfull_request		1
/* 'Pvrs' version			*/
#define kCompileWithCGIversion			1


/**  Interface  **/

/* setting this to off will create a background only application with
	no user interface. The application won't show up in the
	application menu. */
#define kCompileWithForeground			1
#if kCompileWithForeground /* only allow the options below if foreground */

/* setting this on will allow user interface elements such as error 
	dialogs to pop up. Turn it off if you want the application to be able
	to run without requiring any user interaction. User interface elements
	such as menus will still be available (providing CompileWithForeground
	is on). */
#define	kCompileWithFullUserInterface	0	
/* display the splash screen on startup */
#define kCompileWithSplashScreen		0
/* application uses modeless dialogs */
#define kCompileWithModelessDialogs		0
/* application uses custom windows */
#define kCompileWithApplicationWindows	0
/* application handles custom keyboard entry */
#define kCompileWithKeyboardEvents		0
/* support drag and drop manager */
#define kCompileWithDragNDrop			0

#endif /* kCompileWithForeground */


/***  PROCESS / APPLICATION  ***/

/* If this is defined, the application will have a global 'gProcessFSSpec'
	that is automatically, during startup, set to be the application's
	file spec. */
#define kCompileWithProcessFileSpec		1


/***  MEMORY CONFIGURATION  ***/

/* You will need to estimate your memory requirements here */

/* 'comfort' zone of free memory. This must be big enough to allow
	any essential functions to be called.
	This is used as an 'emergency' reserve of memory to be
	deallocated if seriously needed.
	Named kMemCushion in IM-Memory: 1-43, 1-36 */
#define kMemCushionSize			8192L /* 8k */
/* for pre-allocation of master pointers.
	Used in initAppMemory "Main.c"
	IM-Memory: 1-42 */
#define	kMoreMasterCalls		4
/* minimum free space needed (in bytes) for application to be able to run.
	Should probably be a multiple of 1024.
	Should be bigger than kMemChusionSize */
#define kMinSegSize				32768L /* 32k */


/**  profiler support  **/

/* max 21 chars in length.
	End result will be something like "\p<name>-68K.prof" or "...-PPC.prof" */
#define kProfileNameStr			"cgi"

/**  Debbugging Support  **/

/* if you are debugging your code, you should set assertions on (1).
	If you are shipping your application, you should turn assertions off (0). */
#define kCompileWithAssertions	1


/***  EOF  ***/
