
#include	"TransSkel.h"
#include	"EventLog.h"


typedef enum
{
	iPushOK = 1,
	iPushCancel,
	iStaticTextTitle,
	iCheckBackground,
	iCheckTerminate,
	iOutline
};


static pascal void OutlineButton (DialogPtr dlog, short item);


/*
 * Set up a variable to point to the drawing procedure.  For 68K code this
 * is just a direct pointer to OutlineButton().  For PowerPC code it is a
 * routine descriptor into which the address of OutlineButton() is stuffed.
 */

# if skelPPC		/* PowerPC code */

static RoutineDescriptor	drawDesc =
		BUILD_ROUTINE_DESCRIPTOR(uppUserItemProcInfo, OutlineButton);
static UserItemUPP	drawProc = (UserItemUPP) &drawDesc;

# else				/* 68K code */

static UserItemUPP	drawProc = OutlineButton;

# endif


/*
 * Draw heavy outline around default dialog button.  This function is
 * associated with the user item when there is a default button, and
 * is called by ModalDialog() when the outline needs redrawing.
 */

static pascal void
OutlineButton (DialogPtr dlog, short item)
{
	SkelDrawButtonOutline (SkelGetDlogCtl (dlog, iPushOK));
}


/*
 * Install outlining procedure to associate outline user item with
 * the given button and draw the outline item immediately.
 */

static void
InstallOutliner (DialogPtr dlog, short item)
{
Rect	r;

	SkelGetDlogRect (dlog, item, &r);			/* get button rect */
	InsetRect (&r, -4, -4);						/* expand it */
	SkelSetDlogRect (dlog, iOutline, &r);		/* use for outline item */
	SkelSetDlogProc (dlog, iOutline, drawProc);
	SkelDrawButtonOutline (SkelGetDlogCtl (dlog, iPushOK));
}


/*
 * Remove outlining procedure from current default and erase the
 * current outline.
 */

static void
RemoveOutliner (DialogPtr dlog)
{
	SkelSetDlogProc (dlog, iOutline, nil);
	SkelEraseButtonOutline (SkelGetDlogCtl (dlog, iPushOK));
}


void
LaunchSetupDialog (void)
{
ModalFilterUPP	filter;
DialogPtr	dlog;
GrafPtr	savePort;
short	item;
short	value;
short	hilite;
short	loop;

	dlog = GetNewDialog (launchSetupDlog, nil, (WindowPtr) -1L);
	if (dlog == (DialogPtr) nil)
	{
		SysBeep (1);
		return;
	}

	SkelPositionWindow (dlog, skelPositionOnMainDevice, horizRatio, vertRatio);

	GetPort (&savePort);
	SetPort (dlog);

	InstallOutliner (dlog, iPushOK);
	SkelSetDlogCtlValue (dlog, iCheckBackground, launchInBackground);
	SkelSetDlogCtlValue (dlog, iCheckTerminate, terminateOnLaunch);

	ShowWindow (dlog);

	loop = 1;
	while (loop)
	{
		/*
		 * Get the standard filter and specify the default and cancel buttons
		 * to map return/enter and command-period/escape.
		 */
		filter = SkelDlogFilter (nil, false);
		SkelDlogDefaultItem (iPushOK);
		SkelDlogCancelItem (iPushCancel);
		ModalDialog (filter, &item);
		SkelRmveDlogFilter ();

		switch (item)
		{
		case iPushOK:
			loop = 0;
			launchInBackground = SkelGetDlogCtlValue (dlog, iCheckBackground);
			terminateOnLaunch = SkelGetDlogCtlValue (dlog, iCheckTerminate);
			break;
		case iPushCancel:
			loop = 0;
			break;
		case iCheckBackground:
		case iCheckTerminate:
			(void) SkelToggleDlogCtlValue (dlog, item);
			break;
		}
	}
	DisposeDialog (dlog);
	SetPort (savePort);
}
