/****************************************************************************
 **                                                                        **
 **  NAME:           Keith Pomakis                                         **
 **  STUDENT NUMBER: 93801177                                              **
 **                                                                        **
 **                         CS688 Final Project                            **
 **                                                                        **
 **                              "Connect"                                 **
 **                                                                        **
 **                          December 9,  1993                             **
 **                                                                        **
 **  This file contains the main routine for the project.  It sets up all  **
 **  of the appropriate data structures and gl modes.  It contains the     **
 **  event loop which is responsible for polling for input when neither    **
 **  robot arm is in action (note that when an arm is in action, the       **
 **  input is not polled from within this event loop, but from the         **
 **  function which animates the arm).  It also contains the logic as to   **
 **  whether or not a game is in progress, who's turn it is, etc.          **
 **                                                                        **
 **  See the README file for further information.                          **
 **                                                                        **
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gl/gl.h>
#include <gl/device.h>
#include "tb_handler.h"
#include "world.h"
#include "c4.h"
#include "arms.h"
#include "chips.h"
#include "menus.h"

#define WINDOW_TITLE    "Connect"

int desired_move = -1;
Boolean get_event(void);

/* Define some local variables. */

static long win_id;
static int num_of_players = 1;
static int first_player = 0;
static int num_to_connect = 4;
static int skill_levels[2] = { 5, 5 };
static int current_player;
static Boolean first_game = TRUE;
static Boolean game_in_progress = FALSE;
static Boolean in_tb_mode = FALSE;
static Boolean thinking = FALSE;

/* Define some local functions. */

static void setup(void);
static void get_options(int argc, char *argv[]);
static int topmost_in_column(int column);
static Boolean drop_chip(int player, int column, int row);
static void game_won(int player);
static void game_tied(void);
static void start_game(void);
void make_announcement(void);


/****************************************************************************/
/**                                                                        **/
/**  This in the main function of the project.  It reads the command line  **/
/**  to set up any requested state, sets up all of the required data       **/
/**  structures and gl modes, and enters an event loop which polls for     **/
/**  input.  This event loop contains the logic as to whether or not a     **/
/**  game is in progress, who's turn it is, etc.                           **/
/**                                                                        **/
/****************************************************************************/

int
main(int argc, char *argv[])
{
    int row, column;

    get_options(argc, argv);
    setup();

    make_announcement();

    /* Event loop. */

    while(TRUE) {

        /* Get the next input event. */

        get_event();
        if (!game_in_progress) continue;

        if (current_player == 0) {

            /* If it is the Red player's turn, the red player is a human, */
            /* and a move was selected, make the selected move and check  */
            /* for a win or a tie.                                        */

            if (num_of_players > 0) {
                if (desired_move >= 0) {
                    if (make_move(0, desired_move)) {
                        row = topmost_in_column(desired_move);
                        if (drop_chip(0, desired_move, row)) {
                            if (is_winner(0)) game_won(0);
                            else if (is_tie()) game_tied();
                            else {
                                current_player = 1;
                                make_announcement();
                            }
                        }
                    }
                    else
                        ringbell();
                    desired_move = -1;
                }
            }

            /* If it is the Red player's turn and the red player is a   */
            /* computer, then automatically select a move at the proper */
            /* skill level, make it, and check for a win or a tie.      */

            else {
                desired_move = -1;
                thinking = TRUE;
                make_announcement();
                disable_end_game();
                column = automatic_move(0, skill_levels[0]);
                enable_end_game();
                thinking = FALSE;
                make_announcement();
                row = topmost_in_column(column);
                if (drop_chip(0, column, row)) {
                    if (is_winner(0)) game_won(0);
                    else if (is_tie()) game_tied();
                    else {
                        current_player = 1;
                        make_announcement();
                    }
                }
            }
        }
        else {

            /* If it is the Blue player's turn, the blue player is a      */
            /* human, and a move was selected, make the selected move and */
            /* check for a win or a tie.                                  */

            if (num_of_players > 1) {
                if (desired_move >= 0) {
                    if (make_move(1, desired_move)) {
                        row = topmost_in_column(desired_move);
                        if (drop_chip(1, desired_move, row)) {
                            current_player = 0;
                            if (is_winner(1)) game_won(1);
                            else if (is_tie()) game_tied();
                            else {
                                current_player = 0;
                                make_announcement();
                            }
                        }
                    }
                    else {
                        ringbell();
                    }
                    desired_move = -1;
                }
            }

            /* If it is the Blue player's turn and the blue player is a */
            /* computer, then automatically select a move at the proper */
            /* skill level, make it, and check for a win or a tie.      */

            else {
                desired_move = -1;
                thinking = TRUE;
                make_announcement();
                disable_end_game();
                column = automatic_move(1, skill_levels[1]);
                enable_end_game();
                thinking = FALSE;
                make_announcement();
                row = topmost_in_column(column);
                if(drop_chip(1, column, row)) {
                    if (is_winner(1)) game_won(1);
                    else if (is_tie()) game_tied();
                    else {
                        current_player = 0;
                        make_announcement();
                    }
                }
            }
        }
    }

	return 0;
}


/****************************************************************************/
/**                                                                        **/
/**  This function polls for an input event and takes the appropriate      **/
/**  action if one exists.  It only accepts input that is legal for the    **/
/**  current mode.  FALSE is returned if the event was END_GAME.           **/
/**  otherwise TRUE is returned.                                           **/
/**                                                                        **/
/****************************************************************************/

Boolean
get_event(void)
{
    short event;

    /* If an event is waiting, process it. */

    if (qtest()) {
        switch(qread(&event)) {
            case QKEY:
                winclose(win_id);
                exit(EXIT_SUCCESS);
            case LEFTARROWKEY:
                if (event) {
                    ResetTrackBallMatrices();
                    red_lookat();
                    draw_world();
                }
                break;
            case RIGHTARROWKEY:
                if (event) {
                    ResetTrackBallMatrices();
                    blue_lookat();
                    draw_world();
                }
                break;
            case UPARROWKEY:
                if (event) {
                    ResetTrackBallMatrices();
                    top_lookat();
                    draw_world();
                }
                break;
            case DOWNARROWKEY:
                if (event) {
                    ResetTrackBallMatrices();
                    bottom_lookat();
                    draw_world();
                }
                break;
            case GKEY:
                if (event && !game_in_progress) {
                    if (!first_game) init_chip_positions();
                    start_game();
                };
                break;
            case ONEKEY:
                if (event && game_in_progress) desired_move = 0;
                break;
            case TWOKEY:
                if (event && game_in_progress) desired_move = 1;
                break;
            case THREEKEY:
                if (event && game_in_progress) desired_move = 2;
                break;
            case FOURKEY:
                if (event && game_in_progress) desired_move = 3;
                break;
            case FIVEKEY:
                if (event && game_in_progress) desired_move = 4;
                break;
            case SIXKEY:
                if (event && game_in_progress) desired_move = 5;
                break;
            case SEVENKEY:
                if (event && game_in_progress) desired_move = 6;
                break;
            case LEFTMOUSE:
                if (event && game_in_progress) desired_move = get_pick();
                break;
            case RIGHTMOUSE:
                if (event) switch (menu_choice()) {
                    case RED_VIEW:
                        ResetTrackBallMatrices();
                        red_lookat();
                        draw_world();
                        break;
                    case BLUE_VIEW:
                        ResetTrackBallMatrices();
                        blue_lookat();
                        draw_world();
                        break;
                    case TOP_VIEW:
                        ResetTrackBallMatrices();
                        top_lookat();
                        draw_world();
                        break;
                    case BOTTOM_VIEW:
                        ResetTrackBallMatrices();
                        bottom_lookat();
                        draw_world();
                        break;
                    case CONNECT_3:
                        num_to_connect = 3;
                        make_announcement();
                        break;
                    case CONNECT_4:
                        num_to_connect = 4;
                        make_announcement();
                        break;
                    case CONNECT_5:
                        num_to_connect = 5;
                        make_announcement();
                        break;
                    case RED_LEVEL_4:
                        skill_levels[0] = 4;
                        break;
                    case RED_LEVEL_5:
                        skill_levels[0] = 5;
                        break;
                    case RED_LEVEL_6:
                        skill_levels[0] = 6;
                        break;
                    case RED_LEVEL_7:
                        skill_levels[0] = 7;
                        break;
                    case RED_LEVEL_8:
                        skill_levels[0] = 8;
                        break;
                    case BLUE_LEVEL_4:
                        skill_levels[1] = 4;
                        break;
                    case BLUE_LEVEL_5:
                        skill_levels[1] = 5;
                        break;
                    case BLUE_LEVEL_6:
                        skill_levels[1] = 6;
                        break;
                    case BLUE_LEVEL_7:
                        skill_levels[1] = 7;
                        break;
                    case BLUE_LEVEL_8:
                        skill_levels[1] = 8;
                        break;
                    case NO_PLAYERS:
                        num_of_players = 0;
                        make_announcement();
                        break;
                    case ONE_PLAYER:
                        num_of_players = 1;
                        make_announcement();
                        break;
                    case TWO_PLAYERS:
                        num_of_players = 2;
                        make_announcement();
                        break;
                    case RED_FIRST:
                        first_player = 0;
                        break;
                    case BLUE_FIRST:
                        first_player = 1;
                        break;
                                        case SLOW_ARM_SPEED:
                                                slow_arm_speed();
                                                break;
                                        case MEDIUM_ARM_SPEED:
                                                medium_arm_speed();
                                                break;
                                        case FAST_ARM_SPEED:
                                                fast_arm_speed();
                                                break;
                    case TRACKBALL_MODE:
                        in_tb_mode = TRUE;
                        make_announcement();
                        qdevice(ESCKEY); qdevice(MIDDLEMOUSE);
                        while (in_tb_mode)
                            if (qtest())
                                switch(qread(&event)) {
                                    case LEFTMOUSE:
                                    case MIDDLEMOUSE:
                                    case RIGHTMOUSE:
                                        TrackBallMouseHandler();
                                        break;
                                    case ESCKEY:
                                        in_tb_mode = FALSE;
                                        break;
                                }
                        unqdevice(ESCKEY); unqdevice(MIDDLEMOUSE);
                        make_announcement();
                        break;
                    case START_GAME:
                        if (!first_game) init_chip_positions();
                        start_game();
                        break;
                    case END_GAME:
                        init_chip_positions();
                        end_game();
                        game_in_progress = FALSE;
                        use_main_menu();
                        make_announcement();
                        return FALSE;
                        break;
                    case QUIT:
                        winclose(win_id);
                        exit(EXIT_SUCCESS);
                        break;
                }
                break;
            case REDRAW:
                reshapeviewport();
                my_perspective();
                draw_world();
                break;
        }
    }

    return TRUE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function does exactly the same thing as get_event, except        **/
/**  returns void.  This is used as the polling function for the connect 4 **/
/**  algorithm, which is required to return void.                          **/
/**                                                                        **/
/****************************************************************************/

static void
poll_event(void)
{
    get_event();
}


/****************************************************************************/
/**                                                                        **/
/**  This function selects the gl modes necessary for this program, queues **/
/**  the required input events, and sets up all of the required data       **/
/**  structures.                                                           **/
/**                                                                        **/
/****************************************************************************/

static void
setup(void)
{
    win_id = winopen(WINDOW_TITLE);

    /* Select the gl modes necessary for this program. */

    mmode(MVIEWING);
    RGBmode();
    doublebuffer();
    gconfig();
    shademodel(GOURAUD);
    zbuffer(TRUE);
    backface(TRUE);
    initnames();
    picksize(3, 3);
    linewidth(9);

    /* Queue the required input events. */

    qdevice(QKEY); qdevice(GKEY);
    qdevice(LEFTARROWKEY); qdevice(RIGHTARROWKEY); qdevice(DOWNARROWKEY);
    qdevice(UPARROWKEY); qdevice(REDRAW);
    qdevice(LEFTMOUSE); qdevice(RIGHTMOUSE);
    qdevice(ONEKEY); qdevice(TWOKEY); qdevice(THREEKEY); qdevice(FOURKEY);
    qdevice(FIVEKEY); qdevice(SIXKEY); qdevice(SEVENKEY);

    /* Set up all of the required data structures. */

    init_world();
    init_menus();
    announce(0, "");
    announce(1, "");

    /* Instruct the connect 4 algorithm to poll for input at level 4 of */
    /* the game tree, allowing the user to maintain control of the      */
    /* program even while the computer is "thinking" about a move.      */

    poll(poll_event, 4);

    /* Set the proper perspective and lookat matrices. */

    my_perspective();
    SetTrackBallRedrawer(draw_world);
    ResetTrackBallMatrices();
    red_lookat();
}


/****************************************************************************/
/**                                                                        **/
/**  This function scans the command line and sets up any states requested **/
/**  within.  If the arguments on the command line are syntactically       **/
/**  invalid, a "usage" message is printed and the program exits with an   **/
/**  error.                                                                **/
/**                                                                        **/
/****************************************************************************/

static void
get_options(int argc, char *argv[])
{
    int argnum = 0;
    char *usage = "Usage: %s <options>\n\nwhere options are:\n"
                  "\t-h\tto get this help\n"
                  "\t-n <n>\tto specify number to connect\n"
                  "\t-p <n>\tto specify number of players (0-2)\n"
                  "\t-f <s>\tto specify first player (red or blue)\n"
                  "\t-r <n>\tto specify skill level of the red player\n"
                  "\t-b <n>\tto specify skill level of the blue player\n"
                  "\t-s <s>\tto specify animation speed of the arms\n"
				  "\t\t(slow, medium or fast)\n";

    /* Look at each argument in turn. */

    while (++argnum < argc) {

        /* If help is requested, print a "usage" message and exit. */

        if (strcmp(argv[argnum], "-h") == 0) {
            printf(usage, argv[0]);
            exit(EXIT_SUCCESS);
        }

        /* If the number of game chips necessary to connect in a row is */
        /* specified, accept the information.                           */

        else if (strcmp(argv[argnum], "-n") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            num_to_connect = atoi(argv[argnum]);
        }

        /* If the number of players is specified, accept the information. */

        else if (strcmp(argv[argnum], "-p") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            num_of_players = atoi(argv[argnum]);
            if (num_of_players > 2) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
        }

        /* If the first player is specified, accept the information. */

        else if (strcmp(argv[argnum], "-f") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            if (argv[argnum][0] == 'r' || argv[argnum][0] == 'R')
                first_player = 0;
            else if (argv[argnum][0] == 'b' || argv[argnum][0] == 'B')
                first_player = 1;
            else {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
        }

        /* If the skll level of the red player is specified, accept the */
        /* information.                                                 */

        else if (strcmp(argv[argnum], "-r") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            skill_levels[0] = atoi(argv[argnum]);
        }

        /* If the skll level of the blue player is specified, accept the */
        /* information.                                                  */

        else if (strcmp(argv[argnum], "-b") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            skill_levels[1] = atoi(argv[argnum]);
        }

        /* If the animation speed of the arms is specified, accept the */
        /* information.                                                */

        else if (strcmp(argv[argnum], "-s") == 0) {
            if (++argnum >= argc) {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
            if (argv[argnum][0] == 's' || argv[argnum][0] == 'S')
                slow_arm_speed();
            else if (argv[argnum][0] == 'm' || argv[argnum][0] == 'M')
                medium_arm_speed();
            else if (argv[argnum][0] == 'f' || argv[argnum][0] == 'F')
                fast_arm_speed();
            else {
                fprintf(stderr, usage, argv[0]);
                exit(EXIT_FAILURE);
            }
        }

        /* If the option was not recognized, print a "usage" message */
        /* and exit with an error.                                   */

        else {
            fprintf(stderr, "Unrecognized option: %s\n", argv[argnum]);
            fprintf(stderr, usage, argv[0]);
            exit(EXIT_FAILURE);
        }
    }
}


/****************************************************************************/
/**                                                                        **/
/**  This function returns the row where the topmost game chip lies in     **/
/**  the specified column of the game board.  If there are no chips in     **/
/**  the specified column, a value of -1 is returned.                      **/
/**                                                                        **/
/****************************************************************************/

static int
topmost_in_column(int column)
{
    int i;
    Game_state state;

    state = get_game_state();
    for (i=5; i>=0 && state.board[column][i] == EMPTY; i--)
        ;

    return i;
}


/****************************************************************************/
/**                                                                        **/
/**  This function performs all of the necessary animation sequences to    **/
/**  make the specified robot arm pick up a game chip and drop it into the **/
/**  specified column and row of the game board.  It will abort and        **/
/**  return FALSE if the user selected END_GAME at any time during the     **/
/**  animation sequence.  Otherwise it will return TRUE.                   **/
/**                                                                        **/
/****************************************************************************/

static Boolean
drop_chip(int player, int column, int row)
{
    float *chip_position, dest_position[3];

    /* Determine the coordinates of the current chip. */
    chip_position = current_chip_pos(player);

    /* Determine the destination coordinates of the chip. */
    coords_of_board_pos(column, row, dest_position);

    /* Pick up the chip. */

    if (!move_to_chip_pile(player, chip_position)) return FALSE;
    if (!pick_up_chip(player)) return FALSE;

    /* Use the robot's version of the chip to move it. */

    remove_current_chip(player);
    move_chip(player);

    /* Drop the chip in the specified column and row. */

    if (!move_to_column(player, column)) return FALSE;
    if (!drop_to_row(player, row)) return FALSE;

    /* Use the real version of the chip again. */

    dont_move_chip(player);
    new_chip_pos(player, dest_position);
    add_current_chip(player);

    /* Return the robot arm to its normal position. */

    if (!return_to_normal(player)) return FALSE;

    next_chip(player);
    return TRUE;
}


/****************************************************************************/
/**                                                                        **/
/**  This function ends the current game and informs the user that the     **/
/**  specified player has won.  A thick white line is drawn to indicate    **/
/**  the winning set of game chips.                                        **/
/**                                                                        **/
/**  This function should only be called if a win by the specified player  **/
/**  has been confirmed.                                                   **/
/**                                                                        **/
/****************************************************************************/

static void
game_won(int player)
{
    int i, x1, y1, x2, y2;

    win_coords(player, &x1, &y1, &x2, &y2);
    end_game();
    game_in_progress = FALSE;
    use_main_menu();

    if (player == 0)
        announce(0, "The Red player has won!");
    else
        announce(0, "The Blue player has won!");
    draw_win_line(x1, y1, x2, y2);
    draw_world();
    for (i=0; i<5; i++) {
        ringbell();
        sginap(20);
    }
}


/****************************************************************************/
/**                                                                        **/
/**  This function ends the current game and informs the user that there   **/
/**  has been a tie.                                                       **/
/**                                                                        **/
/****************************************************************************/

static void
game_tied(void)
{
    int i;

    end_game();
    game_in_progress = FALSE;
    use_main_menu();

    announce(0, "There has been a tie!");
    draw_world();
    for (i=0; i<5; i++) {
        ringbell();
        sginap(20);
    }
}


/****************************************************************************/
/**                                                                        **/
/**  This function begins a new game of the appropriate type.              **/
/**                                                                        **/
/****************************************************************************/

static void
start_game(void)
{
    game_in_progress = TRUE;
    first_game = FALSE;
    use_game_menu();
    hide_win_line();
    new_game(7, 6, num_to_connect);
    current_player = first_player;

    make_announcement();
}


/****************************************************************************/
/**                                                                        **/
/**  This function makes an announcement to the user informing him/her of  **/
/**  the current state of the game.                                        **/
/**                                                                        **/
/****************************************************************************/

void
make_announcement(void)
{
    char msg[60];

    if (in_tb_mode)
        announce(0, "Trackball mode - Use the mouse to pan, zoom and rotate.");
    else if (!game_in_progress) {
        sprintf(msg, "Welcome to the game of Connect %d!  (%s player mode)",
                num_to_connect,
                (num_of_players == 0)? "zero" : ((num_of_players == 1)?
                "one" : "two"));
        announce(0, msg);
    }
    else if (current_player == 0) {
        if (num_of_players == 0) {
            sprintf(msg, "Red player's turn (computer level %d)",
                                                 skill_levels[0]);
            announce(0, msg);
        }
        else if (num_of_players == 1)
            announce(0, "Red player's turn (that's you!)");
        else
            announce(0, "Red player's turn");
    }
    else {
        if (num_of_players < 2) {
            sprintf(msg, "Blue player's turn (computer level %d)",
                                                skill_levels[1]);
            announce(0, msg);
        }
        else
            announce(0, "Blue player's turn");
    }

    if (in_tb_mode)
        announce(1, "Press ESC when done.");
    else if (thinking)
        announce(1, "Thinking...");
    else
        announce(1, "");

    draw_world();
}
