/****************************************************************************/
/**                                                                        **/
/**  This file contains all of the definitions for the materials used in   **/
/**  this project.                                                         **/
/**                                                                        **/
/****************************************************************************/

#include <gl/gl.h>
#include "materials.h"

static float red_chip_material[] = {
    DIFFUSE, 0.8, 0.0, 0.0,
    AMBIENT, 0.4, 0.1, 0.1,
    LMNULL
};

static float blue_chip_material[] = {
    DIFFUSE, 0.0, 0.0, 0.8,
    AMBIENT, 0.0, 0.1, 0.4,
    LMNULL
};

static float red_arm_material[] = {
    DIFFUSE, 0.7, 0.1, 0.2,
    AMBIENT, 0.4, 0.1, 0.1,
    LMNULL
};

static float blue_arm_material[] = {
    DIFFUSE, 0.2, 0.1, 0.7,
    AMBIENT, 0.1, 0.1, 0.4,
    LMNULL
};

static float metal_material[] = {
    DIFFUSE, 0.8, 0.9, 0.2,
    AMBIENT, 0.4, 0.4, 0.1,
    LMNULL
};

static float floor_material[] = {
    DIFFUSE, 0.0, 0.5, 0.1,
    AMBIENT, 0.0, 0.9, 0.0,
    LMNULL
};

static float board_material[] = {
    DIFFUSE, 0.8, 0.7, 0.2,
    AMBIENT, 0.4, 0.4, 0.1,
    LMNULL
};

static float board_border_material[] = {
    DIFFUSE, 0.0, 0.4, 0.9,
    AMBIENT, 0.1, 0.2, 0.4,
    LMNULL
};


/****************************************************************************/
/**                                                                        **/
/**  This function initializes the materials to be used for all of the     **/
/**  objects used in this project.                                         **/
/**                                                                        **/
/****************************************************************************/

void
init_materials(void)
{
    lmdef(DEFMATERIAL, RED_CHIP_MATERIAL, 0, red_chip_material);
    lmdef(DEFMATERIAL, BLUE_CHIP_MATERIAL, 0, blue_chip_material);
    lmdef(DEFMATERIAL, RED_ARM_MATERIAL, 0, red_arm_material);
    lmdef(DEFMATERIAL, BLUE_ARM_MATERIAL, 0, blue_arm_material);
    lmdef(DEFMATERIAL, METAL_MATERIAL, 0, metal_material);
    lmdef(DEFMATERIAL, FLOOR_MATERIAL, 0, floor_material);
    lmdef(DEFMATERIAL, BOARD_MATERIAL, 0, board_material);
    lmdef(DEFMATERIAL, BOARD_BORDER_MATERIAL, 0, board_border_material);
}
