/****************************************************************************/
/**                                                                        **/
/**  This file contains the definitions of all the menus used in this      **/
/**  project, as well as some functions for manipulating them.             **/
/**                                                                        **/
/****************************************************************************/

#include <gl/gl.h>

long view_menu, connect_menu, red_level_menu, blue_level_menu, level_menu,
     first_menu, nplayers_menu, game_menu, speed_menu, main_menu, current_menu;

/****************************************************************************/
/**                                                                        **/
/**  This function initializes all of the menus to be used in this         **/
/**  project.                                                              **/
/**                                                                        **/
/****************************************************************************/

void
init_menus(void)
{
    view_menu = newpup();
    addtopup(view_menu, "red%x0", 0);
    addtopup(view_menu, "blue%x1", 0);
    addtopup(view_menu, "top%x2", 0);
    addtopup(view_menu, "bottom%x3", 0);

    connect_menu = newpup();
    addtopup(connect_menu, "3%x4", 0);
    addtopup(connect_menu, "4%x5", 0);
    addtopup(connect_menu, "5%x6", 0);

    red_level_menu = newpup();
    addtopup(red_level_menu, "4%x7", 0);
    addtopup(red_level_menu, "5%x8", 0);
    addtopup(red_level_menu, "6%x9", 0);
    addtopup(red_level_menu, "7%x10", 0);
    addtopup(red_level_menu, "8%x11", 0);

    blue_level_menu = newpup();
    addtopup(blue_level_menu, "4%x12", 0);
    addtopup(blue_level_menu, "5%x13", 0);
    addtopup(blue_level_menu, "6%x14", 0);
    addtopup(blue_level_menu, "7%x15", 0);
    addtopup(blue_level_menu, "8%x16", 0);

    level_menu = newpup();
    addtopup(level_menu, "red%m", red_level_menu);
    addtopup(level_menu, "blue%m", blue_level_menu);

    first_menu = newpup();
    addtopup(first_menu, "red%x24", 0);
    addtopup(first_menu, "blue%x25", 0);

    nplayers_menu = newpup();
    addtopup(nplayers_menu, "0%x17", 0);
    addtopup(nplayers_menu, "1%x18", 0);
    addtopup(nplayers_menu, "2%x19", 0);

    speed_menu = newpup();
    addtopup(speed_menu, "slow%x20", 0);
    addtopup(speed_menu, "medium%x21", 0);
    addtopup(speed_menu, "fast%x22", 0);

    game_menu = newpup();
    addtopup(game_menu, "Game Menu%t", 0);
    addtopup(game_menu, "skill level%m", level_menu);
    addtopup(game_menu, "arm speed%m", speed_menu);
    addtopup(game_menu, "view%m", view_menu);
    addtopup(game_menu, "trackball mode%x23", 0);
    addtopup(game_menu, "end game%x27", 0);
    addtopup(game_menu, "quit%x28", 0);

    main_menu = newpup();
    addtopup(main_menu, "Main Menu%t", 0);
    addtopup(main_menu, "number to connect%m", connect_menu);
    addtopup(main_menu, "number of players%m", nplayers_menu);
    addtopup(main_menu, "first player%m", first_menu);
    addtopup(main_menu, "skill level%m", level_menu);
    addtopup(main_menu, "arm speed%m", speed_menu);
    addtopup(main_menu, "view%m", view_menu);
    addtopup(main_menu, "trackball mode%x23", 0);
    addtopup(main_menu, "start game%x26", 0);
    addtopup(main_menu, "quit%x28", 0);

    current_menu = main_menu;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the menu system that the main menu is now     **/
/**  the active menu.                                                      **/
/**                                                                        **/
/****************************************************************************/

void
use_main_menu(void)
{
    current_menu = main_menu;
}


/****************************************************************************/
/**                                                                        **/
/**  This function instructs the menu system that the game menu is now     **/
/**  the active menu.                                                      **/
/**                                                                        **/
/****************************************************************************/

void
use_game_menu(void)
{
    current_menu = game_menu;
}


/****************************************************************************/
/**                                                                        **/
/**  This function pops up the active menu and lets the user make a menu   **/
/**  selection.  The item selected is returned.                            **/
/**                                                                        **/
/****************************************************************************/

long
menu_choice(void)
{
    return dopup(current_menu);
}


/****************************************************************************/
/**                                                                        **/
/**  This function disables the "end game" option of the game menu.  This  **/
/**  is required while the computer is thinking about a move.              **/
/**                                                                        **/
/****************************************************************************/

void
disable_end_game(void)
{
    setpup(game_menu, 4, PUP_GREY);
}


/****************************************************************************/
/**                                                                        **/
/**  This function enables the "end game" option of the game menu.         **/
/**                                                                        **/
/****************************************************************************/

void
enable_end_game(void)
{
    setpup(game_menu, 4, PUP_NONE);
}
