#ifndef TB_HANDLER_H
#define TB_HANDLER_H

extern void TrackBallMouseHandler();
	/*
	 * This function should be called after a mousedown when in track
	 * ball mode. It handles all the transformations and redraws till
	 * all the mouse buttons have been released. (Make sure you call
	 * SetTrackBallRedrawer sometime before TrackBallMouseHandler.)
	 */

extern void SetTrackBallRedrawer(void (*redrawFunc)());
	/*
	 * Give this module a function to call in order to update the screen
	 * while applying trackball transformations
	 *
	 * If you've never used a function pointer like this before, don't
	 * worry it's easy.
	 * Just define a function in, say main.c, as follows
	 *
	 * void redraw()
	 * {
	 *		...
	 *		ApplyTrackBallTransformations();
	 *		...
	 * }
	 *
	 * Then somewhere else as part of your initialization do:
	 *
	 * SetTrackBallRedrawer(redraw);
	 */

extern void ResetTrackBallMatrices();
	/* Sets the view transformation to the identity matrix */

extern void SetTrackBallLookat(float vx, float vy, float vz, float px,
			       float py, float pz, float twist);
	/*
	 * Sets the view in the same manner as lookat. Use this instead of
	 * lookat if you want view transformations to be controlled by the 
	 * trackball 
	 *
	 * Note: This function overwrites any previous trackball positioning
	 * 	 So it should really be called only at very specific times.
	 */

extern void ApplyTrackBallTransformations();
	/*
	 * Applies the trackball transformations to the current gl matrix
	 * using multmatrix calls. (ie effects are appended to current
	 * transformations)
	 * This should be called somewhere in your redraw function
	 */

#endif
