#include "pent editor lists.h"
#include "pent editor files.h"
#include "pent globals.h"
#include "list layer.h"
#include "window layer.h"
#include "text layer.h"
#include "resource layer.h"
#include "memory layer.h"
#include <Icons.h>

enum { kGroupList=0, kBoardList };

#define kGroupListLineHeight	19
#define kBoardListLineHeight	kGroupListLineHeight

#define kLockedIconID			650
#define kGroupLDEF				129
#define kBoardLDEF				128
#define kGrowBoxSize			15

static	short			gWhichListActive;
static	ListHandle		gGroupListHandle=0L;
static	ListHandle		gBoardListHandle=0L;
static	Rect			gGroupListRect;
static	Rect			gBoardListRect;
static	short			gMaxGroupsToDisplay=5;
static	Handle			gLockedIconHandle=0L;

void InitTheEditorLists(void)
/* don't call this more than once */
{
	GetIconSuite(&gLockedIconHandle, kLockedIconID, svAllSmallData);
	gGroupListHandle=gBoardListHandle=0L;
	gWhichListActive=kGroupList;
}

ListHandle GetGroupListHandle(void)
{
	return gGroupListHandle;
}

ListHandle GetBoardListHandle(void)
{
	return gBoardListHandle;
}

void GetGroupListRect(Rect *theRect)
{
	*theRect=gGroupListRect;
}

void GetBoardListRect(Rect *theRect)
{
	*theRect=gBoardListRect;
}

Boolean GroupListActiveQQ(void)
{
	return (gWhichListActive==kGroupList);
}

Boolean BoardListActiveQQ(void)
{
	return (gWhichListActive==kBoardList);
}

Boolean HighlightedBoardLockedQQ(void)
{
	Cell			theCell;
	Str255			itemStr;
	Handle			iconHandle;
	
	if (MyGetFirstSelectedCell(gBoardListHandle, &theCell))
	{
		MyGetCellData(gBoardListHandle, theCell, itemStr);
		Mymemcpy((Ptr)&iconHandle, (Ptr)&itemStr[1], 4);
		return (iconHandle!=0L);
	}
	
	return FALSE;
}

Boolean SetHighlightedBoardLocked(Boolean isLocked)
{
	Cell			theCell;
	Str255			itemStr;
	
	if (MyGetFirstSelectedCell(gBoardListHandle, &theCell))
	{
		MyGetCellData(gBoardListHandle, theCell, itemStr);
		if (isLocked)
		{
			Mymemcpy((Ptr)&itemStr[1], (Ptr)&gLockedIconHandle, 4);
		}
		else
		{
			Mymemset((Ptr)&itemStr[1], 0, 4);
		}
		MySetCellData(gBoardListHandle, theCell, itemStr);
	}
	
	return FALSE;
}

void SetGroupListActive(Boolean redraw)
{
	if (gWhichListActive!=kGroupList)
	{
		gWhichListActive=kGroupList;
		if (redraw)
		{
			UpdateEditorLists((**gGroupListHandle).port);
		}
	}
}

void SetBoardListActive(Boolean redraw)
{
	if (gWhichListActive!=kBoardList)
	{
		gWhichListActive=kBoardList;
		if (redraw)
		{
			UpdateEditorLists((**gBoardListHandle).port);
		}
	}
}

void SetHighlightedGroup(short refNum)
{
	Cell			theCell;
	
	if (refNum<0)
		MyDeselectAllCells(gGroupListHandle);
	else
	{
		SetPt(&theCell, 0, refNum);
		MySelectOneCell(gGroupListHandle, theCell);
	}
}

void SetHighlightedBoard(short boardNum)
{
	Cell			theCell;
	
	if (boardNum<0)
		MyDeselectAllCells(gBoardListHandle);
	else
	{
		SetPt(&theCell, 0, boardNum);
		MySelectOneCell(gBoardListHandle, theCell);
	}
}

short GetNumberOfGroups(void)
{
	return (**gGroupListHandle).dataBounds.bottom;
}

short GetNumberOfBoards(void)
{
	return (**gBoardListHandle).dataBounds.bottom;
}

short GetNumberOfVisibleGroups(void)
{
	return (**gGroupListHandle).visible.bottom-(**gGroupListHandle).visible.top;
}

short GetNumberOfVisibleBoards(void)
{
	return (**gBoardListHandle).visible.bottom-(**gBoardListHandle).visible.top;
}

short GetHighlightedGroup(void)
{
	Cell			theCell;
	
	if (MyGetFirstSelectedCell(gGroupListHandle, &theCell))
		return theCell.v;
	else
		return -1;
}

short GetHighlightedBoard(void)
{
	Cell			theCell;
	
	if (MyGetFirstSelectedCell(gBoardListHandle, &theCell))
		return theCell.v;
	else
		return -1;
}

void GetIndGroupName(Str255 theStr, short index)
{
	Cell			theCell;
	
	if ((index<0) || (index>=GetNumberOfGroups()))
	{
		theStr[0]=0x00;
		return;
	}
	
	SetPt(&theCell, 0, index);
	MyGetCellData(gGroupListHandle, theCell, theStr);
}

void GetIndBoardName(Str255 theStr, short index)
{
	Cell			theCell;
	Str255			itemStr;
	
	if ((index<0) || (index>=GetNumberOfBoards()))
	{
		theStr[0]=0x00;
		return;
	}
	
	SetPt(&theCell, 0, index);
	MyGetCellData(gBoardListHandle, theCell, itemStr);
	Mymemcpy((Ptr)&theStr[1], (Ptr)&itemStr[5], itemStr[5]-4);
	theStr[0]=itemStr[0]-4;
}

void GetHighlightedGroupName(Str255 theStr)
{
	Cell			theCell;
	
	if (MyGetFirstSelectedCell(gGroupListHandle, &theCell))
	{
		MyGetCellData(gGroupListHandle, theCell, theStr);
	}
	else theStr[0]=0x00;
}

void GetHighlightedBoardName(Str255 theStr)
{
	Cell			theCell;
	Str255			itemStr;
	
	if (MyGetFirstSelectedCell(gBoardListHandle, &theCell))
	{
		MyGetCellData(gBoardListHandle, theCell, itemStr);
		Mymemcpy((Ptr)&theStr[1], (Ptr)&itemStr[5], itemStr[5]-4);
		theStr[0]=itemStr[0]-4;
	}
	else theStr[0]=0x00;
}

Boolean DeleteHighlightedGroup(void)
{
	Cell			theCell;
	
	if (MyGetFirstSelectedCell(gGroupListHandle, &theCell))
	{
		MyDeleteItemFromList(gGroupListHandle, theCell.v);
		BuildBoardListHandle(GetBoardsFileFSPtr(), TRUE);
		return TRUE;
	}
	
	return FALSE;
}

Boolean DeleteHighlightedBoard(void)
{
	Cell			theCell;
	
	if (MyGetFirstSelectedCell(gBoardListHandle, &theCell))
	{
		MyDeleteItemFromList(gBoardListHandle, theCell.v);
		return TRUE;
	}
	
	return FALSE;
}

void AddGroupToList(Str255 theStr, Boolean autoSelect)
{
	Cell			theCell;
	
	if ((!GroupListActiveQQ()) && autoSelect)
		SetGroupListActive(TRUE);
	
	MyAddStr255ToList(gGroupListHandle, theStr);
	if (autoSelect)
	{
		SetPt(&theCell, 0, GetNumberOfGroups()-1);
		MySelectOneCell(gGroupListHandle, theCell);
		BuildBoardListHandle(GetBoardsFileFSPtr(), TRUE);
	}
}

void AddBoardToList(Str255 theStr, Boolean isLocked, Boolean autoSelect)
{
	Str255			itemStr;
	Cell			theCell;
	
	itemStr[0]=0x04;
	if (isLocked)
	{
		Mymemcpy((Ptr)&itemStr[1], (Ptr)&gLockedIconHandle, 4);
	}
	else
	{
		Mymemset((Ptr)&itemStr[1], 0, 4);
	}
	Mymemcpy((Ptr)&itemStr[5], (Ptr)&theStr[1], theStr[0]);
	itemStr[0]+=theStr[0];
	if ((!BoardListActiveQQ()) && autoSelect)
		SetBoardListActive(TRUE);
	
	MyAddStr255ToList(gBoardListHandle, itemStr);
	if (autoSelect)
	{
		SetPt(&theCell, 0, GetNumberOfBoards()-1);
		MySelectOneCell(gBoardListHandle, theCell);
	}
}

void EditGroupInList(short index, Str255 newStr)
/* index is 0-based */
{
	Cell			theCell;
	
	SetPt(&theCell, 0, index);
	MySetCellData(gGroupListHandle, theCell, newStr);
}

void EditBoardInList(short index, Str255 newStr, Boolean isLocked)
/* index is 0-based */
{
	Str255			itemStr;
	Cell			theCell;
	
	itemStr[0]=0x04;
	if (isLocked)
	{
		Mymemcpy((Ptr)&itemStr[1], (Ptr)&gLockedIconHandle, 4);
	}
	else
	{
		Mymemset((Ptr)&itemStr[1], 0, 4);
	}
	Mymemcpy((Ptr)&itemStr[5], (Ptr)&newStr[1], newStr[0]);
	itemStr[0]+=newStr[0];
	SetPt(&theCell, 0, index);
	MySetCellData(gBoardListHandle, theCell, itemStr);
}

void SetEditorListRects(WindowRef theWindow, short hSpace, short vSpace, short headerHeight)
{
	short			groupBottom;
	
	SetPort(theWindow);
	groupBottom=kGroupListLineHeight*gMaxGroupsToDisplay+headerHeight+vSpace;
	SetRect(&gGroupListRect, hSpace, headerHeight+vSpace, (GetWindowWidth(theWindow)-hSpace)/2,
		groupBottom);
	SetRect(&gBoardListRect, (GetWindowWidth(theWindow)+hSpace)/2, headerHeight+vSpace,
		GetWindowWidth(theWindow)-hSpace, groupBottom);
}

void CreateEditorLists(WindowRef theWindow)
{
	gGroupListHandle=MyCreateVerticalScrollingList(theWindow, gGroupListRect, 1,
		kGroupLDEF, kGroupListLineHeight);
	MySetListSelectionFlags(gGroupListHandle, lOnlyOne);
	gBoardListHandle=MyCreateVerticalScrollingList(theWindow, gBoardListRect, 1,
		kBoardLDEF, kBoardListLineHeight);
	MySetListSelectionFlags(gBoardListHandle, lOnlyOne);
}

OSErr BuildGroupListHandle(FSSpec *theFS, Boolean redraw)
{
	OSErr			oe;
	short			oldRefNum, refNum, numStringResources, resID, i;
	Boolean			alreadyOpen;
	Str255			resName;
	ResType			resType;
	Handle			resHandle;
	
	if ((oe=OpenTheResFile(theFS, &oldRefNum, &refNum, &alreadyOpen, TRUE))!=noErr)
		return oe;
	
	LSetDrawingMode(FALSE, gGroupListHandle);
	MyClearAllCells(gGroupListHandle);
	
	numStringResources=Count1Resources(kBoardResType);
	for (i=1; i<=numStringResources; i++)
	{
		resHandle=Get1IndResource(kBoardResType, i);
		if (resHandle==0L)
		{
			LSetDrawingMode(TRUE, gGroupListHandle);
			return ResError();
		}
		
		GetResInfo(resHandle, &resID, &resType, resName);
		oe=ResError();
		if (oe!=noErr)
		{
			ReleaseResource(resHandle);
			CloseTheResFile(oldRefNum, refNum, alreadyOpen);
			LSetDrawingMode(TRUE, gGroupListHandle);
			return oe;
		}
		AddGroupToList(resName, FALSE);
	}
	
	CloseTheResFile(oldRefNum, refNum, alreadyOpen);
	
	LSetDrawingMode(TRUE, gGroupListHandle);
	
	if (redraw)
		MyUpdateList(gGroupListHandle);
	
	return noErr;
}

OSErr BuildBoardListHandle(FSSpec *theFS, Boolean redraw)
{
	OSErr			oe;
	short			oldRefNum, refNum, numBoards, j;
	Boolean			alreadyOpen;
	Handle			resHandle;
	Str255			groupName;
	Rect			eraseRect;
	MyBoardRec		boardRec;
	
	LSetDrawingMode(FALSE, gBoardListHandle);
	MyClearAllCells(gBoardListHandle);
	
	if (GetHighlightedGroup()>=0)
	{
		GetHighlightedGroupName(groupName);
		
		if ((oe=OpenTheResFile(theFS, &oldRefNum, &refNum, &alreadyOpen, TRUE))!=noErr)
			return oe;
		
		resHandle=Get1NamedResource(kBoardResType, groupName);
		if (resHandle==0L)
		{
			oe=ResError();
			CloseTheResFile(oldRefNum, refNum, alreadyOpen);
			return oe;
		}
		
		numBoards=GetHandleSize(resHandle)/sizeof(MyBoardRec);
		for (j=0; j<numBoards; j++)
		{
			Mymemcpy((Ptr)&boardRec, (Ptr)((unsigned long)*resHandle+sizeof(MyBoardRec)*j),
				sizeof(MyBoardRec));
			AddBoardToList(boardRec.title, boardRec.locked, FALSE);
		}
		
		ReleaseResource(resHandle);
		CloseTheResFile(oldRefNum, refNum, alreadyOpen);
	}
	
	LSetDrawingMode(TRUE, gBoardListHandle);
	
	if (redraw)
	{
		if (GetHighlightedBoard()<0)
			gCurrentBoardRec.numRows=0;
		eraseRect=gBoardListRect;
		eraseRect.right-=16;
		EraseRect(&eraseRect);
		MyUpdateList(gBoardListHandle);
	}
	
	return noErr;
}

void UpdateEditorLists(WindowRef theWindow)
{
	LUpdate(theWindow->visRgn, gGroupListHandle);
	MyDrawListBorder(gGroupListHandle);
	MyDrawActiveListBorder(gGroupListHandle, GroupListActiveQQ());
	LUpdate(theWindow->visRgn, gBoardListHandle);
	MyDrawListBorder(gBoardListHandle);
	MyDrawActiveListBorder(gBoardListHandle, BoardListActiveQQ());
}

void ResizeEditorLists(WindowRef theWindow, short hSpace, short vSpace, short headerHeight)
{
	LSetDrawingMode(FALSE, gGroupListHandle);
	LSetDrawingMode(FALSE, gBoardListHandle);
	SetEditorListRects(theWindow, hSpace, vSpace, headerHeight);
	MySetListRect(gGroupListHandle, gGroupListRect);
	MySetListRect(gBoardListHandle, gBoardListRect);
	MySetListSize(gGroupListHandle, gGroupListRect.right-gGroupListRect.left-16,
		gGroupListRect.bottom-gGroupListRect.top, kGroupListLineHeight);
	MySetListSize(gBoardListHandle, gBoardListRect.right-gBoardListRect.left-16,
		gBoardListRect.bottom-gBoardListRect.top, kBoardListLineHeight);
	MyDrawListBorder(gGroupListHandle);
	MyDrawListBorder(gBoardListHandle);
	if (GroupListActiveQQ())
		MyDrawActiveListBorder(gGroupListHandle, TRUE);
	if (BoardListActiveQQ())
		MyDrawActiveListBorder(gBoardListHandle, TRUE);
	LSetDrawingMode(TRUE, gGroupListHandle);
	LSetDrawingMode(TRUE, gBoardListHandle);
	InvalRect(&(theWindow->portRect));
}

void GetGrowSizeTheLists(Rect *sizeRect, short hSpace, short vSpace, short headerHeight)
{
	#pragma unused(hSpace)
	
	SetRect(sizeRect, 294,
		48+kGrowBoxSize+1+headerHeight+kGroupListLineHeight*gMaxGroupsToDisplay+vSpace*5,
		32766, 32767);
}

void DisposeEditorLists(void)
{
	if (gGroupListHandle!=0L)
		LDispose(gGroupListHandle);
	if (gBoardListHandle!=0L)
		LDispose(gBoardListHandle);
	gGroupListHandle=gBoardListHandle=0L;
}
