/*************************** MTPrefsFile.c ************************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	10 April 93 

 This file overrides a standard preferences file to customize routines for
 ModemTalk. 
 
*/

#include 	<Serial.h>
#include 	<String.h>
#include	<GestaltEqu.h>
#include	<Traps.h>
#include 	"TrapAvailable.h"
#include	"ModemTalk.h"
#include	"MTalkUtilities.h"
#include	<BPrefsFile.h>
#include 	"BMTPrefsFile.h"

#define baud14400 6
#define baud38400 1
#define kCR	0x0D

OSErr 	BMTPrefsFile::IPrefsFile(OSType fileType, OSType sign,
			short strsID, ResType resTyp, short resID)
{
	OSErr osstat;
	
	osstat = inherited::IPrefsFile(fileType,sign,strsID,resTyp,resID);
	return(osstat);
}

void	BMTPrefsFile::Dispose(void)
{
	inherited::Dispose();
}

/***********************************
 * G E T S E L E C T E D S P E E D *
 ***********************************
 
 Return the speed selected in the Popup menu when the user hit OK. This value
 may be passed to the SerReset() function.
  
*/

short	BMTPrefsFile::GetSelectedSpeed(void)
{
#define kNumberOfSpeeds 10

	long		speed [kNumberOfSpeeds] = 
				{1200, 2400, 4800, 9600, 
				 14400, 19200, 38400, 57600,
				 -1, -1};
	short		speedCode [kNumberOfSpeeds] = 
				{baud1200, baud2400, baud4800, baud9600,
				 baud14400, baud19200, baud38400, baud57600,
				 -1, -1};

	char		theSpeedStr[20];
	long		theSpeed;
	short		i;
	
	theSpeed = GetBitsPerSecond();
	for (i = 0; i<kNumberOfSpeeds; i++)	{
		if (speed[i] == theSpeed) break;
	}
	
	if (i >= kNumberOfSpeeds) {		// bad news...
//		DDebugStr("\pBad Speed value");	// DEBUG!
		return(baud9600);		// default to 9600 bps
	}
	return(speedCode[i]);
}

/***********************************
 * G E T B I T S P E R S E C O N D *
 ***********************************
 
 Return the speed in BPS.
  
*/

long BMTPrefsFile::GetBitsPerSecond()
{
	long	mySpeed;
	char	theSpeedStr[20];
	
	GetAPref(sBPS,theSpeedStr,sizeof(theSpeedStr));
	StringToNum(theSpeedStr, &mySpeed);
	return(mySpeed);
}
	
// GetPortNames - return the port chosen by user as Pascal strings
//	Returns "\p.AIn" & "\p.AOut" unless the setting matches "Printer"

void	BMTPrefsFile::GetPortNames(char *inPort, char *outPort)
{	
	char	portName[20];
	
	BlockMove("\p.AIn", inPort, 5);		// Assume Phone port
	BlockMove("\p.AOut", outPort, 6);
	
	GetAPref(sPort, portName, sizeof(portName));
	if (EqualString(portName, "\pPrinter", FALSE, FALSE)) {
		inPort[2] = 'B';		// no, they want Printer port
		outPort[2] = 'B';
	}
}

Boolean	BMTPrefsFile::GetDialOnOpen(void)
{	
	char	str[2];
	
	GetAPref(sOpen, str, sizeof(str));
	return(str[0]);
}

Boolean	BMTPrefsFile::GetDialOnStartup(void)
{
	char	str[2];
	
	GetAPref(sStartup, str, sizeof(str));
	return(str[0]);
}

// GetDialScript - returns the script to be used to dial. Script will be read 
//	from Prefs file (if present) or from the adev file itself.

Handle	BMTPrefsFile::GetDialScript(void)
{
	return(GetResource('Dilr',128));
}

// GetDialString - return the (Pascal) dialing string "ATMxDx###,,,### <cr>"
// 	Warning: theNumber must point to a buffer of at least 250 characters

void	BMTPrefsFile::GetDialString(char *theNumber)
{
#define kMaxDialStrLen 249

	char	temp[256];
	long	l;
	
	strcpy(theNumber, (char*)"\pATM1DT");		// init to noisy, tone
	GetAPref(sSpeaker, temp, sizeof(temp));
	if (temp[1] == '1') 
		theNumber[4] = '0';		// make it quiet (M-zero)
	GetAPref(sTone, temp, sizeof(temp));
	if (temp[1] == '1')
		theNumber[6] = 'P';		// force pulse dialing
	GetAPref(sPhNum,temp, sizeof(temp));
	PZstrcat(theNumber, temp, kMaxDialStrLen); // now kMaxDialStrLen, (was 40) 24 Aug 94 reb
	GetAPref(sDoCCard, temp, sizeof(temp));
	if (temp[0])	{			// we should do calling card stuff
		PZstrcat(theNumber, (char*)"\p;", kMaxDialStrLen); // add a ";"
	}
	theNumber[0]++;				// tack on final CR
	theNumber[theNumber[0]] = kCR;
}

// GetCCString - return the (Pascal) dialing string "ATDTccccccc <cr>"
// 	Warning: theNumber must point to a buffer of at least 250 characters

void	BMTPrefsFile::GetCCString(char *theNumber)
{
#define kMaxDialStrLen 249

	char	temp[256];
	long	l;
	
	strcpy(theNumber, (char*)"\pATDT");	// Tone dial this one...
	GetAPref(sDoCCard, temp, sizeof(temp));
	if (temp[0])	{			// we should do calling card stuff
		register short 	j,i;

		GetAPref(sPause, temp, sizeof(temp));
		StringToNum(temp, &l);
		l = ((l+1)>>1);			// round up and divide by 2
		for (i = 0; i<l; i++)	{	// add on commas
			j = theNumber[0];
			j++;
			theNumber[j] = ',';
			theNumber[0] = j;
		}
		GetAPref(sCardNum,temp, sizeof(temp));
		PZstrcat(theNumber, temp, kMaxDialStrLen); // and finally the calling card #
	}
	theNumber[0]++;				// tack on final CR
	theNumber[theNumber[0]] = kCR;
}

// GetModemString - return the selected modem initialization string. Tack a <cr>
//	on the end before returning it.
	
void	BMTPrefsFile::GetModemString(char *theString)
{
	GetAPref(sInitStr, theString, 40);
	theString[0]++;
	theString[theString[0]] = kCR;
}

/***********************************
 * M A C N O T C O M P A T I B L E *
 ***********************************
 
 Test to see if the Mac is compatible. Currently no tests are significant:
 This routine simply returns FALSE, to indicate that all macs *are*
 compatible. (We're somewhat ignoring that Express Modems freeze when
 you pull out the RJ-11 plug...)
 	
 If the Mac is not compatible, this displays warning and "Oh." button, 
 then returns TRUE if we're not able to run...
 
*/
#define _PMgrOp 0xA085

Boolean BMTPrefsFile::MacNotCompatible(void)
{
	Boolean hasPowerMgr, hasMacTCP,onPhonePort;
	Boolean	retval;
	OSErr	osstat;
	short	refnum;
	char	portName[10];
	
#define strHasExpModem "\pWarning: ModemTalk may freeze when using an Express Modem."
#define strHasExpModem1 "\pThis is likely to happen when the telephone connection fails."
#define strHasMacTCP "\pWARNING: ModemTalk is not compatible with MacTCP on PowerBook or Duo computers."
#define strHasMacTCP1 "\pIf you use an application which uses MacTCP, your Mac may crash."
	
	retval = FALSE;				// assume the best
	osstat = gestaltUnknownErr;		// assume no Express Modem 
	if (TrapAvailable(_GestaltDispatch)) {	// noGestalt implies compatible
		hasPowerMgr = TrapAvailable(_PMgrOp); // look for Power Mgr
		osstat = AssertDrvrOpen("\p.ipp",&refnum);
		hasMacTCP = (refnum != 0); 	// if refnum == 0, MacTCP's not there
		GetAPref(sPort, portName, sizeof(portName));
		onPhonePort = EqualString(portName, "\pPhone", FALSE, FALSE);
		
		if (hasPowerMgr && hasExpModem && onPhonePort) { // warn 'em
			//ParamText(strHasExpModem, strHasExpModem1, "\p", "\p"); // just warn 'em
			//CautionAlert(262, NULL);
			//retval = TRUE;
		}
		else if	(hasPowerMgr && hasMacTCP)	{ // warn 'em
			//ParamText(strHasMacTCP, strHasMacTCP1, "\p", "\p"); // just warn 'em
			//CautionAlert(262, NULL);
			//retval = FALSE;
		}
	}

	return(retval);
}