/******************************* BMTalkDLOG.c *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	24 March 1993 

 This object handles the user interface of the main ModemTalk configuration
 dialog. This presents the current settings (from the ModemTalk Prefs file), 
 and allows the user to change them. 
 
 When the user clicks Dial/Hangup, BMTalkDLOG updates the preferences file.
 
*/

#include <GestaltEqu.h>
#include <Folders.h>
#include <string.h>
#include "mySetupA4.h"
#include "atlk.h"
#include "MTalkUtilities.h"
#include "ModemTalk.h"
#include "BMTPrefsFile.h"
#include "BMTalkDLOG.h"
#include "Help.h"

#define DLOGMain 256
#define	DLOGMoreInfo 264
#define HELPSTR 257

enum { kSpeedCNTL = 264, kModemTypeCNTL, kPortCNTL };

enum DLOGItems {bConnect=1, bCancel, bMoreChoices, bPhNum, 
		bSpeedPopup, bStatusBox, bModemPopup, bNumTitle, 
		bOKUItem, bStatUItem, bPortPopup, bPICT, bDLOGvers};
enum DLOGMoreItems {bOK=1, bMoreCancel, bStartup, bOpened, bMoreTitle,
		bTone, bPulse, bSpeakerOn, bSpeakerOff, bDialingUItem, 
		bSpeakerUItem, bCreditCardPause, bCreditCardLegend2, 
			bCreditCardNumber, bCreditCardCheckBox, 
		bCreditCardUItem, bMoreOKUItem, bDialingTitle, bSpeakerTitle };

typedef struct popupPrivateData {
	MenuHandle 	mHandle;	// Menu's handle
	short		mID;		// menu ID
	char		mPrivate[1];	// private stuff
} popupPrivateData, *popupPrivatePtr, **popupPrivateHdl;

typedef struct	VersRsrc {
 		char	major;
 		char	minor;
 		char	release;
 		char	non_release;
 		short	country;
 		char	shortVers[20];	/* Pascal String, really */
 		} VersRsrc;
 		
#define kCR	0x0D
#define PSLen(x) ((x) & 0x00FF)

#pragma mark Local functions

pascal Boolean DLOGFilter(DialogPtr theWindow, 
		EventRecord * theEvent, short * itemHit);
void 	DrawRadioButtons(DialogPtr dPtr, short bFirst, short bLast, short theSel);
void 	ToggleCheckBox(DialogPtr dMorePtr, short theCtl);
void	SetUserItemProc(DialogPtr dPtr, short theItem, ProcPtr theProc);
pascal void DoOutline(DialogPtr theWindow, short itemNo);
pascal void DoFrame(DialogPtr theWindow, short itemNo);
pascal void DoVers(DialogPtr theWindow, short itemNo);
ControlHandle GetNewPopupControl(short id, WindowPtr wp);
void 	FixupPopupControl(short id);

// InitDLOG - Initialize the object using the current settings from the prefs
//	file. 

OSErr	BMTalkDLOG::InitMTalkDLOG(BMTPrefsFile *prefs)
{
	OSErr	osstat;
	short	thisResFile;
	Handle	hdl;
 	ControlHandle 	cHdl;
 	CursHandle	cursH;
 	
	cursH = GetCursor(watchCursor);		// get the watch cursor
	HLock((Handle)cursH);
	wCursor = **cursH;			// make a local copy
	ReleaseResource(cursH);
	
 	FixupPopupControl(kSpeedCNTL);
 	FixupPopupControl(kModemTypeCNTL);
 	FixupPopupControl(kPortCNTL); 		// crock up popups for Sys 6
	
	itsPrefs = prefs;			// use the passed prefs file	
	isShowing = FALSE;			// not on screen yet...
	return (noErr);
}	

// Dispose - deallocate buffers, etc.

void	BMTalkDLOG::Dispose(void)
{
	if (dMainPtr) DisposDialog(dMainPtr);		dMainPtr = NULL;
	if (dMorePtr) DisposDialog(dMorePtr);		dMorePtr = NULL;
	if (CDEFrsrc) DisposeHandle(CDEFrsrc);		CDEFrsrc = NULL;
	delete this;
}

// SetWatchCursor - set the watch cursor

void	BMTalkDLOG::SetWatchCursor(void)
{
	HLock(this);			// lock our object
	SetCursor(&wCursor);		// copy it
	HUnlock(this);
}

// ShowDialog - show the dialog that has already been initialized

void	BMTalkDLOG::ShowMTDLOG(Boolean showDial, char *initMsg)
{
	FlushEvents(everyEvent, 0);
	if (!isShowing)		{
		// get the dialog
		dMainPtr = GetNewDialog(DLOGMain, (DialogPeek) NULL, (WindowPtr) -1) ; /* get the dialog */
		if (!dMainPtr) 
			DDebugStr("\pCouldn't open main DLOG");
		SetupMainDLOG(showDial);		// fill in the DLOG items
		ClearDisplay();
		ShowWindow((WindowPtr) dMainPtr);	// make the DLOG visible
	 	DrawDialog(dMainPtr);			// force it to update 
		isShowing = TRUE;
		AddToDisplay(initMsg);			// put up the desired string
	}
}

// DoMTDLOG - Handle any clicks in the user interface. 
//	Return noErr if they clicked OK; 
//	return -1 if they clicked Cancel; 
//	return OSErr (<0) otherwise.
//	If they clicked Connect, saves the settings in the prefs file.

OSErr	BMTalkDLOG::DoMTDLOG(void)
{
	long		gestaltResponse;
	OSErr		retval;
	OSErr		osstat;
	short		item;
	short		itemtype;
	Handle		itemH;
	Rect		box;
	short		i;	
 	
	KeyMap  	theKeys;

#define CmdKey		48

	retval = -1;			// they haven't said yes or no
	
	if (!isShowing)
		DDebugStr("\pDLOG not shown...");
	InitCursor();				// now make an arrow	
 	for (;;)	{			/* ask 'em */
 		ModalDialog((ModalFilterProcPtr) DLOGFilter, &item);
 		 		
 		switch (item)				/* handle enabled items */
 		{
 		case bConnect: 	/* Dial/Hangup */
// 			histlog("bConnect",0,0);
			SetWatchCursor();
	 		GetSettingsFromMain();	
			itsPrefs->SavePrefs();	// save settings in the prefs file
 			retval = noErr;
 			break;
 				
 		case bCancel:	
 			retval = -1;	/* Cancel */
// 			histlog("bCancel",0,0);
			SetWatchCursor();
 			break;
 				
 		case bMoreChoices:	 
 			osstat = SetupMoreDLOG();
 			if (!osstat)	{
 				osstat = DoMoreDLOG();
 				if (osstat == 1)	{	// OK
 					GetMoreSettings();
 				}
 			} 
 			if (dMorePtr)	{
 				DisposDialog(dMorePtr);
 				dMorePtr = NULL;
 			}
			break; 		/* More Choices button... */
				
		case bPICT:			// anywhere in PICT gives help
			Help(HELPSTR, helvetica, 10); /* Help */
			break;
			
		default: 
//			DDebugStr("\pHow'd I get here?");
			break;

 		}
 		if ((item == bConnect) || (item == bCancel)) {
			GetKeys(&theKeys);	// if Command key, show debug
			if (BitTst((Ptr) &theKeys, CmdKey)) {
			useDebugger = TRUE;
			asm	{				// DEBUG!
				lea	adevVersion,A0
			}
			DDebugStr("\patlkg;dm a0;g"); 	// DEBUG!
			asm	{ 				
				lea mtlkg,A0 			// DEBUG!
				lea numIMtx,A1			// DEBUG!
			}	
			DDebugStr("\pMTalkg & MTS;dm a0;dm a1;g");	// DEBUG!
			}
			
 			break;
 		}
 	}
	return(retval);		// and return status
}			

/***************************
 * C L E A R D I S P L A Y *
 ***************************
 
 Set the string displayed to "\p".  

*/

void	BMTalkDLOG::ClearDisplay(void)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;

	if (!isShowing) return;
	
	GetDItem(dMainPtr, bStatusBox, &itemtype, &item, &box);
	SetIText(item, "\p");				// empty string
}

/***************************
 * A D D T O D I S P L A Y *
 ***************************
 
 Add aStr (Pascal format) to the string currently displayed on the screen. 
  
*/
void	BMTalkDLOG::AddToDisplay(char *aStr)
{
#define kNumLinesToDisplay	4

 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	char		curStr[256];		// current string from DITL
 	char		newStr[256];		// local copy of the new string
 	short		newLen, curLen, copyLen;
 	char		*sPtr;
 	short		i;
 	char		*p, *last;

	if (!isShowing) return;			// not on screen now...
	
	newLen = PSLen(*aStr);
	BlockMove(aStr, newStr, newLen+1);	// get new string in local storage
	if (newStr[newLen] == kCR)
		newLen--;			// ignore trailing CR
		
	GetDItem(dMainPtr, bStatusBox, &itemtype, &item, &box);
	GetIText(item, curStr);			// get the current string
	curLen = PSLen(curStr[0]);
	
	sPtr = curStr+1;			// adrs of first char
	if (newLen + curLen >= sizeof(curStr)-2) { // if it won't fit with len+kCR
		copyLen = sizeof(curStr) - 2 - newLen;
		BlockMove(sPtr+curLen-copyLen, sPtr, sizeof(curStr)-newLen-2);
		curLen = copyLen;		// subtract off newlen chars
	}
	if ((curLen > 0) && (curStr[curLen] != kCR))	{// add a CR between entries
		curLen++;				 //  if required
		curStr[curLen] = kCR;
	}
	BlockMove(newStr+1, sPtr+curLen, newLen); // copy in the new characters
	curStr[0] = curLen + newLen;		// update the length byte
	
	// now scan backwards to keep only the last n lines

	p = curStr+PSLen(*curStr);	// point at last character
	last = p;
	i = 0;
	
	while (p != curStr) 	{	// run through the text
		if (*p == kCR) i++;	// if a CR, bump the counter	
		if (i >= kNumLinesToDisplay) break;// if we found this many, break
		p--;			// move to previous char (start of next line)
	}
	
	// p now points to front of curStr or kCR before first char of new string...
	
	if (p != curStr) 		{	// didn't get to beginning
		newLen = last - p + 1;		// include length byte
		BlockMove(p, curStr, newLen);// copy the lines
		curStr[0] = newLen-1;	// update the count
	}
	
	SetIText(item, curStr);
}

// AddToDebug - append a string to the existing debug trace

void	BMTalkDLOG::AddToDebug(char *aStr)
{

}

// SetupMTDLOG - fill in Dialog items from the current settings
//	showDial is true if the button should be labelled "Dial".

void	BMTalkDLOG::SetupMainDLOG(Boolean showDial)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	OSErr		osstat;
 	short		i;
 	Ptr		p;

	SetPopup(dMainPtr, bSpeedPopup, sBPS);
	SetPopup(dMainPtr, bModemPopup, sModemType);
	SetPopup(dMainPtr, bPortPopup, sPort);
	SetEditText(dMainPtr, bPhNum, sPhNum);
	SetUserItemProc(dMainPtr, bOKUItem, &DoOutline); // outline default button
	SetUserItemProc(dMainPtr, bStatUItem, &DoFrame); // frame the status area
	SetUserItemProc(dMainPtr, bDLOGvers, &DoVers);   // draw the version number
	
/* set the Dial/Hangup button, depending on the link state */

	if (!showDial)	{	// default is "Dial"
		GetDItem(dMainPtr, bConnect, &itemtype, &item, &box);
		SetCTitle(item, "\pHang Up");
	}
}

/*****************************************
 * G E T S E T T I N G S F R O M D L O G *
 *****************************************
 
 Update the settings strings from the DLOG items
 
*/

void	BMTalkDLOG::GetSettingsFromMain(void)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	char		buf[256];
	
	GetPopup(dMainPtr, bSpeedPopup, sBPS);
	GetPopup(dMainPtr, bPortPopup, sPort);
	
	GetPopup(dMainPtr, bModemPopup, sModemType);		// get the modem type
	itsPrefs->GetAPref(sModemType, buf, sizeof(buf));
	item = GetNamedResource('mStr', buf);		// fetch its init string
	if (item == NULL) 
		item = (Handle)GetString(-4032);
	HLock(item);
	itsPrefs->SetAPref(sInitStr,(char*)*item);
	HUnlock(item);
	ReleaseResource(item);
	
	GetEditText(dMainPtr, bPhNum, sPhNum);
}

/*****************************
 * S E T U P M O R E D L O G *
 *****************************
 
 Load and fill in all fields for the More... DLOG. Return OSErr if an error occurred. 
  
*/

OSErr	BMTalkDLOG::SetupMoreDLOG(void)
{
	dMorePtr = GetNewDialog(DLOGMoreInfo,(DialogPeek) NULL, (WindowPtr) -1);
	if (!dMorePtr)	
		return (-108);
		
	SetCheckBox(dMorePtr, bStartup, sStartup);
	SetCheckBox(dMorePtr, bOpened, sOpen);
	SetRadioButtons(dMorePtr, bTone, bPulse, sTone);
	SetRadioButtons(dMorePtr, bSpeakerOn, bSpeakerOff, sSpeaker);
	SetEditText(dMorePtr, bCreditCardNumber, sCardNum);
	SetEditText(dMorePtr, bCreditCardPause, sPause);
	SetCheckBox(dMorePtr, bCreditCardCheckBox, sDoCCard);
	SetUserItemProc(dMorePtr, bMoreOKUItem, &DoOutline);
	SetUserItemProc(dMorePtr, bDialingUItem, &DoFrame);
	SetUserItemProc(dMorePtr, bSpeakerUItem, &DoFrame);
	SetUserItemProc(dMorePtr, bCreditCardUItem, &DoFrame);

	ShowWindow((WindowPtr) dMorePtr);// make the DLOG visible
	DrawDialog(dMorePtr);
	return (noErr);
}

/***********************
 * D O M O R E D L O G *
 ***********************
 
 Do the user interface required for the More... DLOG. Return 1 if they clicked
 OK; 0 if they clicked Cancel; < 0 if there was an error.
  
*/

OSErr	BMTalkDLOG::DoMoreDLOG(void)
{
	short	item;
	OSErr	retval;
	
	while (TRUE)	{
		ModalDialog((ModalFilterProcPtr) DLOGFilter, &item);
		switch (item)	{
		
		case bOK:	
			retval = 1;
			break;
			
		case bMoreCancel:
			retval = 0;
			break;
			
 		case bOpened: 			/* Dial when opened */
 			ToggleCheckBox(dMorePtr, bOpened);
 			break;
	 			
 		case bStartup: 			/* Dial on startup */
 			ToggleCheckBox(dMorePtr, bStartup);
 			break;

		case bCreditCardCheckBox:
			ToggleCheckBox(dMorePtr, bCreditCardCheckBox);
			break;
			
		case bTone:
		case bPulse:
			DrawRadioButtons(dMorePtr, bTone, bPulse, item);
			break;
			
		case bSpeakerOn:
		case bSpeakerOff:
			DrawRadioButtons(dMorePtr, bSpeakerOn, bSpeakerOff, item);
			break;
		
		default: 
//			DDebugStr("\pBad Item in DoMoreDLOG");
			break;

		}
		if ((item == bOK) || (item == bMoreCancel)) break;
	}
	return(retval);
}

/*********************************
 * G E T M O R E S E T T I N G S *
 *********************************
 
 Get settings info from the More... DLOG. Save 'em in the prefs file. 
  
*/

void	BMTalkDLOG::GetMoreSettings(void)
{
	GetCheckBox(dMorePtr, bStartup, sStartup);
	GetCheckBox(dMorePtr, bOpened, sOpen);
	GetRadioButtons(dMorePtr, bTone, bPulse, sTone);
	GetRadioButtons(dMorePtr, bSpeakerOn, bSpeakerOff, sSpeaker);
	GetEditText(dMorePtr, bCreditCardNumber, sCardNum);
	GetEditText(dMorePtr, bCreditCardPause, sPause);
	GetCheckBox(dMorePtr, bCreditCardCheckBox, sDoCCard);
}

/*************************
 * S E T C H E C K B O X *
 *************************
 
 Update the display of a checkbox on the screen from the value 
 in the prefs.
 
*/

void	BMTalkDLOG::SetCheckBox(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	char		theStr[2];

	GetDItem(dPtr, dItem, &itemtype, &item, &box);	
	itsPrefs->GetAPref(theSetting, theStr,sizeof(theStr));
	SetCtlValue((ControlHandle) item, theStr[0]);
}

/*******************
 * S E T P O P U P *
 *******************
 
 Update a popup menu on the screen based on the value in the settings.
 
*/

void	BMTalkDLOG::SetPopup(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	short		i;
	char		theStr[256];
	
	itsPrefs->GetAPref(theSetting, theStr,sizeof(theStr));
	GetDItem(dPtr, dItem, &itemtype, &item, &box);	/* get popup item */
	i = MatchingMenuItem(theStr, (ControlHandle) item);			
	if (i == 0) i = 1;				/* use default */
	SetCtlValue((ControlHandle) item, i);
}

/*************************
 * S E T E D I T T E X T *
 *************************
 
 Update the display of editable text from the value in the prefs.
 
*/

void BMTalkDLOG::SetEditText(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	char		str[256];

	itsPrefs->GetAPref(theSetting, str,sizeof(str));	/* get phone number */
	GetDItem(dPtr, dItem, &itemtype, &item, &box);	
	SetIText(item, (StringPtr) str);
	SelIText(dPtr, dItem, 0, 32767);
}

/*********************************
 * S E T R A D I O B U T T O N S *
 *********************************
 
 Update the group of radio buttons from the value in the prefs. 
 Radio button prefs are zero-based - 0 implies to use the first choice, 
 1 the second, etc.
 
*/

void	BMTalkDLOG::SetRadioButtons(DialogPtr dPtr, short bFirst, short bLast, short theSetting)
{
 	char		str[256];
 	long		l;
	
	itsPrefs->GetAPref(theSetting, str, sizeof(str));
	StringToNum(str, &l);
	DrawRadioButtons(dPtr, bFirst, bLast, l+bFirst);
}

/***********************************
 * D R A W R A D I O B U T T O N S *
 ***********************************
 
 Draw the group of radio buttons, selecting the specified value. 
  
*/

void DrawRadioButtons(DialogPtr dPtr, short bFirst, short bLast, short theSel)
{
 	short		i;
 	short 		itemtype;
 	Rect		box;
 	Handle		item;

	if (bFirst > bLast) 
		DDebugStr("\pRadioButtons confused...");
	if (theSel < bFirst) theSel = bFirst;	// sanity checking...
	else if (theSel > bLast) theSel = bLast;

	for (i = bFirst; i <= bLast; i++)	{
		GetDItem(dPtr, i, &itemtype, &item, &box);
		SetCtlValue((ControlHandle) item, (i == theSel));
	}
}

/*******************************
 * T O G G L E C H E C K B O X *
 *******************************
 
 Update the checkbox shown on the screen with its negation.
 
*/

void ToggleCheckBox(DialogPtr dPtr, short theItem)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		itemH;
	short		i;
	
  	GetDItem(dPtr, theItem, &itemtype, &itemH, &box);	
	i = GetCtlValue((ControlHandle) itemH);
	SetCtlValue((ControlHandle) itemH, 1-i);
}

/*************************
 * G E T C H E C K B O X *
 *************************
 
 Update the pref from the value in the checkbox.
 
*/

void	BMTalkDLOG::GetCheckBox(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	char		theStr[2];

	GetDItem(dPtr, dItem, &itemtype, &item, &box);	
	theStr[1] = 'x';
	theStr[0] = GetCtlValue((ControlHandle) item);
	itsPrefs->SetAPref(theSetting, theStr);
}

/*******************
 * G E T P O P U P *
 *******************
 
 Update the setting for a popup from the dialog item
 
*/

void	BMTalkDLOG::GetPopup(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemType;
 	Rect		box;
 	ControlHandle	item;
 	short		i;
	char		theStr[64];
	
	GetDItem(dPtr, dItem, &itemType, &item, &box);
	i = GetCtlValue(item);
	GetPopupItemText(item, i, theStr);
	itsPrefs->SetAPref(theSetting, theStr);
}

/*********************************
 * G E T R A D I O B U T T O N S *
 *********************************
 
 Update the setting from a group of radio buttons from the dialog
 
*/

void	BMTalkDLOG::GetRadioButtons(DialogPtr dPtr, short bFirst, short bLast, short theSetting)
{
 	short 		itemType;
 	Rect		box;
 	ControlHandle	item;
 	long		l;
 	short		i;
	char		theStr[64];
	
	for (i = bFirst; i <= bLast; i++)	{	// spin 'til we get non-zero
		GetDItem(dPtr, i, &itemType, &item, &box);
		if (GetCtlValue((ControlHandle) item)) break;
	}
	if (i > bLast)
		i = bFirst;
	i -= bFirst;
	l = i;
	NumToString(l, theStr);		
	itsPrefs->SetAPref(theSetting, theStr);
}

/*************************
 * G E T E D I T T E X T *
 *************************
 
 Update the setting for a popup from the dialog item
 
*/

void	BMTalkDLOG::GetEditText(DialogPtr dPtr, short dItem, short theSetting)
{
 	short 		itemType;
 	Rect		box;
 	ControlHandle	item;
 	short		i;
	char		theStr[256];
	
	GetDItem(dPtr, dItem, &itemType, &item, &box);	
	GetIText(item, theStr);
	itsPrefs->SetAPref(theSetting, theStr);
}

/*********************************
 * S E T U S E R I T E M P R O C *
 *********************************
 
 Set the UserItem procedure for this item in this DLOG
 
*/

void	SetUserItemProc(DialogPtr dPtr, short theItem, ProcPtr theProc)
{
 	short 		itemtype;
 	Rect		box;
 	Handle		item;

	GetDItem(dPtr, theItem, &itemtype, &item, &box); /* set up outliner */
	SetDItem(dPtr, theItem, itemtype, (Handle) theProc, &box);/**/
}			

/*****************************************************************************
 DoOutline - Outline default button
 *****************************************************************************/

pascal void DoOutline(DialogPtr theWindow, short itemNo)
 {
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	
	GetDItem(theWindow, itemNo ,&itemtype, &item, &box);
 	InsetRect(&box, -4, -4);
 	PenSize(3, 3);
 	FrameRoundRect(&box, 16, 16); 
 	PenSize(1, 1);
  }

/*****************************************************************************
 DoFrame - draw a frame around a user item
 *****************************************************************************/

pascal void DoFrame(DialogPtr theWindow, short itemNo)
 {
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	
	GetDItem(theWindow, itemNo ,&itemtype, &item, &box);
 	InsetRect(&box, -2, -2);
 	FrameRect(&box);
  }

/*****************************************************************************
 DoVers - Draw the version number in 9 pt Geneva
 *****************************************************************************/

pascal void DoVers(DialogPtr theWindow, short itemNo)
 {
 	short 		itemtype;
 	Rect		box;
 	Handle		item;
 	short		oldSize;
 	short		oldFont;
 	VersRsrc	**versH;
 	char		str[64];
 	
/* get the vers=1 string and fill in the title */
	
	versH = (VersRsrc **) GetResource('vers',1);
	BlockMove((**versH).shortVers, str, PSLen((**versH).shortVers[0])+1);
		
	GetDItem(theWindow, itemNo ,&itemtype, &item, &box);
	MoveTo(box.left + 1, box.bottom - 1);	
	oldFont = (*(GrafPort *) theWindow).txFont;
	oldSize = (*(GrafPort *) theWindow).txSize;
	TextFont(geneva);
	TextSize(9);
	DrawString((StringPtr) str);
	TextFont(oldFont);
	TextSize(oldSize);
 }

/*****************************************************************************
 DLOGFilter - filter dialog input.  Cmd-., etc.
 	Treat Return & Enter as a click on the Default Button
 	Treat Cmd-. or Esc as a click on the Cancel button
 	Treat Cmd-Opt-C as a command to copy the History Queue to the Clipboard
 *****************************************************************************/

pascal Boolean DLOGFilter(DialogPtr theWindow, EventRecord * theEvent, short * itemHit)
 {
 	short 		itemType;
 	Rect		box;
 	Handle		item;
 	Boolean		retVal;
 	long		l;
 	char		*p;
 	Boolean		lastLine;
 	Handle		bufHandle;
 	long		lineLen;
 
#define 	ok	1
#define	cancel	2
#define	keyEsc	0x1B
#define	CmdOptShift (cmdKey|optionKey|shiftKey)
#define kKeyCodeZ	0x0600

 	retVal = 0;
 	*itemHit = 0;
	if (((*theEvent).what == keyDown) || ((*theEvent).what == autoKey))  {
 		register unsigned char	theChar;
		theChar = BitAnd((*theEvent).message, charCodeMask);
		switch (theChar)	{
		
		case '.':
			if (!BitAnd((*theEvent).modifiers, cmdKey)) // if not Cmd-., 
				break;				    //   break;
		case keyEsc:
			*itemHit = cancel;			// cmd-. or Esc 
			retVal = true;
			break;
			
		case 0x0D:					// CR
		case 0x03:					// enter
			*itemHit = ok;
			retVal = true;	
			break;	
			
		default:
			bufHandle = NewHandle(0);
			if ((bufHandle)
			&& (BitAnd((*theEvent).modifiers, CmdOptShift) == CmdOptShift)
			&& (BitAnd((*theEvent).message, keyCodeMask) == kKeyCodeZ))
			{
				SysBeep(1);			// give feedback
				retVal = true; 			// we handled it
				GetHistoryText(bufHandle);
				ZeroScrap();
				HLock(bufHandle);
				l = PutScrap(GetHandleSize(bufHandle),'TEXT', *bufHandle);
			}
			if (bufHandle) DisposeHandle(bufHandle);
		} // end switch
	}

	if (retVal && *itemHit)    {
		GetDItem(theWindow, *itemHit, &itemType, &item, &box);
		HiliteControl((ControlHandle)item, inButton);
		Delay(10, &l);
		HiliteControl((ControlHandle)item, 0);
	}
	
	return(retVal);
 }
 
 /*****************************************************************************
  DoAlert - put up the proper (1/2 button) alert, with the right text
 *****************************************************************************/
 
short BMTalkDLOG::DoAlert (short alertType, short numButtons, 
 			short strNum1, short strNum2, char *extraStr)
 {
 	char 	str1[258];
 	char	str2[258];
 	char	str3[2];		/* always empty, may be used */
 	short	alrtID;
 	short	item;
 	char	*p;
 	
 	GetIndString(str1, errSTRNum, strNum1);		/* from error strings */
 	GetIndString(str2, errSTRNum, strNum2);
 	str3[0] = 0;							/* empty PString */
 	
 	ParamText(&str1, &str2, extraStr, nil);
 	if (numButtons == 1) alrtID = SMALLONE;
 	else alrtID = SMALLTWO;
 	
 	p = str3;
 	if (extraStr != nil)
 	{
 		alrtID += 2;						/* makes room for third string */
 		p = extraStr;
 	}
 	ParamText(&str1, &str2, p, nil);		/* set up the strings */
 	
 	InitCursor();
 	switch(alertType)
 	{
 	case Anote: 	item = NoteAlert(alrtID, nil); 		break;
 	case Acaution: 	item = CautionAlert(alrtID, nil); 	break;
 	case Astop:	item = StopAlert(alrtID, nil); 		break;
 	}
 	DrawDialog(dMainPtr);
 	return(item);
 }

 /*****************************************************************************
  MatchingMenuItem - return the item # from the menu which matches the str
  *****************************************************************************/
 
 short BMTalkDLOG::MatchingMenuItem (char *str, ControlHandle theCtl)
 {
 	MenuHandle 	mHdl;			/* handle to the menu */
 	short		numItems;		/* number of items in the menu */
 	short 		i;
 	char		mStr[64];		/* menu item's string */
 	popupPrivateData **pHdl;		/* handle to private data */
 	
 	pHdl = (popupPrivateData**)(*theCtl)->contrlData; // point at the private data
 	mHdl = (*pHdl)->mHandle;	// get the menu handle
 	numItems = CountMItems(mHdl);
 	for (i=numItems; i > 0; i--)
 	{
 		GetItem(mHdl, i, (StringPtr) mStr);
 		if (EqualString((StringPtr) str, (StringPtr) mStr, 0, 0)) break; /* not case or diac. sensitive */
 	}
 	return (i);
 }
 
/*****************************************************************************
 GetPopupItemText - returns the i'th item from a popup menu in str
 *****************************************************************************/

 void BMTalkDLOG::GetPopupItemText(ControlHandle theCtl, short itemNo, char *str)				
 {
 	MenuHandle 		mHdl;
 	popupPrivateData 	**pHdl;		/* handle to popup's private data */
 	
 	pHdl = (popupPrivateData**)(*theCtl)->contrlData; // point at the private data
 	mHdl = (*pHdl)->mHandle;	// get the menu handle
  	GetItem((MenuHandle) mHdl, itemNo, (StringPtr) str);
 }
 
 /*
 * FixupPopupControl() --
 *	Fixing up a CNTL's CDEF id if the System 7 popup menu CDEF isn't available.
 */
#define kOldPopUpMenuCDEF	1280
void FixupPopupControl(short id)
{
	Handle		cdef_h;				/* handle to CDEF resource */
	Handle		cntl_h;				/* handle to CNTL resource */
	short		old_proc_id;			/* proc_id field in CNTL */
	
	/* first load CNTL resource */
	cntl_h = GetResource('CNTL', id);
	if (cntl_h == nil)
		return;
	LoadResource(cntl_h);
	
	/* now try to load System 7 popup menu CDEF */
	cdef_h = GetResource('CDEF', popupMenuProc >> 4);
	
	/* if it failed, fix up CNTL resource to point to our CDEF */
	if (cdef_h == nil)
	{
		old_proc_id = *(short *) (*cntl_h + 16);
		*(short *) (*cntl_h + 16) = kOldPopUpMenuCDEF | (old_proc_id & 0xF);
	}
} /* FixupPopupControl() */

/*
 * GetNewPopupControl() --
 *	Load a pop-up menu control, first fixing up the CDEF id if the System 7 popup menu CDEF
 *	isn't available.
 */
ControlHandle GetNewPopupControl(short id, WindowPtr wp)
{
	FixupPopupControl(id);	
	return GetNewControl(id, wp);
} /* GetNewPopupControl() */

