/******************************* LAPMgrGlue.c *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	17 Jan 93 

 This file encapsulates everything that an application might want
 to do with the LAP Manager.  It provides glue routines to all the
 LAP Manager calls.

*/

#include <stddef.h>
#include <AppleTalk.h>
#include <OSUtils.h>
#include "LAPMgrGlue.h"

/*******************
 * L M W R T G E T *
 *******************

 LMWrtGet returns the currently selected atlk connection.
 
*/

ProcPtr * LMWrtGet(short thePort)
{
	asm 	{
		moveq	#LWrtGet,D0		// LWrtGet parameter
		move.l	D4,-(SP)
		move.w	thePort,D4		// the port
		movea.l	LAPMgrPtr,A0
		jsr	2(A0)
		move.l	(SP)+,D4
		move.l	A0,D0			// D0 is return value
	}
}

/*************************
 * L M W R T I N S E R T *
 *************************

 LMWrtInsert installs a new Atlk connection on the specified port.  The 
 LAPManager uses the flags as described on p. 33 of the MACPG. 
 The LAP Manager also tries to get an unused node adrs maxTries times
 
*/

short LMWrtInsert(void *atlkRsrc, short flags, short maxTries, short thePort)
{
	asm {
		move.l	D4,-(SP)
		moveq	#LWrtInsert,D0		// LWrtInsert parameter
		move.w	thePort,D4		// the port
		movea.l	atlkRsrc,A0
		move.w	flags, D1
		move.w	maxTries,D2
		movea.l	LAPMgrPtr,A1
		jsr	2(A1)
		move.l	(SP)+,D4		// D0 holds return value
	}
}

/*************************
 * L M W R T R E M O V E *
 *************************

 LMWrtRemove removes the AppleTalk connection for the specified port.
 
*/

short LMWrtRemove(short thePort)
{
	asm {
		moveq	#LWrtRemove,D0		// LWrtGet parameter
		move.l	D4,-(SP)		// save D4
		move.w	thePort,D4		// the port
		movea.l	LAPMgrPtr,A1
		jsr	2(A1)
		move.l	(SP)+,D4
	}
}

/***************************
 * L M R D D I S P A T C H *
 ***************************

 LMRdDispatch is called by the atlk whenever a frame has arrived from the link. 
 This call passes the frame up to the .MPP protocol stack. On exit, the packet
 and any error conditions have been handled by the LAP Mgr and LMRdDispatch 
 (this code).
 
 The atlk code passes in a pointer to the (entire) frame, its length (excluding 
 CRC bytes), the addrRefnum for the link address, and a pointer to the MPP 
 variables.
 
 LMRdDispatch stuffs these values into an array to make it easy to call the 
 assembly code:
     offset 	0  (rD1)	remaining packet length
     		1  (rD2)	LAPType + addrRefNum (in high word)
     		2  (rD3)	ptr to link source adrs (extended)
     		3  (rA0)	avail to atlk - preserved until ReadRest call
     		4  (rA1)	avail to atlk - preserved until ReadRest call
     		5  (rA2)	ptr to .MPP variables
     		6  (rA3)	ptr past 5 hdr bytes in .MPP toRHA (non-ext)
     		7  (rA4)	ptr to ReadPacket routine
     		8  (rA5)	avail for .MPP; restored after ReadRest called
     		
 Simplifications to the packet delivery mechanism:
 
 1)	LMRdDispatch always receives control after it calls the LAP Mgr's 
 LRdDispatch routine. To do this, it pushes two return addresses (a "good" and 
 "error" entry point) onto the stack. Then it loads the values above into the 
 proper registers. Finally, LMRdDispatch calls the LAP Manager. 

 The LAPMgr will attempt to pass the packet off to a protocol handler. If it is 
 successful, it'll first pop off the first "error" address" (@1 in this code) 
 transfer to the protocol handler, and wind up returning to @2. If no protocol 
 handler is there, the LAP Mgr will return to @1, which must pop off @2. In 
 either case, the stack is now prepared to return to the original caller of 
 LMRdDispatch.
 
 2)	We keep a pointer to the next unread char of the frame in register A0. 
 Protocol handlers will preserve this value between calls to ReadPacket
 and ReadRest. 
 
 3)	We keep the count of the number of bytes remaining in the frame in 
 register A1 (word). 
 
 4)	The LAPMgr has stringent rules for preserving A4 & A5 and restoring them
 after the ReadRest call. This is because it expects the protocol handler never
 to return to the LMRdDispatch caller. Because of 1) above, we can save A4 &
 A5 on entry, then restore them before exiting this routine. 
 
*/

void	LMRdDispatch (char * theData, short theLen, short theAddrRef, 
			Ptr linkAdrs, Ptr MPPVars)
{ 
	long	regs[9];			// to hold the register values
	char	lapType;			//
	register Ptr	s,d;			// temp pointers
	register long	l;

	asm	{				// save regs we're gonna change...
		movem.l	D1-D3/A0-A5,-(SP)
	}
	
// fill in the RHA with the first five bytes of the frame (LAP S&D, Type, Len)
	
	d = MPPVars+1;				// toRHA (Top Of RHA) is mppVars+1
	s = theData;				// point at start of AALAP Frame
	*d++ = *s++;				// copy LAP Dest addr 
	*d++ = *s++;				// copy LAP Src addr
	lapType = *s;				// remember the LAP Type
	*d++ = *s++;				// copy LAP Type 
	*d++ = *s++;;				// and hibyte of length
	*d++ = *s++;;				// and lobyte of length

// fill in the regs array values

	l = theAddrRef;
	l = (l << 16) + lapType;		
	regs[0] = theLen-5;			// D1: number of char's remaining 
	regs[1] = l; 				// D2: LAPType + addrRefNum in hiword
	regs[2] = (long) linkAdrs;		// D3: link source address
	regs[3] = (long) s;			// A0: pointer to next unused byte
	regs[4] = theLen-5;		 	// A1: length of unread data
	regs[5] = (long) MPPVars;		// A2: the MPP variables
	regs[6] = (long) d;			// A3: points past first 5 bytes in RHA	
	regs[7] = ReadPacketCode();		// A4: points at ReadPacket code	
	regs[8] = 0x50FF8001;			// A5: avail to .MPP	
	
// and call the LAP Manager

	asm {
//		move.w	SR,-(SP)		; push the current interrupt level
//		ori.w	#0x0700,SR
//		and.w	#0xFEFF,SR		; set interrupt level to 6
		moveq	#LRdDispatch,D0		; LRdDispatch parameter
		pea	@2			; push our "good" return adrs
		pea	@1			; and our "error" return adrs
		move.l	LAPMgrPtr,A1		; A1 = LAP Mgr entry point
		pea	2(A1)			; push (LAP Mgr+2)
		movem.l	regs,D1-D3/A0-A5	; load up the registers
		rts				; and transfer to the LAP Mgr	

; The LAP Manager does its stuff here, and returns to @1 or @2

	@1:					; came back here if no prot. hdlr
		nop	;dc.w	0xA9FF			; STOP! DEBUG!
		addq	#4,SP			; pop "good" return address
		
	@2:					; good return point
//		move.w	(SP)+,SR		; restore interrupt level
		movem.l	(SP)+,D1-D3/A0-A5 	; restore the registers
	}
}
			
/*********************
 * L S E T I N U S E *
 *********************

 LSetInUse informs the LAP Manager that the address it suggested (in the LAPEnq
 call) is already in use. 
 
*/

short	LMSetInUse (Ptr MPPVars)
{
	asm {
		moveq	#LSetInUse,D0		// LSetInUse parameter
		move.l	A2,-(sp)
		movea.l	MPPVars,A2		// A2 points to the .MPP vars
		movea.l	LAPMgrPtr,A1
		jsr	2(A1)
		move.l	(sp)+,A2
		moveq	#0,D0			// return noErr
	}
}


short	LMGetSelfSend (Ptr MPPVars) 
		{ DebugStr("\pLMGetSelfSend"); return (noErr); }
short	LMAARPAttach (long protocolType, short drvrRefNum, ProcPtr theListener)
		{ DebugStr("\pLMAARPAttach"); return (noErr); }
short	LMAARPDetach (long protocolType, short drvrRefNum)
		{ DebugStr("\pLMAARPDetach"); return (noErr); }
long	LMGetATalkInfo (void)
		{ DebugStr("\pLMGetATalkInfo"); return (noErr); }
long	LMGetPortInfo (short thePort)
		{ DebugStr("\pLMGetPortInfo"); return (noErr); }
short	LMOpenADEV (short adevFileID, short* index, char* filename, short* fileRefnum)
		{ DebugStr("\pLMOpenADEV"); return (noErr); }

/***********************
 * L N S E T I N U S E *
 ***********************

 LNSetInUse informs the LAP Manager that the address it suggested (in the LAPEnq
 call) is already in use. 
 
*/

short	LMNSetInUse (Ptr MPPVars, short thePort)
{
	asm {
		moveq	#LNSetInUse,D0		// LNSetInUse parameter
		movem.l	A2/D4,-(SP)
		movea.l	MPPVars,A2		// A2 points to .MPP variables
		move.w	thePort,D4		// D4 has the port
		movea.l	LAPMgrPtr,A1
		jsr	2(A1)
		movem.l	(SP)+,D4/A2
		moveq	#0,D0			// return noErr
	}
}


short	LM802Attach (ProcPtr handler, Ptr protSpecification, short ENETrefNum)
		{ DebugStr("\pLM802Attach"); return (noErr); }
		
short	LM802Detach (Ptr protSpecification, short ENETrefNum)
		{ DebugStr("\pLM802Detach"); return (noErr); }

/*******************
 * L M G E T A E Q *
 *******************

 LMGetAEQ - Get pointer to AppleTalk Event Queue
 
*/
 
QHdr *LMGetAEQ(void)
{
	asm {
		moveq	#LGetAEQ,D0		// LGetAEQ parameter
		movea.l	LAPMgrPtr,A1
		jsr	2(A1)			// A1 will hold return AEQ
		move.l	A1, D0									
	}
}

/*******************************
 * R E A D P A C K E T C O D E *
 *******************************

 ReadPacketCode returns the address of the ReadPacket code. This routine
 (and its friend, ReadRest) are discussed below.
 
 ***********************	 *******************
 * R E A D P A C K E T *   and   * R E A D R E S T *
 ***********************	 *******************
  
 ReadPacket and ReadRest are the two routines which return the next piece of a 
 received packet to the caller. This code is written based on the assumption 
 that the entire frame has already been read into memory. This makes it 
 considerably simpler than the LocalTalk case described in IM-II pps 326-328. 
 This assumption is safe for slow links such as a ModemTalk link (which might 
 take as long as 5 seconds to assemble a complete frame) or for fast links 
 (such as an EtherTalk link) where processor speed prohibits switching the flow 
 of control based on bits of the arriving frame.
 
 ReadPacket - read in the specified number of bytes into the specified
 	buffer. It is an error to request more bytes than have been received.

 ReadRest - read in the rest of the packet, putting the specified number
	of bytes into the specified buffer.  Error if packet longer than buffer.

 Call:
	A0,A1 = registers reserved by the atlk code. In particular, A0 points 
		to the next unconsumed byte of the packet; A1.w holds the 
		count of unconsumed bytes of the packet.
	A2 =  MPP variables
	A3 -> buffer to read into
	A4 -> start of ReadPacket
	D1 =  number of bytes left to read in the frame
	D3 =  byte count to read (word)

 Return:
	D0 changed
	D1 number of chars still unread (ReadPacket); modified (ReadRest)
	D2 preserved
	D3 = number of bytes of request (in D3) that were *not* copied. Thus,
	    = 0 if exact number of bytes requested were read
	    > 0 indicates packet smaller than requested count
	    < 0 indicates packet larger than requested, data was truncated
	A0,A1 preserved by ReadPacket, modified by ReadRest
	A3 -> one past where last character went
	A4,A5 restored (by ReadRest when packet's all in or error)
	Z flag set if no error was detected. ReadPacket returns an error
		(Z cleared) if D3 != 0; ReadRest always returns Z set true.
	
	NOTE: CRC bytes are not included in counts
*/

long ReadPacketCode (void) { 

	asm {
	lea	@ReadPacket,A0
	move.l	A0,D0
	rts
	
@ReadPacket: bra.s   @DoRP		; Entry point for ReadPacket code

@ReadRest:   
	move.l	A1,D1			; get remaining length into D1
	moveq	#0,D0			; clear high word
	move.w	D1,D0			; plan to copy entire length
	sub.w	D1,D3			; D3 = amount of request *not* copied
	bpl.s	@10			; go if we copy entire length
	add.w	D3,D0			; D0 = D1 + (D3-D1) = orig D3
	clr.w	D3			; gonna copy orig D3; D3 = 0 shows that
@10	
	sub.w	D0,D1			; D1 now holds count of remaining bytes
		    
; At this point:
;	A0 points to next unread byte
;	A3 points to dest buffer
;	D0 has number of bytes to copy
;	D1 has number of bytes unused 
;	D3 has proper return value

	move.l  D0,-(SP)     		; push length to be copied
        move.l  A3,A1               	; point at the dest buffer
        _BlockMove                  	; copy the data
	move.l (SP)+,D0			; get back D0
        adda.w	D0,A0     		; update the unconsumed ptr
        adda.w	D0,A3       		;    and the dest pointer
        move.l	D1,A1			; save the un-consumed length
        moveq   #0,D0			; set the CCR (Z always set - no link err)
        rts

; The real ReadPacket code

@DoRP:
	bsr.s	@ReadRest		; copy the data
	tst.w	D3			; set CCR depending on D3
	beq.s	@1			; DEBUG!
	nop		;dc.w	0xA9FF			; DEBUG!
@1:					; DEBUG!
	rts				; and return
}
}