/******************************** MTalkGlobals.c ******************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	25 April 93 

 This file contains definitions for the global variables used by ModemTalk.
 
*/

#include 	<Serial.h>
#include 	<Retrace.h>
#include 	<AppleTalk.h>
#include	"atlk.h"
#include	"ModemTalk.h"
#include 	"mySetupA4.h"			// local copy of A4 global stuff
#include	"LAPMgrGlue.h"
#include	"MTadev.h"
#include	"BPrefsFile.h"
#include 	"BMTPrefsFile.h"
#include 	"BMTalkDialer.h"
#include 	"BMTalkDLOG.h"

#pragma mark ModemTalk-specific globals

long		mtlkg;			// magic cookie - should be 'mtlk'
long		modemTalkVers;		// DLM converted to seconds since 1904 
short		myRouter;		// other end's number (from IM/UR)
short		propNode;		// LAPMgr's proposed node number
Boolean		hasAdrs;		// true if IM/UR handshake worked 
Boolean		AALAPup;		// true if interrupt activity should progress
short		linkState;		// current state of the link (enum linkStates)
short		serPortConfig;		// serial port configuration (SerReset)
Ptr		histBuf;		// pointer to the history log buffer
char		magicChar[32];		// char's to be escaped in frame

// 	Input side globals - used only by input code

short		inRefNum;		// refnum for the input driver
ProcPtr		pbReadCompl;		// proc to be used in inPB's completion routine
long		inGlobPtr;		// pointer to globals info
IOParam		inPB;			// PB for input driver
char		*inBuf;			// pointer to the input buffer
char		*inFree;		// next free character in inBuf
char		*inNext;		// next un-consumed char in inBuf
char		*inLast;		// last char of PBRead buffer
short		inputState;		// current state of input buffer
short		inCRC;			// accumulated CRC
short		rcvdLen;		// count of de-escaped chars
char		*interruptBuf;		// Serial driver's interrupt buffer

// 	Output side globals - used only by output code

short		outRefNum;		// refnum for the output driver
Boolean		nIM;			// true if we should send IM frame
Boolean		nUR;			// true if we should send UR frame
short		numDisc;		// number of disc frames to be sent
long		outGlobPtr;		// pointer to globals info
IOParam		outPB;			// PB for output driver
char		*outBuf;		// pointer to the output buffer
char		*outFree;		// next free character in outBuf
short		IFSeqLen;		// length of the InterFrame sequence
char		interFrame[2];		// data to be sent between frames
short		outCRC;			// output side CRC

// 	VBL globals - Timers and such...

long		tStartedIM;		// time we started sending IM's
long		tLastIM;		// last time we sent an IM
long		tLastRx;		// last time we received complete frame
long		tLastRTMP;		// last time we sent up an RTMP pkt
long		tXOffHold;		// last time we received an XOff
long		tNextDisc;		// time to send the next DISC frame
long		tStartedNotif;		// time we put up the LinkDown notif
long		tLastVBL;		// time we last did "major" vbl task
long		tStartedConn;		// time we started this connection
short		tVBLtask;		// # ticks 'til next vbl task call

VBLTask		myVBL;			// the actual VBL queue element
IOParam		vblPB;			// used to send frames at VBL time
char		*vblBuf;		// pointer to buffer for VBL's AACtl's
DeferredTask	inDTask;		// deferred task for LAPReadCompl
DeferredTask	outDTask;		// deferred task for LAPWriteCompl
NMRec		theNotifRec;		// used to notify when link went down...
char		notifMsg[256];		// the loss-of-link notification message
SleepQRec	mySleepQElem;		// the sleep queue element
VBLTask		notifVBL;		// VBL to tear down the notification
Boolean		hasExpModem;		// true if we really have a Express/GeoPort modem
Boolean		hasWNE;			// true if WaitNextEvent is implemented
short		prevNet;		// previous network (before sudden disc)
short		prevNode;		// previous node number

//	Statistics - only updated one place in the code

long		numIMtx 	= 0;	// # IM's transmitted
long		numURtx 	= 0;	// # UR's transmitted
long		numDataTx 	= 0;	// # data frames transmitted
long		numPBWrites	= 0;	// # PBWrite calls
long		numTxErrs 	= 0;	// # times PBWrite failed
long		numTxTO 	= 0;	// # times LAPWrite timed out during IM/UR
long		numForcedXOn	= 0;	// # times we forced XOn
long		numENQs		= 0;	// # times LAPEnq called
long		numNodeReject	= 0;	// # times we called LSetInUse

long		numLAPReads	= 0;	// # times LAPReadCompl called
long		numIMrx 	= 0;	// # IM's received
long		numURrx 	= 0;	// # UR's received
long		numIFrx 	= 0;	// # IF's received
long		numDISCrx 	= 0;	// # DISC's received
long		numDataRx 	= 0;	// # data frames received
long		numRxCRCErr 	= 0;	// # received CRC errors
long		numShortRx	= 0;	// # short frames received
long		numLAPframes	= 0;	// # good LAP frames received
long		numUnkLAP	= 0;	// # unknown LAPtypes received
long		numLongLAP	= 0;	// # LAP frames that were too long
long		numFrameStarts	= 0;	// # opening kFrames seen

long		numVBL 		= 0;	// # times vbl task entered

