/******************************** MTalkLink.c *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	25 April 93 

 This file contains routines for starting and shutting down the ModemTalk
 link. 
 
*/

#include 	<Serial.h>
#include 	<Retrace.h>
#include 	<AppleTalk.h>
#include 	<GestaltEqu.h>
#include 	<Power.h>
#include	<LoMem.h>
#include	<string.h>
#include	<stdio.h>
#include	"atlk.h"
#include	"ModemTalk.h"
#include	"MTalkUtilities.h"
#include 	"mySetupA4.h"			// local copy of A4 global stuff
#include	"LAPMgrGlue.h"
#include	"MTadev.h"
#include 	"CCODLoad.h"
#include	"BPrefsFile.h"
#include 	"BMTPrefsFile.h"
#include 	"BMTalkDialer.h"
#include 	"BMTalkDLOG.h"
#include	"TrapAvailable.h"
#include 	"histlog.h"

/* Offsets within regs[] array for atlkCode */
#define rD0 (*regs).r[0]
#define rD1 (*regs).r[1]
#define rD4 (*regs).r[2]
#define rA0 (*regs).r[3]
#define rA1 (*regs).r[4]

#define strMTalkEstablished  	"\pModemTalk connection established."
#define strNoServerResponse 	"\pNo response from the dialup server."
#define strConnecting		"\pConnecting"
#define strLostMTalkConnection	"\pYour ModemTalk connection has been disconnected.\r\rUse the Network Control Panel to reconnect."
#define	strSTOP			"\pSTOP"
#define strNeedATalk56		"\pModemTalk requires AppleTalk 56 or greater"
#define strBadatlkOther		"\pBad call to atlkOther"
#define strFoundSERDopen	"\pFound SERD open"
#define	strAALAP		"\pAALAP\r"
#define strLinkDown		"\pLinkDown"
#define strBadPrefs1 		"\pWarning: Your preferences file is obsolete.\r\r"
#define strBadPrefs2 		"\pPlease delete the 'ModemTalk Prefs' file from the Preferences folder in the System folder."
#define strPrefsVers 		"\p1"
#define strHangingUp		"\pHanging up"
#define strManualDial		"\pManual Dial"
#define	strDirectConnect	"\pDirect Connect"
#define	strAin			"\p.AIn"
#define strAout			"\p.AOut"
#define strCantCloseSERD	"\pDidn't close SERD"
#define strLoadingCCOD		"\pLoading CCOD..."
#define strCouldntGetCCOD	"\pCouldn't get CCOD rsrc"

#define strCPrevPort	   	"Your previous connection was on port %s"
#define strCConnForMinutes 	"You have been connected with ModemTalk to port %s for %ld minutes."
#define strCConnForHours   	"You have been connected with ModemTalk to port %s for %ld hours, %ld minutes."

#pragma mark Internal routines

void 	BogusInit(void);			// bogus prototype
OSErr	InitSerial(void * inName, void * outName, short baud); 
				// set up port, buffers, flow ctl, data rate
OSErr	DoAAConnect(void);	// handshake with AALAP device
Boolean	SendCR(void);		// send a CR, return TRUE if we heard reponse
OSErr	StartupAALAP(void);	// 
OSErr 	DoUsersBidding(void); 	// do user interface, call dialer, etc
OSErr 	BringUpLink(BMTalkDialer *dialer, BMTPrefsFile *thePrefs, BMTalkDLOG *theUI); 
						// dial, handshake w/ remote network
void	BringLinkDown(BMTalkDialer *dialer); // knock down link gracefully
void	DisposeSerial(void);	// close serial drivers
void	DisposeLink(void);	// close out the link
void	SendDiscPackets(void);	// sends three kDISC packets
void	VBLtask(void);		// periodic action
void	SendRTMP(short aNet, short aRouter);
long	SleepTask(void);	// handle sleep demands
void 	NotifVBLTask(void);	// VBL to remove the notification 
OSErr	LoadCCODIntoMemory(void); // load CCOD into appl or sys heap
Boolean MacNotCompatible(BMTPrefsFile *thePrefs); // return TRUE if Mac won't support ModemTalk
void	GetDurationLegend(long ticks, char * theStr);
void	GetPortString(char * theStr, short aNet, short aNode);
void	GetPrevPortString(char * theStr);

#define INTBUFLEN 512

#define PSLen(x) ((x) & 0x00FF)
enum dialFailures {configFail=-10034, dialFail, connectFail, handshakeFail};

/* these are bit positions, not indexes into Packed array [0..127] of Boolean */
#define CmdKey		48
#define ShiftKey 	63
#define OptKey		61

/* * * * * * * * * *
 * A T L K I N I T *
 * * * * * * * * * *
 
 Initialize the ModemTalk link. Allocate storage, set values, and start off 
 link
 
*/

OSErr	atlkInit(tRegs * regs)
{
	OSErr		retval;
	KeyMap  	theKeys;
	long		t;
	BMTalkDLOG	*theDLOG;
	OSErr		osstat;
	char		gPort[206];
	Str255		localMsg = strLostMTalkConnection;
	short		item;
	short		refnum;
	long		l;
	short		iCnt;
	long		expModmVers;
	short		resAttrs;
	
	{
	register long *lp;
	
	mtlkg = 'mtlk';				// fill in magic cookie
	modemTalkVers = 0;			// fill in date
	retval = controlErr;			// assume the worst
	
	hasWNE = TrapAvailable(0xA860);		// WaitNextEvent = 0xA860
	
	histBuf = GetHistBufAdrs();		// pointer to history buffer
	
	lp = (long*)&CurApName;			// check for running at INIT time
	if (*lp == -1)	{ 			//   (INIT process puts -1 there)
		GetKeys(&theKeys);			// if Shift key, abort
		if (BitTst((Ptr) &theKeys, ShiftKey)) {	//   installation
			SysBeep(1);
			goto bailout;
		}
		else  asm	{			// we need a graphics world
			pea	gPort[202]	
			_InitGraf
			_InitFonts
			_InitWindows
			_TEInit
			clr.l	-(sp)
			_InitDialogs
			clr.l	DeskHook	
			clr.l	DragHook
		}
	}
	}

	if (mppVersion < kMinATalkVers)	{ // bad version of AppleTalk
		ParamText(strSTOP, strNeedATalk56 , "", "");
		StopAlert(260, NULL);
		goto bailout;
	}
	
	if (hasGestalt)	{
		hasExpModem = (Gestalt ('3615',&expModmVers) == noErr);
	}
	
	inRefNum = outRefNum = 0;		// initialize important globals
	inBuf = outBuf = vblBuf = interruptBuf = NULL;
	AALAPup = hasAdrs = MPPup = AOpened = MTInserted = FALSE;
	linkState = kLinkNotConnected;
	myNet = 0;
	myNode = 0;
	myRouter = 0;
	prevNet = 0;
	prevNode = 0;
	propNode = 0;			
	mySleepQElem.sleepQType = 0;		// bug it...
	
#ifndef TESTAPPL
	inGlobPtr = (long)&inDTask;		// adrs of the deferred task
#else
	inGlobPtr = SetCurrentA5();		// remember our globals (test app only)
	outGlobPtr = inGlobPtr;
#endif

	retval = memFullErr;			// assume we can't get the space

	outBuf = NewPtrSys(MAXAABUF);		// allocate output buffer
		if (!outBuf) goto bailout;
	inBuf = NewPtrSys(MAXAABUF);		// allocate input buffer
		if (!inBuf) goto bailout;
	vblBuf = NewPtrSys(3*MAXAACTLFRAME);	// allocate buffer for VBL AActls
		if (!vblBuf) goto bailout;
	interruptBuf = NewPtrSys(INTBUFLEN);	// allocate serial int. buffer
		if (!interruptBuf) goto bailout;
	
	retval = LoadCCODIntoMemory();		// load the User interface code in
	if (retval) goto bailout;		//   bail out if not available
	CCODLoad();				// link in the segment...
	
// If we got this far, things are sailing along just fine. We'll always return
//	noErr from this point on...

	BlockMove(localMsg, notifMsg, PSLen(localMsg[0])+1);	// DEBUG!
	
	if (hasGestalt)		{
		osstat = Gestalt(gestaltPowerMgrAttr,&l);	// find out about power mgr
		if ((!osstat)
		&& (BitTst(&l, 31-gestaltPMgrExists)))  {
			mySleepQElem.sleepQLink = NULL;
			mySleepQElem.sleepQType = slpQType;
			mySleepQElem.sleepQProc = &SleepTask;
			mySleepQElem.sleepQFlags = 0;
			SleepQInstall(&mySleepQElem);		// no result: bull ahead
		}
	}
	
// Talk to the user, get their settings, and try to dial up the link
	
	linkState = kLinkNotConnected;		// since we just installed it
	retval = DoUsersBidding();		// try to connect, let settings 
						//  determine whether to show DLOG
	retval = noErr;				// don't report any problems:
						// we just couldn't dial...
	UnloadA4Seg(CCODLoad);			// push out the user interface...
bailout:	

	return (retval);
}

/* * * * * * * * * * * *
 * A T L K R E M O V E *
 * * * * * * * * * * * * 
 
 Shuts down the ModemTalk link.
 
*/

OSErr	atlkRemove(tRegs * regs)
{
	CntrlParam	pb;
	OSErr		retval;
	long		t;
	
#ifndef TESTAPPL
	SendDiscPackets();			// send out kDISC packets
#endif
	DisposeLink();				// let go of the serial stuff
	
	// release ModemTalk buffers
	
	if (outBuf) 	DisposPtr(outBuf); outBuf = NULL;
	if (inBuf) 	DisposPtr(inBuf); inBuf = NULL;	
	if (vblBuf) 	DisposPtr(vblBuf); vblBuf = NULL;
	if (interruptBuf) DisposPtr(interruptBuf); interruptBuf = NULL;
	
	if (hasGestalt)		{
		retval = Gestalt(gestaltPowerMgrAttr, &t);	// find out about power mgr
		if ((!retval)
		&& (BitTst(&t, 31-gestaltPMgrExists))
		&& (mySleepQElem.sleepQType == slpQType))  {
			SleepQRemove(&mySleepQElem);
		}
	}
	
#ifndef	TESTAPPL
	retval = VRemove((QElemPtr) &myVBL);
#endif
}

/*********************
 * A T L K O T H E R *
 *********************

 atlkOther may be used for back-door communication channel to link-specific
 code. For ModemTalk, the kDisplayUI selector will force the atlk to display 
 the user interface, allowing the user to reconfigure and connect/disconnect
 the link.
  
*/

OSErr	atlkOther(tRegs *regs)
{
	short	item;
	OSErr	osstat;
	
	histlog("atlkOther",rD0,0);
	if (rD0 == kDisplayUI)	{
		osstat = LoadCCODIntoMemory();		// load in user interface code
		if (osstat) return (osstat);		// bail out if it didn't work
		
		CCODLoad();				// call a routine in UI segment
		osstat = DoUsersBidding();		// now do what they want

		UnloadA4Seg(CCODLoad);			// push out the user interface...
		return (osstat);
	}
	else 	{
		DDebugStr(strBadatlkOther);
		return (controlErr);
	}
}

/* * * * * * * * * * * *
 * I N I T S E R I A L *
 * * * * * * * * * * * *
 
 Initialize the Serial port. First KillIO on both drivers, then close them to clear
 out any screwy state, then finally open them afresh.
 
 The parameters are the names of the input and output ports, and the 
 baudXXXX which should be passed to the SerReset() function.
 
*/

OSErr	InitSerial(void * inName, void * outName, short baud)
{
	OSErr	retval;
	SerShk 	hShkInfo;
	OSErr	osstat1;
	OSErr	osstat2;
	short	refnum;
	long	t;
	EventRecord theEvent;
	long	expModmVers;
	
	// first open in & out drivers 
	
	osstat1 = AssertDrvrOpen(inName,&refnum);
	osstat2 = AssertDrvrOpen(outName,&refnum);
	if ((osstat1 == 0) || (osstat2 == 0))	{
		DDebugStr(strFoundSERDopen);
		histlog("OpenSERD",osstat1, osstat2);
	}

	if (outRefNum || inRefNum)	{	// they've already been opened
		histlog("CloseSERD",inRefNum,outRefNum);
		DisposeSerial();		// close 'em to clear their state
	}
	
	retval = OpenDriver((char *) outName,&outRefNum); // Now the real opens
	if (retval) return (retval);			// 	output first, 
	retval = OpenDriver((char *) inName,&inRefNum);	// 	then input side
	if (retval) return (retval);
	
// If we have an Express Modem, call WNE while its background 
// 	application is starting up (as a result of OpenDriver...)

	if (hasExpModem)	{	// if we have express modem
		t = TickCount();	// spin loop calling WNE
		while ((TickCount() - t) < SECOND) { // for one second
			if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }
		}
	}
	
// Now finish serial driver initialization

	retval = SerSetBuf(inRefNum, interruptBuf, INTBUFLEN);
//	histlog("SetBuf",retval,0);
	if (retval) return (retval);
	
	hShkInfo.fXOn = TRUE;
	hShkInfo.fCTS = FALSE;
	hShkInfo.xOn  = 0x11;
	hShkInfo.xOff = 0x13;
	hShkInfo.errs = 0;
	hShkInfo.evts = 0;
	hShkInfo.fInX = TRUE;
	hShkInfo.fDTR = FALSE;
	retval = SerHShake(outRefNum, &hShkInfo);	// only tell output side
//	histlog("SerHShake",retval,0);
	if (retval) return (retval);
	
	serPortConfig = baud + data8 + stop10 + noParity;
	retval = SerReset(outRefNum, serPortConfig);	// set the serial config
//	histlog("SerReset",retval,0);
	if (retval) return (retval);
	
	retval = OutControl(dtrControl,leaveOff);	// drop DTR when closing
//	histlog("dtrControl",retval,leaveOff);
	if (retval) return (retval);
	retval = OutControl(assertDTR,0);		// assert DTR
	
	histlog("InitSerial",retval,0);
	return (retval);			
}
 
/* * * * * * * * * * * * * * *
 * D I S P O S E S E R I A L *
 * * * * * * * * * * * * * * *

 Close the Serial drivers, deallocating the interrupt time buffer as well
 
*/

void	DisposeSerial(void)
{
	OSErr	retval;
	EventRecord theEvent;
	
	histlog("DispSerial",0,0);
	if (!hasExpModem)	{
		retval = KillIO(inRefNum);	// stop any pending IO
		histlog("KillIOIn",retval,0);
		retval = KillIO(outRefNum);
		histlog("KillIOout",retval,0);
		retval = SerSetBuf(inRefNum, NULL, 0);	// restore default input buffer
		histlog("SerSetBuf",retval,0);
		retval = CloseDriver(inRefNum);	// close input side first
		histlog("CloseIn",retval,0);
	}
	retval = CloseDriver(outRefNum);	// then output side
	histlog("CloseOut",retval,0);
	inRefNum = outRefNum = 0;		// no more drivers
}

/* * * * * * * * * * * * *
 * D O A A C O N N E C T *
 * * * * * * * * * * * * *
 
 Handshake with an AALAP-capable device to cause it to go into AALAP mode. We 
 send up to 10 CR's to the device, until we hear 0xA5 or "@ ". If we get a good
 response, we send "AALAP <CR>" and declare the link up.
 
 If we don't hear a good response, we should whine to the user.
 
*/

OSErr	DoAAConnect(void)
{
	short 	i;
 	
 	for (i = 1; i<=15; i++)		{	// send CR 15 times, look for response
 		if (SendCR()) 	{		// they said something good
 			SendSlowString((char *) strAALAP);
 			return(noErr);
 		}
 	}
 	return (-1);			 	// bad status - didn't hear from 'em
}

/* * * * * * * *
 * S E N D C R *
 * * * * * * * *
 
 Send a CR, listen for 0xA5 or '@ ' within 30 ticks. If one arrives, return
 TRUE, else FALSE. NED's don't really honor "AALAP" when elicited by the ENQ.
 
*/

 Boolean SendCR (void)
 {
 	long		t;				// temp for times
 	Boolean		sawAT;				// true if we saw "@"
 	OSErr 		osstat;				// return status
 	char		c;				// last char read in
 	long		numChars;			// num chars avail on input
 	
 	c = kCR;
	osstat = SendChar(&c);				// send the CR
		
 	sawAT = FALSE;
 	t = TickCount();				// start timer
 	while ((TickCount() - t) < 30L)		{	// spin up to 30 ticks
 	
		osstat = GetChar(&c);			// get next character

		if (osstat != eofErr)		{
			switch (c)		{
		case kFrame:				// Frame char?
//		case kENQ:				// *NOT* ENQ - NED's not ready
			return(TRUE);			//  BINGO!
		case kAT:				// @-sign?
			sawAT = TRUE;
			break;			
		case kSpace:				// saw a space?
			if (sawAT) return (TRUE); 	// "@ "? return TRUE
			else sawAT = FALSE;		// otherwise, false alarm
			break;
			}
		}	
 	}						// endwhile
 	
 	return (FALSE);					// bad news...
}

/*****************
 * V B L T A S K *
 *****************
 
 This VBL task checks state of the ModemTalk link, and supplies any periodic
 actions required. The VBL is scheduled every tick. In particular, we do this:
 
 if (LAPWrtRtn was deferred)
 	handle the LAPWrtRtn
 	
 if (PBRead(inbuf) completed)
 	handle the arrived data
 	
 if (last "major" VBL was > 1/2 second ago) 
 	remember this time 
	 
	if (MPP is up, and we haven't send an RTMP for 10 seconds)
		SendRTMP(current net & router)
	 
	if (we sent an IM, and haven't heard back)
		queue another IM until 20 seconds have passed,
		else declare the link down

	if (we've received an Xoff more than one second ago)
		Force an XOn to keep things going
	 	
	if (we need any IM/UR/DISC frames)
	 	queue and send the required frames
 endif	 
 
 if (link is still up)
 	reschedule the VBL task for next tick
 else remove the VBL from the queue (we're going down)
  
*/

void 	VBLtask (void)
{
	OSErr		osstat;
	long		t;			// current TickCount() value
	long		aaLen;			// length of AA control frames
	CntrlParam	aPB;			// param block for Immed. Status
	
#ifndef TESTAPPL
	SetUpA4();				// got globals again (in A4)
#endif	

	if (deferLAPWrtRtn)	{		// has the completion been held off?
		LAPWrtRtn();			// if so, do it now...
	}
	
	if (inPB.ioResult <= noErr)	{
		LAPReadCompl();
	}
			
	t = TickCount();			// get the current time
	if ((t - tLastVBL) > HALFSECOND) {	// Do we need major work?
		
		tLastVBL = t;				// remember this time
		numVBL++;				// update stats...
			
		if (MPPup && (t - tLastRTMP > 10*SECOND)) { 	// need an RTMP?
			histlog("SendRTMP",myNet, myRouter);	// DEBUG!
			SendRTMP(myNet, myRouter);		// keep sysABridge happy
			tLastRTMP = t;
		}
		
		CheckRxTimer();				// do we need an IM?
		
		aPB.ioCRefNum = outRefNum;		// DEBUG
		aPB.ioCompletion = NULL;
		aPB.csCode = 8;				// SerStatus code
		osstat = PBStatusImmed(&aPB);		// Immediate Status call
		if (osstat) histlog("vblPBStat",osstat,0);	// DEBUG!
		if (aPB.csParam[2] & 0x00FF)	{	// received XOff?
			if (!tXOffHold)		{	// and we haven't started timer
				tXOffHold = t;		// remember time
			}
			else if (t-tXOffHold > SECOND)	{ // if it has been a long time
				numForcedXOn++;
				histlog("xoffhold",0,0);
				osstat = OutControl(forceXon,0); // force Xon
			}
		}
		else 	tXOffHold = 0;			// no Xoff - clear the timer
		
	// should we send AALAP ctl pkts?
	
		if (tStartedIM)			{ 	  // link going away?
			if (t - tStartedIM > 20*SECOND) { // gone for (too long) a time
				LinkWentDown();	  	  // warn the user
				tStartedIM = 0;		  // and stop worrying
			}
			else if (t - tLastIM > SECOND) 	{ // should we poke him again?
				nIM = TRUE;
				tLastIM = t;
	//			histlog("sentIM",0,0);
			}
		}
		if ((nIM || nUR || numDisc) && (vblPB.ioResult <= 0))	{ 
			aaLen = FrameAALAPCtls(vblBuf); // frame up the packets
			vblPB.ioReqCount = aaLen;
			vblPB.ioBuffer = vblBuf;
			osstat = PBWriteAsync(&vblPB);	// Async write
			histlog ("VBL PBW",aaLen,t);	// DEBUG!
		}
		
	}
	if (AALAPup)				// if we're still running
		myVBL.vblCount = 1;		// 	reschedule the VBL task
	else 		{
		osstat = VRemove((QElemPtr) &myVBL); // otherwise, remove it
		if (osstat) histlog("VRemove", osstat, 0);
	}
	
#ifndef TESTAPPL
	RestoreA4();				// back to the real world
#endif
}


/*******************
 * S E N D R T M P *
 *******************
 
 SendRTMP passes an RTMP Plus packet to the LAP manager, supplying the
 a net and router node number for the .MPP to munch on.
 
 This code lays out the information in a buffer on the stack and calls the
 LMRdDispatch code. This works because the LAP Manager glue code is
 synchronous, and returns immediately.
 
 The format of an RTMP Response packet is shown below. All fields are one byte
 wide, unless otherwise specified:
 
	 LAP Dest
	 LAP Src
	 LAPType (= 1)
	 Length (2 bytes, = 9)
	 Dest Skt (=1)
	 Src Skt (=1)
	 DDPType (=1)
	 Sender's network number (2 bytes)
	 ID length (one byte, measured in bits, = 8)
	 Sender's ID (Router node number, for ModemTalk)
 
*/
void	SendRTMP(short aNet, short aRouter)
{
	char	rtmpData[13];
				
// start packet to odd byte, so length will begin on even byte...

	rtmpData[1] = 0x00FF;			// LAP Dest
	rtmpData[2] = aRouter;			// LAP Src
	rtmpData[3] = 1;			// DDP Type
	rtmpData[4] = 0;			// length
	rtmpData[5] = 9;			// length
	rtmpData[6] = 1;			// Dest skt
	rtmpData[7] = 1;			// Src skt
	rtmpData[8] = 1;			// DDPType 
	rtmpData[9] = (aNet>>8) & 0x0FF;	// network #
	rtmpData[10] = (aNet) & 0x0FF;		// network #
	rtmpData[11] = 8;			// 8-bit long ID
	rtmpData[12] = aRouter;			// router node number
	
	LMRdDispatch(&rtmpData[1], 12, addrRefNum, NULL, mppVars);
}

/***************************
 * L I N K W E N T D O W N *
 ***************************
 
 LinkDown is called whenever the atlk detects that the link has gone down. 
 This is detected by a failure to hear good frames in response to IM packets.
 
 This code clears myNet and forces a RTMP packet to the .MPP stack. If
 the user re-dials, the newly arrived network number may force another
 MPP Open (with the appropriate LAPEnq's...).
 
 This code fires off a notification to the user as well. 
 
*/
 
void	LinkWentDown(void) 
{
	OSErr	osstat;
	
//	DDebugStr(strLinkDown);			// tell 'em we're down

	if (sysVersion >= 0x0600)		{	// if NM available
		tStartedNotif = TickCount();		// remember when it started
		notifVBL.qType = vType;			// initialize the VBL task block
		notifVBL.vblAddr = (ProcPtr) &NotifVBLTask;
		notifVBL.vblCount = 60*SECOND;		// go for a while
		notifVBL.vblPhase = 1;			// mark it to look different
		osstat = VInstall((QElemPtr) &notifVBL);// install the VBL task

		theNotifRec.qType = nmType;
		theNotifRec.nmMark = 0;
		theNotifRec.nmIcon = NULL;
		theNotifRec.nmSound = (Handle) -1;
		theNotifRec.nmStr = (StringPtr) notifMsg;
		theNotifRec.nmResp = (NMProcPtr) -1;
		theNotifRec.nmRefCon = 0;
		osstat = NMInstall(&theNotifRec);
	}
	else SysBeep(5);				// long, sad beep...
	
	prevNet = myNet;
	prevNode = myNode;				// remember for diagnostics
	
	DisposeLink();					// and set link to ground zero	
}

/*******************************
 * D O U S E R S B I D D I N G *
 *******************************
 
 If the user's willing, open the proper serial port, dial the phone, and get the 
 link going.
 
 Returns with linkState set properly
 
*/

OSErr DoUsersBidding(void)
{
	BMTPrefsFile	*thePrefs = NULL;
	BMTalkDLOG	*theDLOG = NULL;
	BMTalkDialer	*theDialer = NULL;
	OSErr		osstat;
	short		i,n;
	Handle		theScript;
	short		**s;
	char		showMsg[256];
	KeyMap		theKeys;
	Boolean		dialOnOpen;
	Boolean		isBootTime;
	Boolean		keyPressed;
	register long 	*lp;
	Boolean		linkIsDown;
	Boolean		uiDone;			// true if the User I/F is done

	thePrefs = new BMTPrefsFile;				// get the prefs file
	osstat = thePrefs->IPrefsFile('adev', 'AaTk', 1024, 'STR#', 256);
	if (osstat) goto bailout;
	
	// check for old preference file...

	thePrefs->GetAPref(sPrefsVers, showMsg, sizeof(showMsg));
	if ((thePrefs->NumPrefs() != kMTNumPrefs)
	|| (!EqualString(showMsg, strPrefsVers,false, false))) { // bad prefs file
		ParamText(strBadPrefs1, strBadPrefs2, "", ""); // just warn 'em
		NoteAlert(262, NULL);
	}

	theDLOG = new BMTalkDLOG;				// make a dialog
	theDLOG->InitMTalkDLOG(thePrefs);
	theDLOG->ClearDisplay();

	theDialer = new BMTalkDialer;				// make a dialer
	osstat = theDialer->IBMTalkDialer(thePrefs, theDLOG);
	
	GetKeys(&theKeys);				// look for keys to force UI
	keyPressed = BitTst(&theKeys, CmdKey) || BitTst(&theKeys, OptKey) 
			|| BitTst(&theKeys,ShiftKey);

	lp = (long*)&CurApName;				// check for INIT time
	isBootTime = (*lp == -1);			// CurApName will == -1

	dialOnOpen = thePrefs->GetDialOnOpen();		// should we dial on open?
	linkIsDown = (linkState != kLinkUp);		// is the link currently down?

	if ((prevNet != 0) && (prevNode != 0))		// we used to be connected
		GetPrevPortString(showMsg);
	else showMsg[0] = 0;
	if (isBootTime) {				// if it's boot time, 
		if (thePrefs->GetDialOnStartup()) {	// and we should dial...
			if (thePrefs->MacNotCompatible()) goto bailout;
			osstat = BringUpLink(theDialer,thePrefs,theDLOG); // dial 'em
			goto bailout;
		}
		else		{				
			osstat = noErr;				// no problem
			goto bailout;				// just exit
		}
	}
	else if (dialOnOpen && linkIsDown && !keyPressed) {
		theDLOG->ShowMTDLOG(linkIsDown, showMsg);		// showing the UI
		if (thePrefs->MacNotCompatible()) goto bailout;
		osstat = BringUpLink(theDialer,thePrefs,theDLOG); 
		if (!osstat) goto bailout;	// if noErr, exit, otherwise fall thru
	}
	else if (!linkIsDown && tStartedConn)	{
		GetDurationLegend(tStartedConn, showMsg);
	}

	// otherwise, we must do the user interface...
		
	uiDone = false;
	while (!uiDone) {
	
		linkIsDown = (linkState != kLinkUp);	// is the link currently down?
		theDLOG->ShowMTDLOG(linkIsDown, showMsg); // show the dialog

		uiDone = TRUE;				// assume we're quitting
		osstat = theDLOG->DoMTDLOG();		// talk to the user...
		
		if (!osstat)	{			// if they hit OK or Hangup
			theDLOG->ClearDisplay();
			
			if (!linkIsDown)	{	// if it's connected now
				theDLOG->AddToDisplay((char*) strHangingUp);
				BringLinkDown(theDialer);// drop the connection
			}
			else			{	// else, dial 'em up
				if (thePrefs->MacNotCompatible()) goto bailout;
				osstat = BringUpLink(theDialer,thePrefs,theDLOG);
				if (osstat)	{
					showMsg[0] = 0;
					DisposeSerial();
					uiDone = FALSE;
				}		
			}
		}
	} 

bailout:
	if (theDLOG) 	theDLOG->Dispose();
	if (thePrefs) 	thePrefs->Dispose();
	if (theDialer)	theDialer->Dispose();
	
	return(osstat);
}

/*************************
 * B R I N G U P L I N K *
 *************************
 
 Make the link go. Use the preferences to init the port, execute a dialing script,
 and handshake. Returns noErr or an OSErr which tells what failed.
 
*/

OSErr BringUpLink(BMTalkDialer *dialer, BMTPrefsFile *thePrefs, BMTalkDLOG *theUI)
{
	OSErr		osstat,retval;
	char		inPort[20], outPort[20];
	short		bpsValue;		// passed to SerReset()
	long		t;
	long		l, l1;
	EventRecord	theEvent;
	char		localBuf[80];
	char		portStr[12];
	

	InitCursor();
	bpsValue = thePrefs->GetSelectedSpeed(); // what did they choose
	myBPS = thePrefs->GetBitsPerSecond();
	thePrefs->GetPortNames(inPort, outPort);
	
	t = TickCount();
	AALAPup = FALSE;				// no more interrupt stuff
	osstat = InitSerial(inPort, outPort, bpsValue); // initialize the port
	histlog("InitSer",osstat,0);
	if (osstat) 	{
		retval = configFail;
		goto bailout;
	}
	while(TickCount()-t < SECOND) { /* spin loop to allow VBL to time out */
		if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }
	}

	thePrefs->GetAPref(sModemType,inPort, sizeof(inPort));
	
	// 	Manual Dial...
	if (EqualString(strManualDial,inPort, false, false))	{
		thePrefs->GetAPref(sPhNum,outPort,sizeof(outPort));
		StringToNum(outPort, &l);		// DEBUG!
		Delay(l*60, &l1);			// delay a while
	}
	// 	Direct Connect...
	else if (EqualString(strDirectConnect,inPort, false, false))	 			{			// else Direct connect
		theUI->AddToDisplay((char*)strConnecting); // give feedback
	}
	// 	Auto dial...
	else { 
		osstat = dialer->Dial();		// exec dialing script
		histlog("dialResult",osstat, 0);
		if (osstat) 	{
			retval = dialFail;		// and whine about it
			goto bailout;
		}
	}
	
// in any event, we should try to "AALAP<cr>"

	osstat = DoAAConnect();
	histlog("AAConnect",osstat,0);
	if (osstat) 	{
		theUI->AddToDisplay((char*)strNoServerResponse);
		retval = connectFail;
		goto bailout;	
	}
	
	osstat = StartupAALAP();	
	histlog("StartAA",osstat,0);
	if (osstat) 		{
		retval = handshakeFail;
		goto bailout;
	}

	tStartedConn = TickCount();			// remember when we started
	theUI->AddToDisplay((char*) strMTalkEstablished); // say it worked
	Delay(60, &l);					// wait a second...

	retval = noErr;
	
bailout:				
	if (retval)	{
		DisposeSerial();			// Close .xIn/.xOut
	}
	return(retval);
}

/*****************************
 * B R I N G L I N K D O W N *
 *****************************
 
 Bring the link down gently - send kDISC packets, then issue Dialer
 hangup commands.
 
*/

void	BringLinkDown(BMTalkDialer *dialer)
{
	long	l;
	
	tStartedConn = 0;
	SendDiscPackets();		// send disc's
	Delay(90,&l);			// wait a second
	dialer->Hangup();		// and tell dialer to hangup
	DisposeLink();			// finally, close vestiges of the link
}

/*************************
 * D I S P O S E L I N K *
 *************************
 
 Make the link non-functional, free up the serial ports.
 
*/

void	DisposeLink(void)
{
#define SysABridge 0x19

	TNetworkTransition	myTransEl;	// for AT Trans Queue
	long	t;
	OSErr	osstat1;
	OSErr	osstat2;
	short	refnum;
	
	AALAPup = FALSE;			// no more vbl/read stuff
	histlog("Link was",(myNet<<8)+myNode,0);	// DEBUG!
	myNet = 0;
	myNode = 0;				// we don't have an identity
	if (mppVersion >= 0x5700)	{	// if AT57, tell 'em...
		myTransEl.netValidProc = &isNetValid;
		myTransEl.newConnectivity = FALSE;
		myTransEl.private = FALSE;
		ATEvent(ATTransNetworkTransition, (Ptr)&myTransEl);
		histlog("atTransQ", ATTransNetworkTransition, FALSE);
	}	
	hasAdrs = FALSE;				
	asm	{
		move.l	mppVars,A0		// point at the real ATalk globals
		clr.b	SysABridge(A0)		// clear the router adrs
	}
	linkState = kLinkDropped;		// not connected anymore
	DisposeSerial();			// free the drivers
//	Delay(60, &t);				// wait for VBL etc. to time out
	osstat1 = AssertDrvrOpen(strAin,&refnum);
	osstat2 = AssertDrvrOpen(strAout,&refnum);
	if ((osstat1 == 0) || (osstat2 == 0))	{
		DDebugStr(strCantCloseSERD);
		histlog("closeSERD",osstat1, osstat2);
	}
}

/***********************
 * S T A R T A A L A P *
 ***********************
 
 Kick off interrupt time processing...
 
*/

OSErr	StartupAALAP(void)
{
	OSErr	retval;
	long	t;

	rcvdLen = 0;
	hasAdrs = FALSE;
	addrRefNum = 0xFFFF;			// no multi-node; all for User Node
	
	BitSet(magicChar,kEsc);			// set the special char's 
	BitSet(magicChar,kXOn);
	BitSet(magicChar,kXOff);
	BitSet(magicChar,kXOnP);
	BitSet(magicChar,kXOffP);
	BitSet(magicChar,kFrame);
	
	IFSeqLen = 0;
	inputState = noFrame;
	
	tStartedIM = 0;
	tLastIM = 0;
	tLastRx = 0;
	tLastRTMP = 0;
	tXOffHold = 0;
	tNextDisc = 0;
	tVBLtask = 1;				// every tick
	
	numDisc = 0;
	nUR = FALSE;
	nIM = TRUE;	
	
	prevNet = 0;				// we don't need this information
	prevNode = 0;				// any more

	myVBL.qType = vType;			// initialize the VBL task block
	myVBL.vblAddr = (ProcPtr) &VBLtask;
	myVBL.vblCount = 1;			// start real soon
	myVBL.vblPhase = 1;			// mark it to look different

	inDTask.qType = dtQType;		// initialize the deferred
	inDTask.dtAddr = &LAPReadCompl;		// 	task block
	inDTask.dtReserved = 0;			// must be zero
	pbReadCompl = &LAPReadCompl;		// point at completion routine
	 
#define _DTInstall	0xA082
	if (TrapAvailable(_DTInstall))		// use deferred if avail... 
		pbReadCompl = &DeferLAPReadCompl;
	
	InitPB(&outPB, outRefNum);
	outFree = outBuf;			// initialize output PB
	
	InitPB(&inPB, inRefNum);		// init the Input PB
	*inBuf = 'Q';				// set bogus data in first byte
						//	of input buffer (never changed)
	ClearReceiver();			// reset receiver's variables
		
	InitPB(&vblPB, outRefNum);		// initialize vbl PB
	
	AALAPup = TRUE;				// we can do int-time tasks
	
#ifndef	TESTAPPL
	LAPReadCompl();				// kick off async read

	retval = VInstall((QElemPtr) &myVBL);	// install the VBL task
	if (retval) return(retval);
	
	linkState = kLinkUp;		
	return(noErr);
	
#endif
}

/***********************
 * I S N E T V A L I D *
 ***********************
 
 ATTransNetworkTransition network validation function to indicate that
 the requested network is reachable or not...

*/

pascal long isNetValid(TNetworkTransition* theTrans, unsigned long theAdrs)
{
	long	msg[2];
	
	msg[0] = (long) 'isVa';
	msg[1] = (long) 'lid ';
	histlog((char*)&msg,theAdrs,theTrans->private);
	return(theTrans->private);
}

/*********************************
 * S E N D D I S C P A C K E T S *
 *********************************
 
 Send three kDISC packet (or wait up to 5 seconds), then turn off 
 interrupt processing.
 
*/

void	SendDiscPackets(void)
{
	long	t;
	EventRecord theEvent;
	
	histlog("SendDisc",AALAPup,0);
	if (AALAPup)	{			// if interrupt stuff active
		t = TickCount();
		numDisc = 3;
		while (numDisc)	{		// spin loop 'til we've sent 'em all
			if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }
			if (TickCount()-t > 5*SECOND) // or we time out
				break;
		}
		AALAPup = FALSE;		// we don't wanna play anymore	
	}
	linkState = kLinkSentDisc;
}

/*********************
 * S L E E P T A S K *
 *********************
 
 Shut down ModemTalk if we get a SleepDemand 

*/

long	SleepTask(void)
{
	long	sleepCommand;
	
	asm	{  move.l	D0,sleepCommand	}
	
	SetUpA4();
	
	histlog("SleepDemand",0,0);
	
	if (sleepCommand == sleepDemand)   {
		DisposeLink();		// shut down link, drivers, etc.
	}
	
	RestoreA4();
	
	return(0);
}

/***************************
 * N O T I F V B L T A S K *
 ***************************
 
 Remove the LinkDown notification after it's been on the screen for a while.

*/

void NotifVBLTask(void)
{
	OSErr osstat1, osstat2;
	
#ifndef TESTAPPL
	SetUpA4();				// got globals again (in A4)
#endif
	osstat1 = NMRemove(&theNotifRec);	// remove the Notification
	osstat2 = VRemove((QElemPtr) &notifVBL); // remove the VBL as well
	histlog("NMRemove",osstat1,osstat2);
	
#ifndef TESTAPPL
	RestoreA4();				// back to the real world
#endif
}

/****************************************
 * L O A D C C O D I N T O M E M O R Y  *
 ****************************************
 
 Load the CCOD resource into memory somewhere Try Appl heap first, 
 	then try System heap

*/

OSErr	LoadCCODIntoMemory(void)
{
	OSErr		retval;
	short		resAttrs;
	long		size;
	register Handle rCCODrsrc;

//	DebugStr(strLoadingCCOD);

	retval = memFullErr;
	SetResLoad(FALSE);			// don't really load it
	rCCODrsrc = NULL;
	rCCODrsrc = Get1IndResource('CCOD',1); 	// get the first one that's there...
	retval = ResError();
	if ((!rCCODrsrc) || retval)  {		// won't fit/not there!
		DDebugStr(strCouldntGetCCOD);
		goto bailout;
	}
	SetResLoad(TRUE);

// we know our CCODrsrc is actually there

	size = SizeResource(rCCODrsrc);		// get the size of the resource
	resAttrs = GetResAttrs(rCCODrsrc); 	// get the current attributes
	resAttrs &= ~resSysHeap;		// clear the system heap bit
	SetResAttrs(rCCODrsrc, resAttrs);  	// update the resAttrs
	ReleaseResource(rCCODrsrc);		// release the resource
	
// now the CCOD resource is aimed at the appl heap

	rCCODrsrc = Get1IndResource('CCOD',1); 	// try to load it into appl heap
	retval = ResError();
	
	if (retval)		{		// But wait! It won't fit
		resAttrs |= resSysHeap;		// set the system heap bit
		SetResAttrs(rCCODrsrc, resAttrs);  // and update them
		ReleaseResource(rCCODrsrc);	// get rid of this copy
		rCCODrsrc = Get1IndResource('CCOD',1); 	// and try again 
		retval = ResError();
		if (retval) goto bailout;
	}
	
	CCODLoad();				// call this to force in the CCOD rsrc
	HLock(rCCODrsrc);			// lock it down...
	retval = noErr;
	
bailout:
	return(retval);
}

// GetDurationLegend - given the TickCount that the connection started up,
//	return a Pascal string stating how long the connection has been up

void	GetDurationLegend(long ticks, char * theStr)
{
	long	minutes;
	long	hours;
	char	thePort[12];	// xxxxx/yyy, where xxxxx = net, yyy = node

	GetPortString(thePort, myNet, myNode);		// get our network address
	minutes = (TickCount()-ticks+1799)/3600;	// round time up to nearest minute
	if (minutes <= 60)
		sprintf(theStr,strCConnForMinutes, thePort, minutes);
	else 	{
		hours = minutes/60;
		minutes = minutes - (hours * 60);
		sprintf(theStr,strCConnForHours, thePort, hours, minutes);
	}
	c2pstr(theStr);

}

// GetPrevPortString - pass back a Pascal string which indicates previous port...

void	GetPrevPortString(char * theStr)
{
	char	thePort[12];
	
	GetPortString(thePort, prevNet, prevNode);
	sprintf(theStr, strCPrevPort, thePort);
	c2pstr(theStr);
}

// GetPortString - pass back a C string of the form xxxxx/yyy, which indicates
//	our network address

void	GetPortString(char * theStr, short aNet, short aNode)
{
	sprintf (theStr, "%d/%d", aNet, aNode);	// use ModemTalk's global values
}

