/*************************** MTalkUtilities.c *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	8 April 93 

 This file contains utility routines used by the ModemTalk connection. 
 
*/

#include	<Script.h>
#include 	<Serial.h>
#include 	<Retrace.h>
#include 	<AppleTalk.h>
#include 	<stdio.h>
#include 	<string.h>
#include	<AppleTalk.h>
#include	"atlk.h"
#include	"ModemTalk.h"
#include 	"mySetupA4.h"			// local copy of A4 global stuff
#include	"LAPMgrGlue.h"
#include	"MTadev.h"
#include	"BPrefsFile.h"
#include 	"BMTPrefsFile.h"
#include 	"BMTalkDialer.h"
#include 	"BMTalkDLOG.h"
#include 	"MTalkUtilities.h"
#include 	"histlog.h"

#define CHARPAUSE	3	 /* ticks between sending chars */

Boolean	GetFirstLine(char *theLog, long *logLen);
Boolean GetNextLine(char *logLine, long *logLen);
Boolean IsCancelEvent( EventRecord *theEvent );

/* these are bit positions, not indexes into Packed array [0..127] of Boolean */
#define CmdKey		48
#define ShiftKey 	63
#define OptKey		61

/*************
 * M U T E X *
 *************

 Mutex sets a bit in a semaphore (a short), and returns TRUE if the semaphore
 was initially zero. This implies that the semaphore is now set, and the caller
 may use the resource. When the caller is finished with the resource, it
 may simply clear the semaphore value by setting it back to zero.
 
*/

Boolean	Mutex (short *theSema)
{
	asm	{
	    movea.l	theSema,A0
	    tas		(A0)
	    seq		D0
	}
	
}

/* * * * * * * * * * * * * * * *
 * S E N D S L O W S T R I N G *
 * * * * * * * * * * * * * * * *
 
 Send a (Pascal) string, with many ticks between each at "typing speeds", since
 some modems cannot handle high speed data entry.
 
 This routine runs synchronously - that is, it doesn't return 'til it has sent
 the entire string.
*/

void	SendSlowString(char * s)
{
	short	numChars;
	char	*p;				// point into the string to send
	long	t,t1;
	
	p = s;
	numChars = *p++;			// get the length of the string
	
	t = TickCount();			// get the current time
	t1 = t;
	while (numChars--)		{
		SendChar(p);
		p++;
		while ((t1 - t) < CHARPAUSE)	{
			t1 = TickCount();
		}
		t = t1;
	}
}

/* * * * * * * * * *
 * S E N D C H A R *
 * * * * * * * * * *
  
 Send a single character (synchronously). 
  
*/

OSErr	SendChar(char *c)
{
	OSErr	osstat;
	IOParam	myPB;
	long	t;
	EventRecord theEvent;
	
	if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }
	InitPB(&myPB, outRefNum);
	myPB.ioBuffer = c;
	myPB.ioReqCount = 1;
	osstat = PBWriteAsync(&myPB);		// write out the data
	t = TickCount();
	while (myPB.ioResult>0)		{
		if (TickCount() - t > 2) {	// or we time out
			histlog("pbWTimeO",(long) *c,t);
			KillIO(outRefNum);
			break;
		}
	}
	histlog("sentChar",*c,t);
	return (myPB.ioResult);
}

/* * * * * * * * *
 * G E T C H A R *
 * * * * * * * * *
  
 Get a single character. Return eofErr if no char is really there, otherwise, return noErr
  
*/
OSErr	GetChar(char *c)
{
 	IOParam		aPB;			// param block used here
	OSErr		osstat;
	long		t,numChars;
	EventRecord theEvent;
	
	if(MyWaitNextEvent(1<<nullEvent,&theEvent,1,NULL)) { }	
	*c = '?';				// give 'em something to chew on if timeout
	osstat = SerGetBuf(inRefNum, &numChars); // check input driver
	if (numChars == 0) return (eofErr);	// nothing to return 
	
	InitPB(&aPB, inRefNum);			// set up the local PB
 	aPB.ioBuffer = c;
 	aPB.ioReqCount = 1;			// get one character
	osstat = PBReadAsync(&aPB);	
	t = TickCount();
	while (aPB.ioResult>0) { 		// spin loop 'til it's been sent
		if (TickCount() - t > 2) {	// or we time out
			KillIO(inRefNum);
			histlog("pbRdTimO",0,0);
			break;
		}
	}
	histlog("got char",*c,t);
	return(aPB.ioResult);
}
		
/***********************
 * O U T C O N T R O L *
 ***********************
 
 OutControl sends the specified csCode, with the specified csParam to the 
 output driver. It's a synchronous operation, on a PB which is a local param.
 
*/

OSErr OutControl(short csCod, short csParm)
{
	CntrlParam 	pb;

	pb.ioCompletion = NULL;				// no completion routine
	pb.ioCRefNum = outRefNum;			// our output driver
	pb.csCode = csCod;				// set the csCode
	pb.csParam[0] = csParm;				//   and the csParam
	return (PBControlImmed(&pb));			// Immed call
}

/***************
 * I N I T P B *
 ***************
 
 Fill in the fields of a PB so that it can be used.
 
*/

void	InitPB(IOParam *pb, short refNum)
{
	pb->ioCompletion = NULL;		
	pb->ioResult = 0;
	pb->ioRefNum = refNum; 
	pb->ioPosMode = 0;
	pb->ioPosOffset = 0;		

}

/*****************
 * N E X T C R C *
 *****************

 NextCRC -- compute a CRC using the accumulated CRC and the char passed in

 This routine computes a CRC-16 on a stream of bytes.  It uses a
 table lookup scheme to implement a x^16 + x^15 + x^2 + 1 polynomial.
 The interested reader is referred to McNamara's Technical Aspects
 of Data Communications, second edition, pps 110-122 for an obliquely
 related discussion.

 This routine takes the storage short cut of looking up two four-bit
 values in a 16-entry table instead of one eight-bit value in a 256
 word table.  This saves a considerable amount of space (it requires 32 
 bytes instead of 512 bytes for the 8-bit lookup table).

 One pass through this routine (one character) is about 262 cycles, or
 33.45 usec on a Mac Plus.  This is a data rate of ~29,900 char/sec,
 or much faster than any RS-232 link we're likely to see.

 Entry:	short *theCRC 	CRC accumulated so far
	char  *p	the data char (8 bits)

 Exit: 	theCRC 		updated properly
 	D0-D2/A0/A1 	changed
	Other regs 	unchanged
	
 This code was taken nearly verbatim from the Async AppleTalk driver,
 and modified only to work with THINK C compiler.

*/
void	NextCRC(short *theCRC, unsigned char *p)
{

asm	{
	   bra.s	   @1
	  
@0:	DC.W	0x0000,0xCC01,0xD801,0x1400
       	DC.W	0xF001,0x3C00,0x2800,0xE401
       	DC.W	0xA001,0x6C00,0x7800,0xB401
       	DC.W	0x5000,0x9C01,0x8801,0x4400
@1:
       move.l	   p,A0		   ; point at the data byte
       move.b	   (A0),D0	   ; D0 = the data
       move.l	   theCRC,A0	   ; A0 points at the CRC accumulator 
       move.w      (A0),D2         ; D2 is the temp accumulator
       lea	   @0,A1	   ; point at CRC table
       
; first work on the least significant nibble
       move.w      D0,D1           ; make a copy of the input character in D1
       
       eor         D2,D1           ; xor the accumulator with the data char
       and         #0x0F,D1        ;    to get an index into the CRCTable
       add.w       D1,D1           ; to make a word index
       lsr         #4,D2           ; shift the CRC right four bits
       move.w      (A1,D1),D1
       eor         D1,D2           ; and mask it with the approp. table entry
       
;  and do it again for the high nibble
       move        D0,D1	   ; put data back into D1
       lsr         #4,D1           ; shift the data char right four bits

       eor         D2,D1           ; xor the accumulator with the data char
       and         #0x0F,D1        ;    to get an index into the CRCTable
       add.w       D1,D1           ; to make a word index
       lsr         #4,D2           ; shift the CRC right four bits
       move.w      (A1,D1),D1 	   ; and mask it with the approp. table entry
       eor         D1,D2
       
       move        D2,(A0)         ; remember this CRC for next time
}
}

/* * * * * * * * * *
 * P Z S T R C P Y *
 * * * * * * * * * *
  
 Copy s2 (Pascal string) into s1, leaving a PZString.
  
*/

void	PZstrcpy(char *s1, char *s2, short maxlen)
{
	short	len;
	
	len = PSLen(*s2);
	if (len < maxlen)	{
		BlockMove(&s2[1], &s1[1], PSLen(*s2));
		s1[0] = len;
		s1[len+1] = 0;
	}
}

/* * * * * * * * * *
 * P Z S T R C A T *
 * * * * * * * * * *
  
 Append s2 (PZStrings) to s1, leaving a PZString.
  
*/

void	PZstrcat(char *s1, char *s2, short maxlen)
{
	short	len;
	
	len = PSLen(*s1) + PSLen(*s2);
	if (len < maxlen)	{
		BlockMove(&s2[1], &s1[PSLen(*s1)+1], PSLen(*s2));
		s1[0] = len;
		s1[len+1] = 0;
	}
}

/* * * * * * * * * * * * * *
 * M A K E P Z S T R I N G *
 * * * * * * * * * * * * * *
  
 Make the passed (Pascal) string into a PZString by adding the NUL.
  
*/

void	makePZString(char *s1)
{
	s1[PSLen(*s1)+1] = 0;
}

/* * * * * * * * * * * * * * * *
 * G E T H I S T O R Y T E X T *
 * * * * * * * * * * * * * * * *
  
 GetHistoryText - return (in the passed handle) the text stored in the history buffer.
  
*/

#define kDebugSamples 64
#define kBytesPerLine 27
#define	kMinBufferSize kDebugSamples * kBytesPerLine

void	GetHistoryText(Handle bufHandle)
{
	long	bufSize;
	Boolean	isLastLine;
	long	lineLen;
	
	bufSize = (3*kMinBufferSize)/2;			// try for 1.5 times expected
	SetHandleSize(bufHandle, bufSize);	
	if (MemError())		{
		bufSize = kMinBufferSize;
		SetHandleSize(bufHandle, bufSize); 	// no dice, actual size
	}
	if (MemError())
		return;					// screw it - return empty handle
	
	{
		register char	*p, *q;

		HLock(bufHandle);
		q = *bufHandle; 			// q points at start
		p = q;					// p walks down the string
		isLastLine = GetFirstLine(p, &lineLen);
		while (!isLastLine)	{
			if ((p - q + kBytesPerLine) > bufSize) {
				SysBeep(1);		// bail out if too much stuff...
				break;		
			}
			p += lineLen; 
			isLastLine = GetNextLine(p, &lineLen);
		}
		lineLen = p - *bufHandle + lineLen;
		HUnlock(bufHandle);
		SetHandleSize(bufHandle, lineLen);
	}
}

// return the first line from the log. Line is ASCII; %s:8, %Lx, %Lx \r\0

Boolean	GetFirstLine(char *theLog, long *logLen)
{
	SetFirstLog();					// point at first one
	return(GetNextLine(theLog, logLen));
}

// return the next line from the log; returns TRUE if it's the last

Boolean GetNextLine(char *logLine, long *logLen)
{
	register long	*l1, *l2;
	long	val[4];
	register short i;
	Boolean	 isLast;
	register char *s, *d;
	
	isLast = GetNextLog(&val[0]);
		
	s = (char*) &val[0];
	d = logLine;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
	*d++ = *s++;
		
	sprintf(d, " %p %p\r", val[2], val[3]); // add remainder of string
	*logLen = 27;				// always 27 = 3 * (8+1)
	
	return(isLast);
}

/* * * * * * * * * * * * * * * *
 * A S S E R T D R V R O P E N *
 * * * * * * * * * * * * * * * *
  
 AssertDrvrOpen - Has two functions:
   1) it will determine whether a named DRVR is open, returning notOpenErr or noErr. 
   2) it returns the DRVR's refnum if the DRVR exists; zero otherwise
 	
 This code stolen from SerDemo.c, in the PBxxx Demo code in the Snippets folder
 of the Dev. CD Sep. 93. 
  
*/

OSErr AssertDrvrOpen (Str255 name, short *refNum)
{
	DCtlHandle	*pUTEntry;
	Ptr		pDrvr;
	OSErr		result;		/* assume not open */
	short		unitNo;
	char		*aDrvrName;
	
	/* The point here is to determine whether a driver is open, given its name.  */
	/* This allows one to check a driver to see if it's open without hard coding */
	/* its reference number. (Normally, the way to get the refNum is to open     */
	/* the driver--but that defeats the whole purpose!)                          */
	/* This is an extension of the code discussed in Tech Note #71.              */
	
enum	{drvrName = 0x12};	/* offset to driver name in 'DRVR' std. header */
#define drambased 6
#define dopened   5

	result = notOpenErr;	// assume the worst
	*refNum = 0;		// return 0 for the refnum by default
	pUTEntry = (DCtlHandle *) UTableBase;
	for (unitNo = 0; unitNo < UnitNtryCnt; unitNo++, pUTEntry++) {
		if (*pUTEntry != nil && **pUTEntry != nil) {
			if (((***pUTEntry).dCtlFlags & 1 << drambased) != 0)
				pDrvr = *(Handle) (***pUTEntry).dCtlDriver;
			else
				pDrvr = (***pUTEntry).dCtlDriver;
			
			if (pDrvr != nil) {
				aDrvrName = pDrvr + drvrName;
				if (EqualString(aDrvrName,name,false,false)) {
					/* We found the one we're after. */
					*refNum = ~unitNo;
					if (((***pUTEntry).dCtlFlags & 1 << dopened) != 0)
						result = noErr;
					break;
				}
			}
		}
	}
	
	return result;
}
/******************************************************************************
 AbortInQueue			TCL 1.1.1 DLP 9/18/91
 
 	Walk the event queue, return TRUE if a Command-'.' is there. The
 	event is removed from the queue.
 	
 ******************************************************************************/

Boolean	AbortInQueue( void)
{
	EvQElPtr	qEvt = (EvQElPtr) GetEvQHdr()->qHead;
	Boolean		foundAbort = FALSE;

	while (qEvt)
	{
			// Starting at the evtQWhat field, and EvQEl is identical
			// to an EventRecord
			
		if ( IsCancelEvent( (EventRecord*) &qEvt->evtQWhat))
		{
			// Flush this and all other key-downs from the queue
			FlushEvents( keyDownMask, 0);
			foundAbort = TRUE;
			break;
		}
		qEvt = (EvQElPtr) qEvt->qLink;
	}
	return foundAbort;
}

/******************************************************************************
 IsCancelEvent
 
 	Returns TRUE if event is a command-period keydown event. Uses 
 	the technique in TN #263 to handle international keyboards.
 	
 ******************************************************************************/


#define kMaskModifiers  0xFE00     // we need the modifiers without the
                                   // command key for KeyTrans
                        
#define kUpKeyMask      0x0080
#define kMaskASCII1     0x00FF0000 // get the key out of the ASCII1 byte
#define kMaskASCII2     0x000000FF // get the key out of the ASCII2 byte

Boolean IsCancelEvent( EventRecord *theEvent )
{
	Boolean  isCancel = FALSE;
	short    keyCode;
	long     virtualKey, keyInfo, lowChar, highChar, state, keyCId;
	Handle   hKCHR = NULL;
	Ptr 	 KCHRPtr;

	if ((theEvent->what == keyDown) || (theEvent->what == autoKey))
	{
  
		// see if the command key is down.  If it is, find out the ASCII
  		// equivalent for the accompanying key.

		if (theEvent->modifiers & cmdKey) 
		{
				// get virtual key and keycode for KeyTrans
				
			virtualKey = (theEvent->message & keyCodeMask) >> 8;
	    
		    // and out the command key and or in the virtualKey
		    
		    keyCode = (theEvent->modifiers & kMaskModifiers)  |  virtualKey;
		   
		   	// make it look like a keyup event, to prevent dead key processing
		   	
		   	keyCode |= kUpKeyMask;
		    
		    KCHRPtr = (Ptr) GetEnvirons( smKCHRCache);
		
		    if (!KCHRPtr) 
		    {
				keyCId = GetScript( GetEnvirons(smKeyScript), smScriptKeys );
				hKCHR = GetResource( 'KCHR', keyCId );
				KCHRPtr = *hKCHR;
			}
			
			if (KCHRPtr) 
			{
				state = 0;
				keyInfo = KeyTrans( KCHRPtr, keyCode, &state);
				if (hKCHR)
					ReleaseResource(hKCHR);
			}
			else
				keyInfo = theEvent->message;
		
		    lowChar = keyInfo &  kMaskASCII2;
		    highChar = (keyInfo & kMaskASCII1) >> 16;
		    if (lowChar == '.' || highChar == '.')
				isCancel = true;
		}
	}
	return isCancel;
}

/* * * * * * * * * * * * * * * * *
 * M Y W A I T N E X T E V E N T *
 * * * * * * * * * * * * * * * * *
  
 MyWaitNextEvent - calls WaitNextEvent if we have an Express or GeoPort modem
 during dialing...
  
*/

Boolean MyWaitNextEvent(short eventMask, EventRecord *theEvent, 
	long sleep, RgnHandle mouseRgn)
{
	if (hasWNE)	{
		return(WaitNextEvent(eventMask, theEvent, sleep, mouseRgn));
	}
	else 
		return (GetNextEvent(eventMask, theEvent));
}