/*************************** ModemTalk.h **************************************/ 
#define _H_ModemTalk
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	17 Jan 93 

 This file contains #defines and global variables for the ModemTalk connection. 
 
*/

#include <Retrace.h>
#include <Power.h>
#include "LAPMgrGlue.h"

#define SECOND		60
#define	HALFSECOND	30

#define kMinATalkVers	5600

// Special AALAP types
#define kIM		0x86
#define kUR		0x87
#define kIF		0x88
#define kDISC		0x99
#define kSDDP		0x01
#define kLDDP		0x02
// kIB no longer supported - it was never used anyway...
// #define kIB		0x89

// Specially framed characters
#define kEsc		0x10
#define	kXOn		0x11
#define kXOff		0x13
#define	kXOnP		0x91
#define kXOffP		0x93
#define kFrame		0xA5

#define kENQ		0x05
#define	kCR		0x0D
#define kAT		0x40
#define kSpace 		0x20

// Serial Driver control calls

#define dtrControl	16
#define	leaveOn		8
#define leaveOff	0
#define assertDTR 	17
#define forceXon 	22

#define LAPHDRLEN   5
	/* flag + LAPType + <zero data> + <2 CRC> + flag */
#define LAPFRAMING  7		/* CRC bytes could each be escaped */
#define	DDPHDRLEN   13		/* length of long DDP header */
#define	MAXLAPDATA 600
	/* IM says 603 char's for LAP Hdr + LAP data; LAP Hdr = 3 */
#define MAXAALAPFRAME ((LAPFRAMING)+(2*MAXLAPDATA))
#define MAXAACTLDATA 8
	/* IM & UR are 4 bytes; IF is 3 bytes max. These could double when escaped */
#define MAXAACTLFRAME (LAPFRAMING+MAXAACTLDATA)
#define MAXAABUF (MAXAALAPFRAME+3*MAXAACTLFRAME)
	/* input & output buffer could contain 1 AALAP data + three AALAP ctl's */
#define	noMoreFrames	-10020

#define PSLen(x) ((x) & 0x00FF)
enum { noFrame, inData, escaping };	// states of receiver FSM

enum { 	kLinkNotConnected, 		// not dialed at all
	kLinkdialing, 			// executing dialing script
	kLinkHandshaking, 		// sending AALAP<cr> & IM/UR's
	kLinkUp, 			// got an address, all's right with the world
	kLinkDropped, 			// lost contact, no resp to IM's
	kLinkSentDisc, 			// we sent kDISC packets to hang up
	kLinkHangingUp 			// executing dialer->Hangup()
	};
	
typedef struct DeferredTask {
	QElemPtr 	qLink;
	short		qType;
	short		dtFlags;
	ProcPtr		dtAddr;
	long		dtParm;
	long		dtReserved;
} DeferredTask;

/*****************
 * G L O B A L S *
 *****************

 The following variables are used by the ModemTalk connection. They are stored
 as offsets from A4, as are the globals in the atlk.h file. 
 
*/
#pragma mark ModemTalk GLOBALS

extern long	mtlkg;			// magic cookie - should be 'mtlk'
extern long	modemTalkVers;		// DLM converted to seconds since 1904 
extern short	myRouter;		// other end's number (from IM/UR)
extern short	propNode;		// LAPMgr's proposed node number
extern Boolean	hasAdrs;		// true if IM/UR handshake worked 
extern Boolean	AALAPup;		// true if VBL & LAPRead interrupt ongoing
extern short	linkState;		// current state of the link (enum linkStates)
extern short	serPortConfig;		// serial port configuration (SerReset)
extern Ptr	histBuf;		// pointer to the history log buffer
extern char	magicChar[32];		// char's to be escaped in frame

// 	Input side globals - used only by input code

extern short	inRefNum;		// refnum for the input driver
extern ProcPtr	pbReadCompl;		// proc to be used in inPB's completion routine
extern long	inGlobPtr;		// pointer to globals info
extern IOParam	inPB;			// PB for input driver
extern char	*inBuf;			// pointer to the input buffer
extern char	*inFree;		// next free character in inBuf
extern char	*inNext;		// next un-consumed char in inBuf
extern char	*inLast;		// last char of PBRead buffer
extern short	inputState;		// current state of input buffer
extern Boolean	oXOffSent;		// true if we've sent XOff
extern short	inCRC;			// accumulated CRC
extern short	rcvdLen;		// count of de-escaped chars
extern char	*interruptBuf;		// Serial driver's interrupt buffer

// 	Output side globals - used only by output code

extern short	outRefNum;		// refnum for the output driver
extern Boolean	nIM;			// true if we should send IM frame
extern Boolean	nUR;			// true if we should send UR frame
extern short	numDisc;		// number of disc frames to be sent
extern long	outGlobPtr;		// pointer to globals info
extern IOParam	outPB;			// PB for output driver
extern char	*outBuf;		// pointer to the output buffer
extern char	*outFree;		// next free character in outBuf
extern short	IFSeqLen;		// length of the InterFrame sequence
extern char	interFrame[2];		// data to be sent between frames
extern short	outCRC;			// output side CRC

// 	VBL globals - Timers and such...

extern long	tStartedIM;		// time we started sending IM's
extern long	tLastIM;		// last time we last sent an IM
extern long	tLastRx;		// last time we received complete frame
extern long	tLastRTMP;		// last time we sent up an RTMP pkt
extern long	tXOffHold;		// last time we received an XOff
extern long	tNextDisc;		// next time to send a DISC frame
extern long	tStartedNotif;		// time we put up the LinkDown notif
extern short	tVBLtask;		// # ticks 'til next vbl task call
extern long	tStartedConn;		// time we started this connection
extern long	tLastVBL;		// time we last did "major" vbl task

extern VBLTask	myVBL;			// the actual VBL queue element
extern IOParam	vblPB;			// used to send frames at VBL time
extern char	*vblBuf;		// pointer to buffer for VBL's AACtl's
extern DeferredTask inDTask;		// deferred task for LAPReadCompl
extern DeferredTask outDTask;		// deferred task for LAPWriteCompl
extern NMRec	theNotifRec;		// used to notify when link went down...
extern char	notifMsg[256];		// the loss-of-link notification message
extern SleepQRec mySleepQElem;		// the sleep queue element
extern VBLTask	notifVBL;		// VBL to tear down the notification
extern Boolean	hasExpModem;		// true if we really have a Express/GeoPort modem
extern Boolean	hasWNE;			// true if WaitNextEvent is implemented
extern short	prevNet;		// previous network (before sudden disc)
extern short	prevNode;		// previous node number

//	Statistics - only updated one place in the code

extern long	numIMtx;		// # IM's transmitted
extern long	numURtx;		// # UR's transmitted
extern long	numDataTx;		// # data frames transmitted
extern long	numPBWrites;		// # PBWrite calls
extern long	numTxErrs;		// # times PBWrite failed
extern long	numTxTO;		// # times LAPWrite timed out during IM/UR
extern long	numForcedXOn;		// # times we forced XOn
extern long	numENQs;		// # times LAPEnq called
extern long	numNodeReject;		// # times we called LSetInUse

extern long	numLAPReads;		// # times LAPReadCompl called
extern long	numIMrx;		// # IM's received
extern long	numURrx;		// # UR's received
extern long	numIFrx;		// # IF's received
extern long	numDISCrx;		// # DISC's received
extern long	numDataRx;		// # data frames received
extern long	numRxCRCErr;		// # received CRC errors
extern long	numShortRx;		// # short frames received
extern long	numLAPframes;		// # good LAP Frames received
extern long	numUnkLAP;		// # times we received unknown LAPtype
extern long	numLongLAP;		// # LAP frames that were too long
extern long	numFrameStarts;		// # opening kFrames seen

extern long	numVBL;			// # times vbl task entered

/* * * * * * * * * * * *
 * P R O T O T Y P E S *
 * * * * * * * * * * * * 
 
*/

#pragma mark Prototypes for MTalkRead.c

void	LAPReadCompl(void);
void	DeferLAPReadCompl(void);
OSErr	UnEscape(void);
void	ProcessFrame(void);
void	ClearReceiver(void);

#pragma mark Prototypes for MTalkWrite.c

long 	FrameAALAPCtls(char * theBuf);
void	CheckRxTimer(void);

#pragma mark Prototypes for MTalkLink.c

void	LinkWentDown(void);
pascal long isNetValid(TNetworkTransition* theTrans, unsigned long theAdrs);

#pragma parameter __D0 PBStatusImmed(__A0)
pascal OSErr PBStatusImmed(CntrlParam *paramBlock)
    = 0xA205; 

#pragma parameter __D0 PBControlImmed(__A0)
pascal OSErr PBControlImmed(CntrlParam *paramBlock)
    = 0xA204; 

#ifdef	TESTAPPL
	#define main atlkmain
	extern void  atlkmain(void);
#else
	extern void  main(void);
#endif
