/* Popup control definition function.
 *
 *	by Paul Merchant, Jr.		January 1993
 *
 *	Copyright  1993 Trustees of Dartmouth College
 */

#ifndef THINK_C
#include <PLStringFuncs.h>
#include <Menus.h>
#include <Types.h>
#include <QuickDraw.h>
#include <Controls.h>
#include <ToolUtils.h>
#include <Memory.h>
#include <Resources.h>
#else
#define PLstrcpy(s1, s2)	BlockMove(&s2, &s1, s2[0] + 1)
#define PLstrlen(s)			(s[0])
#endif

#define kLeftMargin			14
#define kRightMargin		6
#define kVerticalMargin		1
#define kArrowMargin		2
#define kShadowInset		3
#define kArrowHeight		6
#define kArrowWidth			12

#define SysFontFam			((Ptr) 0xba6)	/* a byte that is the system font */
#define SysFontSize			((short *) 0xba8)	/* a word that is the system font size */

typedef struct t_private {
	MenuHandle		mh;
	short			menuID;
	PicHandle		downArrow;
	Boolean			ReadMenu;		/* True if we created the menu, false otherwise */
	Boolean			fixedWidth;		/* True if the control should be a fixed width */
	short			fontNum;
	short			fontSize;
	Style			fontStyle;
	short			titleWidth;
	short			just;			/* How to justify the title */
} t_private;

#define GetCData(ch) 	((t_private **) (*ch) -> contrlData)

static void doDraw(ControlHandle ch, long part);
static void doInit(ControlHandle ch, short resFlags);
static long doTest(ControlHandle ch, Point where);
static void doCalcRgn(ControlHandle ch, RgnHandle rgn, Boolean thumbRgn);
static void doDispose(ControlHandle ch);
static long	doText(ControlHandle ch);
static void doAction(ControlHandle ch);


pascal long main(short varCode, ControlHandle theControl,
					short message, long param) {
	long	result = 0;
	
	switch(message) {
		case drawCntl:	doDraw(theControl, param);					break;
		case testCntl:	result = doTest(theControl, *(Point *) &param);	break;
		case calcCRgns:
			if (param & 0x80000000) {
				param &= 0x00ffffff;
				doCalcRgn(theControl, (RgnHandle) param, true);
			}
			else 
				doCalcRgn(theControl, (RgnHandle) param, false);
			break;
			
		case calcCntlRgn:
			doCalcRgn(theControl, (RgnHandle) param, false);					
			break;	
			
		case calcThumbRgn:
			doCalcRgn(theControl, (RgnHandle) param, true);					
			break;
			
		case initCntl:	doInit(theControl, varCode);		break;
		case dispCntl:	doDispose(theControl);				break;
		case posCntl:	break;
		case thumbCntl:	break;
		case dragCntl:	break;
		case autoTrack:	doAction(theControl);				break;
		default:		break;
	}
	
	return result;
}


static void doDraw(ControlHandle ch, long) {
	PenState		ps;
	GrafPtr			SavedPort;
	t_private 		**private = GetCData(ch);
	Rect			r = (*ch) -> contrlRect;
	Str255			s;
	Boolean			elipsis;
	short			savedFont, savedSize;
	Style			savedStyle;
	short			elipsisWidth;
	FontInfo		fi;
	Point			p;
	MenuHandle		mh;
	short			itemWidth;
	short			lineHeight;
	Style			itemStyle;

	if (!private) return;
	if (!(*ch) -> contrlVis) return;
	
	mh = GetMHandle((*private) -> menuID);
	
	GetPort(&SavedPort);
	SetPort((*ch) -> contrlOwner);
	savedFont = (*ch) -> contrlOwner -> txFont;
	savedSize = (*ch) -> contrlOwner -> txSize;
	savedStyle = (*ch) -> contrlOwner -> txFace;
	GetPenState(&ps);
	PenMode(patCopy);

/*	
	if ((*ch) -> contrlHilite == 255)
		PenPat(qd.gray);
	else
		PenPat(qd.black);
*/	
	TextFont((*private) -> fontNum);
	TextSize((*private) -> fontSize);
	TextFace((*private) -> fontStyle);
	EraseRect(&r);
	
	CalcMenuSize(mh);
	GetFontInfo(&fi);

	lineHeight = fi.leading + fi.descent + fi.ascent;
	
	r.bottom = r.top + lineHeight + 2 * kVerticalMargin;
	
	/* Compute the portion of the title that will fit: */
	
	PLstrcpy(s, (*ch) -> contrlTitle);
	
	elipsis = (StringWidth(s) > (*private) -> titleWidth);
	
	if (elipsis) {
		elipsisWidth = CharWidth('');
		
		while((StringWidth(s) > (*private) -> titleWidth - elipsisWidth)
				&& (PLstrlen(s) > 0))
			s[0] -= 1;
	}
	
	MoveTo(r.left, r.top + fi.ascent + kVerticalMargin);
	DrawString(s);
	if (elipsis) DrawChar('');
	
	
	/* Compute the portion of the selected item that will fit:
	 * That's everything within the margins and beside the arrow.
	 */
	
	r.left += (*private) -> titleWidth;
	itemWidth = r.right - r.left - kLeftMargin - kRightMargin
									- kArrowMargin - kArrowWidth;
	
	GetItem(mh, (*ch) -> contrlValue, s);
	GetItemStyle(mh, (*ch) -> contrlValue, &itemStyle);
	TextFace(itemStyle);
	
	elipsis = (StringWidth(s) > itemWidth);
	
	if (elipsis) {
		elipsisWidth = CharWidth('');
		
		while((StringWidth(s) > itemWidth - elipsisWidth)
			&& (PLstrlen(s) > 0))
		s[0] -= 1;
	}
	
	MoveTo(r.left + kLeftMargin, r.top + fi.ascent + kVerticalMargin);
	DrawString(s);
	if (elipsis) DrawChar('');
	
		
	if (!(*private) -> fixedWidth) {
		GetPen(&p);
		r.right = p.h + kArrowWidth + kArrowMargin + kRightMargin;
	}
	
	FrameRect(&r);
	
	MoveTo(r.left + kShadowInset, r.bottom);
	LineTo(r.right, r.bottom);
	LineTo(r.right, r.top + kShadowInset);
	
	r.right -= kRightMargin;
	r.left = r.right - kArrowWidth;
	r.top = (r.top + r.bottom ) / 2 - kArrowHeight/2;
	r.bottom = r.top + kArrowHeight;
	
	DrawPicture((*private) -> downArrow, &r);
	
	SetPenState(&ps);
	TextFont(savedFont);
	TextSize(savedSize);
	TextFace(savedStyle);
	SetPort(SavedPort);
}

static long doTest(ControlHandle ch, Point where) {
	t_private 		**private = GetCData(ch);
	Rect			r = (*ch) -> contrlRect;

	if (!private) return 0;
	if (!(*ch) -> contrlVis) return 0;
	if ((*ch) -> contrlHilite == 255)	return 0;
	
	if (PtInRect(where, &r)) return 1;
	
	return 0;
}


static void doCalcRgn(ControlHandle ch, RgnHandle rgn, Boolean thumbRgn) {
	t_private 		**private = GetCData(ch);
	Rect			r = (*ch) -> contrlRect;

	if (thumbRgn) {
		if (private)
			r.left += (*private) -> titleWidth;
	}
	else {
		if (private)
			r.right = r.left + (*private) -> titleWidth;
	}

	RectRgn(rgn, &r);
}


#define kDownArrowPict			9000

static void doInit(ControlHandle ch, short resFlags) {
	short			menuID = (*ch) -> contrlMin;
	short 			flags = (*ch) -> contrlValue;
	short			width = (*ch) -> contrlMax;
	t_private		**private;
	MenuHandle		mh;
	PicHandle		p;
	Style			st;
	
	/* Get our private storage */
	
	if (!(private = (t_private **) NewHandle(sizeof(t_private))))
		goto error1;
	
	
	/* Get the Menu to display */
	
	if (mh = GetMHandle(menuID)) {
		(*private) -> ReadMenu = false;
	}
	else if (mh = GetMenu(menuID)) {
		(*private) -> ReadMenu = true;
		InsertMenu(mh, -1);
	}
	else
		goto error2;
	
	(*private) -> menuID = menuID;
	(*private) -> mh = mh;
	
	/* Get the picture of the popup arrow */
	
	if (!(p = GetPicture(kDownArrowPict))) goto error3;
	DetachResource((Handle) p);
	HNoPurge((Handle) p);
	
	(*private) -> downArrow = p;
	
	
	/* Set the control title flags */
	
	(*private) -> just = (short) ((char ) (flags & 0xff));
	
	st = 0;
	
	if (!(flags & popupTitleNoStyle))
		st = (Style) ((char) ((flags & 0xff00) >> 8));
	
	(*private) -> fontStyle = st;
	
	(*private) -> fixedWidth = resFlags & popupFixedWidth ? true : false;
	
	if (resFlags & popupUseWFont) {
		(*private) -> fontNum = (*ch) -> contrlOwner -> txFont;
		(*private) -> fontSize = (*ch) -> contrlOwner -> txSize;
	}
	else {
		(*private) -> fontNum = 0;
		(*private) -> fontSize = 0;
	}
	
	if (resFlags & popupUseAddResMenu) 
		AddResMenu(mh, (ResType) (*ch) -> contrlRfCon);
	
	(*private) -> titleWidth = width;
	
	(*ch) -> contrlData = (Handle) private;
	
	(*ch) -> contrlValue = 1;
	(*ch) -> contrlMin = 1;
	(*ch) -> contrlMax = CountMItems(mh);
	(*ch) -> contrlAction = (ProcPtr) -1L;
	return;
	
error3:	if ((*private) -> ReadMenu) {
			DeleteMenu(menuID);
			DisposeMenu(mh);
		}
error2:	DisposeHandle((Handle) private);
error1:	(*ch) -> contrlData = NULL;
	
}


static void doDispose(ControlHandle ch) {
	t_private **private = GetCData(ch);
	
	if (!private) return;
	
	DisposeHandle((Handle) (*private) -> downArrow);

	if ((*private) -> ReadMenu) {
		DeleteMenu((*private) -> menuID);
		DisposeMenu((*private) -> mh);
	}
	DisposeHandle((Handle) private);
	(*ch) -> contrlData = NULL;
}


static void doAction(ControlHandle ch) {
	t_private 		**private = GetCData(ch);
	GrafPtr			SavedPort;
	Rect			r = (*ch) -> contrlRect;
	Point			p;
	long			selectResult;
	
	if (!private) return;
	if (!(*ch) -> contrlVis) return;
	if ((*ch) -> contrlHilite == 255)	return;
	
	GetPort(&SavedPort);
	SetPort((*ch) -> contrlOwner);

	r.right = r.left + (*private) -> titleWidth;
	p.h = r.right;
	p.v = r.top;
	
	LocalToGlobal(&p);
	InvertRect(&r);
	
	CheckItem((*private) -> mh, (*ch) -> contrlValue, true);
	CalcMenuSize((*private) -> mh);		
	selectResult = PopUpMenuSelect((*private) -> mh, p.v, p.h, (*ch) -> contrlValue);
	CheckItem((*private) -> mh, (*ch) -> contrlValue, false);
	
	InvertRect(&r);
	
	if (selectResult) {
		selectResult &= 0xffff;
		
		if (selectResult != (*ch) -> contrlValue) SetCtlValue(ch, selectResult);
	}
	
	SetPort(SavedPort);
}
