/******************************* Test.c ***************************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	21 Jan 93 

 This file is a main routine which can be used to debug the atlk code of the 
 ADEV. It presents a text-window interface, and allows a person to enter
 commands which will be turned into calls on the various atlk routines.
 
 Several changes are required to switch between the debugging application
 version and the adev/atlk code resources. To make an application:
 
 	Change the "main" of the atlk resource to "atlkmain"
 	Comment out any calls to SetupA4(), RememberA0(), and RestoreA4()
 	Build with the "FakeLAPMgr.c" file, instead of the "LAPMgr.c" file
 	Use a project set up to build an application.
 
 To switch to an atlk code resource, undo the changes above. The first two 
 points can be done automatically by un/defining the TESTAPPL constant in the 
 atlk.h #include file.
 
 When running as an application, the ADEV globals will simply be part of the
 application's globals. 
 
 */
 
 
 /* * * * * * 
  * M A I N *
  * * * * * *
  
  This routine is a framework for sending off commands to the atlk code. It 
  allows the developer to build up sequences of calls which exercise the
  atlk. The ModemTalk test does this:
  
  	Call the AInstall code of the atlk code
  	Send a BroadCast Echo to a particular network number
  	Call the AShutDown code
  
*/
 
#include <stdio.h>
#include <string.h>
#include <console.h>
#include <Retrace.h>
#include <AppleTalk.h>
#include "atlk.h"
#include "LAPMgrGlue.h"
#include <CObject.h>

#define SECOND 60

#pragma Mark #Prototypes

OSErr	DoAInstall(void);
void	DoPoll(void);
OSErr	DoAShutdown(void);
void	PollLoop (short seconds);
void	LAPHook(long hookType, WDSElement *wds, long dest);
void	DoBcstEcho(short theNet);

	Ptr		bogusMPP;		// the .MPP globs
	WDSElement	wds[5];
	Ptr		DDPHdr;			// globals to make a WDS 
	Ptr		DDPData;		// 

main()
{
	OSErr		retval;
	long		l;
	long		t;
  	long		dest;
  	
	InitGraf(&thePort);					/* Standard initialization calls	*/
	InitFonts();
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(NULL);					/* ??? Add a ResumeProc				*/
	InitCursor();
	MaxApplZone();
  	
	retval = DoAInstall();			// bring up the world
	
/*	bogusMPP = NewPtr(100);			// bogus MPP world*/
/*	DDPHdr = NewPtr(100);*/
/*	DDPData = NewPtr(100);*/
/*	*/
/*	dest = 0x0023;*/
/*	LAPHook(0x00FF, NULL, dest); 		// fire off initial LAPEnq*/
/*	*/
/*	PollLoop(5);				// poll for five seconds*/
/**/
/*	DoBcstEcho(0x0123);			// send broadcast Echo to net 0x123*/
/*	*/
/*	PollLoop(5);				// poll for five more seconds*/

	retval = DoAShutdown();			// and tear down the connection
 
 	DisposPtr(bogusMPP);
}
#include "ModemTalk.h"
 
/* * * * * * * * * * * * 
 * D O A I N S T A L L *
 * * * * * * * * * * * *
 
*/

OSErr DoAInstall(void)
{

	asm		{
		moveq	#AInstall,D0		// get the AInstall selector
		jsr	atlkCode		// and call to the atlkcode 
	}
}

/* * * * * * * * 
 * D O P O L L *
 * * * * * * * *
 
 Simply call the atlk's polling routine - in this case, it's ModemTalk's 
 VBL task.
 
*/

void DoPoll(void)
{
	VBLtask();
}

/* * * * * * * * * * * *
 * D O S H U T D O W N *
 * * * * * * * * * * * *
 
*/

OSErr	DoAShutdown(void)
{
	asm		{
		moveq	#AShutdown,D0		// get the AShutdown selector
		jsr	atlkCode		// and call to the atlkcode 
	}
}

/* * * * * * * * * * * * 
 * D O A I N S T A L L *
 * * * * * * * * * * * *
 
*/

void PollLoop(short secs)
{
	long	t;
	
	t = TickCount();
	while ((TickCount() - t) < secs * SECOND) { // spin loop for a while
		DoPoll();			  // to let the atlk init itself
	}
}

/* * * * * * * * * * * * 
 * D O A I N S T A L L *
 * * * * * * * * * * * *
 
*/

void	LAPHook(long hookType, WDSElement *wds, long dest)
{
	
	asm				{	// fill the registers
		move.l		hookType,D0	// FF = LAPEnq; 0 = Atalk pkt
		move.l		dest,D2
		lea		@1,A0
		move.l		wds,A1
		move.l		bogusMPP,A2
		jsr		LAPWriteCode	
	  @1:
	}
}

/* * * * * * * * * * * * 
 * D O A I N S T A L L *
 * * * * * * * * * * * *
 
*/

void	DoBcstEcho(short theNet)
{
	Ptr		p;
	
	wds[0].entryLength = 16;		// build up the WDS
	wds[0].entryPtr = DDPHdr;
	wds[1].entryLength = 9;
	wds[1].entryPtr = DDPData;
	wds[2].entryLength = 0;
	
	p = DDPHdr;
	*p++ = myRouter;		// LAPDest
	*p++ = myNode;			// LAPSrc
	*p++ = 2;			// LAPType
	*p++ = 0;			// DDPLen
	*p++ = 22;			// DDPLen
	*p++ = 0;			// DDPCksum
	*p++ = 0;			// DDPCksum
	*p++ = (theNet>>8) & 0x00FF;	// DDPDnet
	*p++ = theNet & 0x00FF;		// DDPDnet
	*p++ = (myNet>>8) & 0x00FF;	// DDPSnet
	*p++ = myNet & 0x00FF;		// DDPSnet
	*p++ = 0xFF;			// DDPDnode
	*p++ = myNode;			// DDPSnode
	*p++ = 4;			// DDPDskt
	*p++ = 0x99;			// DDPSskt
	*p++ = 4;			// DDPType
	
	p = DDPData;
	*p++ = 1;			// echo command
	*p++ = 'H';			// and a byte of data
	*p++ = 'i';			// and a byte of data
	*p++ = ' ';			// and a byte of data
	*p++ = 'R';			// and a byte of data
	*p++ = 'i';			// and a byte of data
	*p++ = 'c';			// and a byte of data
	*p++ = 'h';			// and a byte of data
	*p++ = '!';			// and a byte of data
	
	LAPHook( 0, (WDSElement *)&wds, myRouter);	// LAPWrite the data pkt
}