/******************************* adev.c ***************************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	17 Jan 1993

 The adev is a code resource which is part of the 'adev' file. The adev code 
 implements any user interface which the particular connection might require.
 For full details of programming adev's, consult the Macintosh AppleTalk 
 Connections Programmer's Guide.
 
 The adev code is loaded into the application heap by the Network control panel.
 It must have the Locked bit set. The GetADEV call to the adev returns the icon 
 and string for each instance of the possible connections the adev supports. 
 When the user selects an adev by clicking its icon, the Network CP issues a 
 SelectADEV call. Clicking on an already selected adev icon causes the Network 
 CP to send a ReselADEV call to the adev. When the Network CP closes, it
 releases all adev resources it loaded.
 
 The resource ID of the adev resource must match that of its associated atlk
 resource, and must be in the range 128..254. Apple Computer reserves the
 range 1..127. The resource ID must be assigned by Apple's Developer Support.
 The Dartmouth ModemTalk adev has been assigned 158 (0x9E).
 
 The resource ID is kept in the low eight bits of a 32-bit value associated with
 this port/adev. It is stored in parameter RAM. The three most-significant bytes
 are available for the adev, and may be used for non-volatile storage of 
 configuration, settings, etc.

 An adev file must contain the following resources:
 	BNDL	-4032	to tie together the ICN# & FREF resources
 	FREF	-4032	to indicate files
 	ICN#	-4032	icon to display in the Network control panel
 	STR 	-4032	string to display under the icon in the Network CP
 	adev	1..254	adev code for a connection
 	atlk	1..254	atlk code for a connection
	xtlk	-4032 	a 32-bit configuration value for extended adev's only

 An adev written for AppleTalk v56 or greater may expect to work with 
 System 6.0.4 or later.
 
 There are three entry points to the adev code, determined by a selector in 
 register D0 on entry: 
 
 GetADEV	The Network control panel is requesting information about the 
 				connection(s) supported by the adev.
 SelectADEV &	The user has selected (or reselected) an instance of the 
 ReselADEV	adev. These two calls should present any user interface and
 		create a value to store in the parameter RAM for this 
 		connection. The ReselADEV message is typically used by phase 2
 		adev's to display a zone selection dialog.
 
*/ 

#include "adev.h"
#define DEBUG 0
#define PSLen(x) ((x) & 0x00FF)

/***********
 * M A I N *
 ***********

 The main routine of the adev code examines the selector (passed in D0) and 
 dispatches to the proper routine.
  
*/

main()
{
	adevRegs	regs;

// offsets to the regs[] array...
#define rD0 regs.r[0]
#define rD1 regs.r[1]
#define rD2 regs.r[2]
#define rA0 regs.r[3]

	asm	{
		movem.l	D0/D1/D2/A0,regs.r[0]
	}
	
	switch (rD0 & 0x00FFL)	{
	
	case GetADEV:
		doGetADEV(&regs);
		break;
	
	case SelectADEV:
		doSelADEV(&regs);
		break;
	
	case ReselADEV:
		doReselADEV(&regs);
		break;
	
	default:
 		if (DEBUG) DebugStr("\p; printf \"Unknown!!! %d\" rd0");
		break;
	}
	
	asm { 
		movem.l	regs.r[0],D0/D1/D2/A0
	}
	
}

#undef 	rD0
#undef	rD1
#undef	rD2
#undef	rA0

// offsets within regs for routines called from adevCode
#define rD0 (*regs).r[0]
#define rD1 (*regs).r[1]
#define rD2 (*regs).r[2]
#define rA0 (*regs).r[3]

		
/*********************
 * D O G E T A D E V *
 *********************

 The GetADEV code returns values which describe a connection which can be made
 through this adev. These values include the string to display beneath the icon
 and the parameter RAM value for that icon. Some adev's support multiple 
 connections: consequently, the Network CP will call the adev several times. 
 
 
 The GetADEV code receives and returns the following values. 
 
 On entry:	
 	D1 =	value of parameter RAM for the currently selected adev
 	D2 =	the previous "ADEV value" (see below) or zero if it's the first call 	
 On exit:
 	D2 =	an "ADEV value" to associate with this connection
 	A0 ->	a Pascal string to display under the icon for this connection
 	D0 = 	a flag to indicate whether this connection has been selected, 
 		and to indicate whether there are more connections supported 
 		by the adev
 	   = -1 if this is the currently selected connection
 	   =  0 if this is a supported connection, but not currently selected
 	   =  1 if this is not a supported connection (if D2 was 0) or if there 
 	   	are no additional supported connections (if D2 was non-zero).
 
 The GetADEV should examine the ADEV value it receives in D2, and determine the
 "next" value in its sequence. If D2 is zero, the GetADEV is being called the
 first time, and it should return the first ADEV value. If D2 = the first ADEV
 value, then GetADEV should return its second ADEV value, and so on. For
 ADEVs which support NuBus slot connections, it is typical for the ADEV to check
 NuBus slots betinning with slot zero. When a supported connection is found,
 return the slot number of the supported card. When the ADEV is called again, 
 then start searching at the next slot.
 
 When the GetADEV code is called with the value of its last supported ADEV (in 
 D2), it should return the flag in D0 set to 1, indicating that this is not a 
 supported connection and that there are no other connections.
 
 The GetADEV code also returns a pointer to the string to display under that
 connection's icon, and a flag indicating whether the returned ADEV value 
 matches the value in the parameter RAM. 
 
*/

void	doGetADEV(adevRegs *regs)
{
		 	void	*retStr;
	register	long	thisADEV;
	register 	long	retFlag;
			long 	lPRAM;
			StringHandle strH;
			
	if (DEBUG) DebugStr("\pGetADEV ; dm ra3");
	
	lPRAM = rD1 & 0xFFFFFF00L;
	switch (rD2)
	{
	case 0:
		thisADEV = (long) phoneValue;		/* Phone port return code */
		strH = GetString(-4032);		/* get the name */
		HLock(strH);
		retStr = *strH;				/* and our name is... */
		retFlag = (phoneValue == lPRAM) ? -1L : 0L; /* we handle at least one... */
		break;
		
	case phoneValue:
		thisADEV = (long) noValue;	/* return the noValue */
		retStr = NULL;			/* pointer to Printer string */
		retFlag = 1L; 			/* return 1 so they won't ask again */
		break;
		
	case printerValue:								
		thisADEV = (long) noValue;		/* return noValue */
		retStr = NULL;				/* (non) pointer */										
		retFlag = 1L;				/* return 1 (don't ask again) */
		break;
	
	default:										/* we won't get here */
		retFlag = 1;
		retStr = NULL;
		break;
	}
	rD2 = thisADEV;
	rA0 = (long) retStr;
	rD0 = retFlag;
}

/*********************		*************************
 * D O S E L A D E V *	  and	* D O R E S E L A D E V *
 *********************		*************************

 doSelADEV and doReselADEV are called when the user clicks an adev in the
 Network control panel. The first is called to select a new adev, the second 
 when the user clicks the currently selected adev. These calls return with a 
 (possibly new) value to set into the parameter RAM for this adev.
 
 The adev may display a dialog to obtain configuration from the user. These
 calls may use this configuration information to create a value for the 
 parameter RAM. The Network CP will pass the returned value to the atlk code
 with the AInstall message.
 
 On entry, D2 contains the ADEV value returned from the associated GetADEV call.
 On exit, D1 contains the new value for the parameter RAM. 
 
*/

void	doSelADEV (adevRegs *regs)
{
	short	item;
	OSErr	osstat;
		
 	if (DEBUG) DebugStr("\pSelectADEV");
 	osstat = LinkSelADEV(regs);
 	if (!osstat) 			// if noErr
 		rD1 = rD2; 		// return the value from GetADEV call
 	else 	rD1 = badValue;		// else error (atlk will reject it...)
}


void	doReselADEV (adevRegs *regs)
{
	short	item;
	OSErr	osstat;
	
 	if (DEBUG) DebugStr("\pReselADEV");
	
 	osstat = LinkReselADEV(regs);
 	if (!osstat) 			// if noErr
 		rD1 = rD2; 		// return the value from GetADEV call
 	else 	rD1 = badValue;		// else error (atlk will reject it...)
}