/******************************* atlk.h ***************************************/ 
#define _H_atlk

// This #define must be *commented out* to build the atlk code resource; 
// it should be defined if to make a debugging application.
 
//#define TESTAPPL 1

#define AInstall		1
#define AShutdown		2
#define	AGetInfo		3
#define	AGetMultiCast		4
#define	ASetMultiCast		5
#define	ADelMultiCast		6
#define	AOpen			7
#define	AClose			8
#define	AAddNode		9
#define	ADelNode		10
#define	AGetNodeRef		11

#define kMinIntStackSpace	0x0400L /* min size of stack at interrupt time */

typedef struct AGetInfoRec {	// returned from AGetInfo call
	short	version;	// version of the adev/atlk; can be 2 or 3
	short	length;		// length of this record (bytes)
	long	speed;		// speed of link in bits/sec
	Byte	BandWidth;	// link speed weight factor
	Byte	reserved1;	// set to zero
	Byte	reserved2;	// set to zero
	Byte	reserved3;	// set to zero
	Byte	flags;		// see description on p 45 of M.NW.AppleTalk2
	Byte	linkAddrSize;	// size of link addr in bytes
	char	linkAddress[6];	// the link address
	short	maxnodes;	// maximum number of nodes 
} AGetInfoRec, *AGetInfoPtr;

typedef struct	{		// define an array of longs to hold entry regs
	long	r[15];
} tRegs;	
	
/*****************
 * G L O B A L S *
 *****************

 The following variables are global to the atlk resource. We use Think C's 
 scheme for allocating globals relative to register A4. The LAPWrite and atlk 
 command code are each responsible for saving entry registers and then setting
 A4 to point at these globals. 
 
*/

extern long	atlkg;			// should be 'atlk' 
extern short	adevVersion;		// we tell the adev/Network CP this version
extern Ptr	mppVars;		// .MPP variable pointer
extern Ptr	portUse;		// pointer of the portUse byte for this porit
extern short	myPort;			// our port as assigned by LAP Manager
extern short	addrRefNum;		// for the (only) address on this port
extern long	LAPWrtRtnRegs[4];	// return reg's after async LAPWrite
extern Boolean	AOpened;		// True after AOpen; false after AClose
extern Boolean  MPPup;			// True if AOpen'ed & rcvd LAPEnq
extern Boolean	MTInserted;		// True if we issued LWrtInsert
extern Boolean  hasGestalt;		// True if we have Gestalt trap
extern short	mppVersion;		// the AppleTalk stack's version
extern short	sysVersion;		// System software version number
extern short	myNet;			// net we're attached to
extern short	myNode;			// our node number
extern long	myBPS;			// current data rate
extern Boolean  useDebugger;		// true if we may trap to the debugger
extern Boolean	deferLAPWrtRtn;		// true if we should defer LAPWrtRtn 
					//  there isn't enough stack space
extern Boolean	hasDTInstall;		// true if we can use DTInstall

/* atlk entry points */

void	LAPWriteCode(void);	// LAPWrite "hook" which writes the packet
void	atlkCode(void);		// entry point for commands to the atlk

OSErr	atlkInstall 	(tRegs * regs);		// AInstall code
OSErr	atlkShutdown 	(tRegs * regs);		// AShutdown code
OSErr	atlkGetInfo 	(tRegs * regs);		// AGetInfo code
OSErr	atlkGetMultiCast (tRegs * regs);	// AGetMultiCast code
OSErr	atlkSetMultiCast (tRegs * regs);	// ASetMultiCast code
OSErr	atlkDelMultiCast (tRegs * regs);	// ADelMultiCast code
OSErr	atlkOpen 	(tRegs * regs);		// AOpen code
OSErr	atlkClose 	(tRegs * regs);		// AClose code
OSErr	atlkAddNode 	(tRegs * regs);		// AAddNode code
OSErr	atlkDelNode 	(tRegs * regs);		// ADelNode code
OSErr	atlkGetNodeRef 	(tRegs * regs);		// AGetNodeRef code

/* Auxiliary entry points 

These routines provide atlk-specific behavior, and must be implemented by the
atlk programmer. The following routines are defined:

LAPWrt is code to send an AppleTalk packet out the link. It is called with the 
	same registers as LAPWriteCode; it returns 0 or an 0SErr if the write 
	operation has completed. It returns 1 to indicate that the operation is 
	still in progress. If the operation is asynchronous (still in progress)
	it is the responsibility of the LAPWrt routine to jump through the 
	address in A0 at the conclusion of the operation.
 	
LAPEnq is called to verify the passed-in AppleTalk network address. LAPEnq will
	call the LAP Mgr's SetInUse routine if it decides to reject the address. 
 
atlkInit is called to establish the atlk's link. This would include allocating
	and initializing memory for global data, dialing a modem, obtaining a 
	network address, and other housekeeping required for the particular
	ADEV. If errors occur during this routine, atlkInit should return an
	OSErr indicating the trouble.
	
atlkRemove is called to tear down the atlk link. This should undo all the
	actions of the atlkInit call.
	
atlkOther may be called to communicate information to the atlk. This might
	be done as a "back door" communication link to the atlk. The command
	to be performed is passed in rD0 as with other calls to the atlkCode.
	
LAPWrtRtn is called to as a "LAP Mgr completion" routine, signalling that the
	current LAPWrite operation has completed. It must reload all registers
	to their state at the LAPWrite call (except D1, D3 & A3), and transfer
	to the original contents of A0.
*/

OSErr	LAPWrt(tRegs *regs);	// the LAPWrite code
void	LAPEnq(tRegs *regs);	// the code to send ENQ's
OSErr	atlkInit(tRegs *regs);	// initialize the atlk code
OSErr	atlkRemove(tRegs *regs);// remove the atlk code
OSErr	atlkOther(tRegs *regs);	// connection-specific command code
void	LAPWrtRtn(void);	// return from Async LAPWrite call