/******************************* BPrefsFile.h *********************************/ 
/*
	Copyright  1993, Trustees of Dartmouth College
	
	Richard E. Brown
	
	Release date:	29 March 1993 

 Defines for a Preference File Object. Based on Jamie McCarthy's CPrefsFile.
 
*/

#define _H_BPrefsFile

class 	BPrefsFile : indirect		{

public:
// Creation and Disposal...
// IPrefsFile opens/creates a preferences file in the Preferences folder
//	of type and creator (fileType/sign). The settings are in a resource
//	type resTyp = resID. The strsID parameter is an STR# resource of 
//	useful strings.

	virtual OSErr 	IPrefsFile(OSType fileType, OSType sign,
			short strsID, ResType resTyp, short resID);
	virtual void	Dispose(void);

// LoadPrefs & SavePrefs - Read/Write the preferences to/from the prefs
//	file. LoadDefaultPrefs reads the preferences from the atlk's 
//	resource file. SaveAsSettings saves the current file as a separate
//	settings file.

	virtual OSErr	LoadPrefs(void);
	virtual OSErr	LoadDefaultPrefs(void);
	virtual OSErr	SavePrefs(void);
	virtual OSErr	SaveAsSettings(short refNum);
	
// GetAPref/SetAPref - get/set the n'th preference from the pref file. 
//	GetAPref copies into thePref; SetAPref copies from thePref 
//	GetAPref's destLen is the entire length of the buffer; the PString
//		will be as long as destLen-1 bytes long.

	virtual void	GetAPref(short n, char * thePref, long destLen);
	virtual void	SetAPref(short n, char * thePref);
	virtual	short	NumPrefs(void);	
private:

	Boolean ExistsOnDisk(Str63 filename);
	OSErr	FindPrefsFolder(void);	// find/create preferences folder
	OSErr	CreateNewPrefs(Str63 filename, OSType creator, OSType fType);
	OSErr	TryToOpen(Str63 filename);
	
// Class variables	

	short		itsApplRefNum;		// Reference number for application's resource fork
	short		itsProjRsrcRefNum;	// If run from ThC, refNum for ..rsrc file
	short		itsRefNum;		// refNum of the (real) prefs file
	short		itsStrsID;		// 'STR#' rsrc to search for the "Prefs" appendads
	OSType		itsFileType;		// Type of prefs file
	OSType		itsFileSign;		// and its signature
	OSType		itsRsrcType;		// Resource type to use to store data in prefs file
	short		itsRsrcID;		// Resource ID to use
	Handle		itsPrefRsrc;		// handle to the STR# rsrc
	short		volNum;			// volume and directory
	long		dirID;			//   of the Pref's folder
	short		numStrs;		// number of strings here
	Boolean		dirty;			// true if we've made changes
};